#include "CompoLibrary.h"
#include "GUIManager.h"
#include "utils.h"
#include "PathQualifier.h"


GlyphProc::GlyphProc(const char* proc_name,const char* descr_)
          :tcl_name(proc_name)
{ 
  if (descr_) descr = descr_;                                   //if no description given, make one from
  else {  descr = "Procedure "; descr += tcl_name;  }           //the Tcl proc name        
}

GlyphProc::~GlyphProc()                                         //Dtor. Remove this proc from the Tcl interpreter
{                                                               //We know that 'tcl_name' is the name of a Tcl command 
    char buf[200];                                              //defined in a script. Thus, to remove it from the interp,        
    sprintf(buf,"rename %s \"\"",(const char*)tcl_name);        //we execute 'rename <name> ""'. It's not terribly elegant
    GUIManager::exec(buf);                                      //but it does the removal job
}

ComponentLibrary::ComponentLibrary(const char* tcl_file_)
                 :tcl_file(tcl_file_)
{ 
}

void ComponentLibrary::loadFromTclFile()                        //Second part of constr. Loads lib in the interpreter
{                                                               //by sourcing its Tcl file.
   char buf[250];

   STRING p = PathQualifier::qualify(tcl_file);

   sprintf(buf,"source %s",(const char*)addQuotes(p));
   GUIManager::exec(buf);
}

ComponentLibrary::~ComponentLibrary()                           //Dtor. Destroy entries for all procs in this
{
   for(int i=0;i<numGlyphProcs();i++)
      delete glyph_procs[i];
}


void ComponentLibrary::addGlyphProc(const char* proc_name,const char* descr)
{                                                               //Make & add new entry for glyph-proc
   GlyphProc* p = new GlyphProc(proc_name,descr);
   glyph_procs.Add(p);
}


void ComponentLibrary::saveDefinition(FILE* fp)
{
        //!!! write all contents of this in a Tcl file...
}

void ComponentLibrary::save(FILE* fp,std::set<STRING>* tcl_procs) //Ask this lib to save cmd to restore itself
{                                                                 //in a Tcl-file. Also remove from tcl_procs the 
   fprintf(fp,"pr_compo_lib load %s\n",(const char*)addQuotes(tcl_file));    
                                                                  //Tcl-proc-names registered in this lib as compos.

   if (tcl_procs)                                                 //Remove, from tcl_procs, the Tcl-proc-names
     for(int i=0;i<glyph_procs.Count();i++)                       //registered in this lib as compos.
     {
       std::set<STRING>::iterator it = tcl_procs->find(glyph_procs[i]->getTclName());
       if (it!=tcl_procs->end()) tcl_procs->erase(it);
     }
}

//-----------  ComponentLibrarySet  --------------------------------------------------


ComponentLibrarySet::ComponentLibrarySet()                      //Ctor
                    :current_lib(0)                             //No current lib yet
{  }


ComponentLibrarySet::~ComponentLibrarySet()                     //Dtor
{  
   reset();                                                     //Clean up all loaded libs
}

void ComponentLibrarySet::reset()                               //Unload all loaded libs
{               
   while(libs.Count())                                          //The unload operation will also destroy the 
      unloadLib(0);                                             //lib-entries
}

void ComponentLibrarySet::unloadLib(int i)                      //Unload i-th loaded lib
{
   current_lib = 0;                                             //Just to be sure: we have no current-lib now since we're 
                                                                //not loading stuff but rather unloading...
   ComponentLibrary* l = libs[i];                               //Get lib        
   libs.Delete(i);                                              //Remove lib-entry from this
   delete l;                                                    //Delete lib-entry, we're done with it
}

void ComponentLibrarySet::loadLib(const char* tcl_file)         //Load a new compo-lib
{
   if (getLib(tcl_file)) return;                                //Check we're not loading the lib twice
                                                                //!!!WARNING: this test implies the tcl_file path is an unique-id
                                                                //            for the lib

   ComponentLibrary* l = new ComponentLibrary(tcl_file);        //Make a new entry for the lib
   current_lib = l;                                             //This is the lib undegoing loading
   l->loadFromTclFile();                                        //Finalize constr. of 'l' by loading data in the interpreter.
   libs.Add(l);                                                 //Store the new lib entry in this
}

ComponentLibrary* ComponentLibrarySet::getLib(const char* tcl_file)
{                                                               //Search a library by name
   for(int i=0;i<libs.Count();i++)
      if (libs[i]->getTclFile()==tcl_file) return libs[i];
   return 0;
}

void ComponentLibrarySet::addGlyphProc(const char* tcl_name,const char* descr)
{                                                               //Add a glyph proc to lib undergoing loading (if any)
   if (current_lib)
      current_lib->addGlyphProc(tcl_name,descr);
}

void ComponentLibrarySet::save(FILE* fp,std::set<STRING>* s)    //Save cmds to restore this in a Tcl-file
{                                                               //Remove from 's' the Tcl-procs registered as compos in this
   for(int i=0;i<libs.Count();i++)                              //Ask each lib to save itself
      libs[i]->save(fp,s);
}

ComponentLibrarySet::GlyphProcIterator ComponentLibrarySet::iterGlyphProcs()
{  return GlyphProcIterator(this);  }                           //Make iterator over this's glyph-procs

ComponentLibrarySet::GlyphProcIterator::GlyphProcIterator(ComponentLibrarySet* l)
                                       :libs(l),lib(0),proc(0)  //Ctor for iterator.
{  }

GlyphProc* ComponentLibrarySet::GlyphProcIterator::next()       //Iterator. Return next glyph-proc.
{
   for(;;)                                                      //Get to next glyph-proc in this's libs
   {
     if (lib>=libs->numLibs()) return 0;                        //Is 'lib' valid? If no, ready iterating.
     if (proc>=libs->getLib(lib)->numGlyphProcs())              //Is 'proc' valid? If not, go to next lib.
     { proc=0;lib++; }                                          
     else break;                                                //We have now a valid 'lib' and 'proc', so return it
   }
   GlyphProc* ret = libs->getLib(lib)->getGlyphProc(proc);      //Return current glyph-proc
   proc++;                                                      //Go to next proc in current lib
   return ret;
}



