#include "EventHandler.h"
#include "IVMapper.h"
#include "IVViewer.h"
#include "GraphManager.h"
#include "ObsList.h"
#include "Node.h"
#include "Selection.h"
#include "SelList.h"
#include "Ops.h"
#include <iostream.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoEvents.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/SoPickedPoint.h>
#include "UI_imp.h"
#include "IVUtils.h"
#include <glut.h>



static void project2Dto3D(SoUIViewer* viewer,float nx,float ny,float& wx,float& wy);
//A simple 'projector' that converts from norm-screen-coords nx,ny to world-coords (wx,wy,0)
//This needs also a viewer to infer which camera/viewport there is to be used.


const int ZBUFFER_HIT_SZ = 6;
//Size, in pixels, of the z-buffer-hit object used by EventHandler to pass pick-point info to IVViewer.
//A few pixels should be enough.


const int MOVE_THRESHOLD = 3;	
//Determines how much an object has to be moved on the screen (in pixels)
//by the event-handler to be considered that it really moved.
//Useful for filtering small unwanted moves that happen together with selection clicks.


static IVZBufferHit zbuffer_hit;
//The ZBufferHit is used for correct picking of wireframe-objs. Basically, we want to pick _through_ them to the first
//not-wireframe (solid or semi-transparent) object. Since Inventor can't do this, we simulate it by combining its
//SoRayPickAction with reading the z-values directly form the z-buffer. See the code in IVViewer::pickElement().


static void initZBufferHit(SoUIViewer* vi,const SbVec2s& sz,float nx,float ny) 
{                                                   //Initializes zbuffer-hit from position nx,ny    
    MAKE_CURRENT(vi);                               //copy a (small) portion of the zbuffer around the click
    int px=int(sz[0]*nx), py=int(sz[1]*ny);         //this'll be used for detecting clicks on wirerame-objects
    zbuffer_hit.init(px-ZBUFFER_HIT_SZ/2,py-ZBUFFER_HIT_SZ/2,ZBUFFER_HIT_SZ,ZBUFFER_HIT_SZ);
}
      


EventHandler::EventHandler(Callback cb, void* user_dt)
	     :user_cb(cb),user_data(user_dt),sweep_cb(0),sweep_data(0),
	      del_cb(0),del_sweep_cb(0),sweep_cb_on(FALSE)
{

}

EventHandler::~EventHandler()
{
   setUserCallback(0);					//relinquish our usage of the user-cb and its data
   setSweepCallback(0);					//relinquish our usage of the sweep-cb and its data
}

Node* EventHandler::getPickedNode()			//to be overriden
{  return 0; }

Edge* EventHandler::getPickedEdge()			//to be overriden
{  return 0; }

Selection* EventHandler::getPickedSelection()		//to be overriden
{   return 0;  }


//---------------  DefaultEventHandler  -------------------------------------------------

DefaultEventHandler::DefaultEventHandler(Callback cb,void* user_dt)
		    :EventHandler(cb,user_dt),
		      picked_node(0), picked_edge(0), picked_selection(0),
		      return_node(0), return_edge(0),
		      moved(FALSE),large_move(FALSE),
		      down_clicked(FALSE),rubber(FALSE)
{  }


Node* DefaultEventHandler::getPickedNode()
{   return return_node;  }

Edge* DefaultEventHandler::getPickedEdge()
{   return return_edge;  }

Selection* DefaultEventHandler::getPickedSelection()
{   return picked_selection;  }


void DefaultEventHandler::brushText(const STRING& text,IVViewer* vw, float nx,float ny)
//Draws 'text', if not "", at the normalized coords nx,ny in viewer 'vw'. This function manages the
//drawing and erasing of previously drawn text automatically by cooperating with the viewer's store_***
//data. This is called only by event().
{
    SoUIViewer*     vi = vw->getViewer();	
    const SbVec2s&  sz = vi->getViewportRegion().getWindowSize();

    MAKE_CURRENT(vi);                                   //make current he viewer's GL context
    saveGLState();                                      //save the GL state since we're changing it...
    glLogicOp(GL_COPY);
    glDisable(GL_BLEND);
    glMatrixMode(GL_PROJECTION);                        //establish a pixel coord-system in the current window       
    glLoadIdentity();
    glOrtho(0,sz[0],0,sz[1],-1,1);
           
    if (vw->rendered)                                   //invalidate store-buffer if 'rendered' by Inventor...
    { delete[] vw->store; vw->store=0; vw->store_sz=0; }
    
    if (vw->store_sz)                                   //If we have a valid store-buffer,
    {                                                   //erase whatever is in it (draw it, thus)
        
        glRasterPos2i(vw->store_x,vw->store_y);
        glDrawPixels(vw->store_dx,vw->store_dy,GL_RGB,GL_UNSIGNED_BYTE,vw->store);
    }

    STRING local_text = text;                           //Copy text locally, since we're going to
                                                        //modify it (replace newlines with \0's)
    DARRAY<char*> txts; int maxl=-1;                    //Create in txts[] the text-spans in local_text
    for(char* t=(char*)(const char*)local_text;*t;)     //delimited by newlines. maxl is the longest
    {                                                   //text-span's pixel-length.
       txts.Add(t);                                     //add current text-pos as string-start
       t=strchr(t,'\n');                                //search for next newline in text
       if (t) {*t = '\0'; t++;}                         //newline found at t. Make this the end of the previous text-span
                                                        //and increment t for next span...
       maxl = max(maxl,glutStringLength(txts[txts.Count()-1]));
       if (!t) break;                                   //no next newline found, so we're done
    }
            
    if (maxl!=-1)                                      //If anything to draw, draw it and
    {                                                  //cache it in the store-buffer for erasing later
      int px=int(sz[0]*nx), py=int(sz[1]*ny);          //Get pix-coords of text-origin
      int ln  = maxl+1;                                //Get text length, in pixels 
      int ht0 = glutStringHeight(txts[0]);             //Get text height, in pixels  
      int ht  = ht0*txts.Count()+1;		       //(the 1-offset is needed to manage GLUT's underscore...)
      ln = (px+ln>sz[0])? sz[0]-px : ln;               //Clip text-bbox to top and right to make it  
      if (py+ht>sz[1]) py -= py+ht-sz[1];              // always fit in the window. On X, we adjust length. On Y, we adjust origin
                                                       // (since glutBitmapString draws text only when _completely_ in window)
      int bt= (ln+1)*(ht+1)*3;                         //Get size of new store, in bytes
      if (vw->store_sz<bt)                             //Realloc store, if needed
      { delete[] vw->store; vw->store = new unsigned char[vw->store_sz=bt]; }
      vw->store_dx=ln; vw->store_dy=ht;                //Store the size/coords of the text to be drawn
      vw->store_x=px; vw->store_y=py;                  
                
      glReadPixels(px,py,ln,ht,GL_RGB,GL_UNSIGNED_BYTE,vw->store);
      glColor3f(0.9,0.9,0.9);                          //Draw text with a background so it shows better 
      glRecti(px,py,px+ln,py+ht);
      glColor3f(0,0,0.7);
      for(int i=0,y=py+ht-ht0+1;i<txts.Count();i++,y-=ht0) //Draw the text-spans atop of each other
         glutBitmapString(px+1,y,txts[i]);
      vw->rendered=FALSE;                              //Mark the viewer as 'rendered' into.
   }
   else                                                //Nothing to draw, so invalidate the store-buffer
   {  delete[] vw->store; vw->store=0; vw->store_sz=0;  }
            
   restoreGLState();                                    //Restore the GL state we received
   SWAP_BUFFERS(vi);                                    //Drawing ready, swap buffers
}

  

void DefaultEventHandler::event(IVViewer* vw,EVENT_TYPE e,MODIFIER_TYPE mt,float nx,float ny,int key)
{
   IVMapperBase*    m = vw->getMapper();
   GraphManager*   gm = GraphManager::getInstance();
   SoUIViewer*     vi = vw->getViewer();	
   BOOL		   mv = !(vw->getOutputOnly());
   SoSceneManager* sm = vi->getSceneManager();
   Selection*  hl_sel = gm->getSelList()->getSelection(vw->getHighlightSelection());
   const SbVec2s&  sz = vi->getViewportRegion().getWindowSize();
   BOOL         shift = ((mt & SHIFT)!=NONE);
   BOOL          ctrl = ((mt & CTRL)!=NONE);
   const char*   posx = m->getPosAttrX(), *posy = m->getPosAttrY();
   SelectionList*  sl = gm->getSelList();
   ObserverList*  obs = gm->getObsList();
   float wx,wy;
   STRING brush_txt;

   if (e==DOWN_1)					//DOWN_1 mouseclick:
   {							//       initialize selecting/deselecting or moving nodes/edges
	moved=large_move=rubber = FALSE;		//	 initialize rubberbanding
	down_clicked = TRUE;
        initZBufferHit(vi,sz,nx,ny);                    //       init zbuffer-hit for correct wireframe-obj picking
        picked_node = return_node = vw->pickNode(nx,ny,&zbuffer_hit);// remember what we picked, if anything
        picked_edge = return_edge = vw->pickEdge(nx,ny,&zbuffer_hit);	
 	nx_orig     = nx; ny_orig = ny;			//	 remember crt pos			
	if (picked_node)				//	 if we picked a node, remember its current coords
	{
	  nx_pos = picked_node->getFloat(posx);
	  ny_pos = picked_node->getFloat(posy);
	}
   }

   if (e==MOVE)						//MOVE:	 so far, we only allow moving picked nodes
   {							//       (makes sense, moving edges is so far meaningless)
	int px=int(sz[0]*fabs(nx_orig-nx)), py=int(sz[1]*fabs(ny_orig-ny));
	if (px+py>MOVE_THRESHOLD) large_move=TRUE;	//	 draw rubberband if moved enough only
	if (picked_node && mv)				//       if we move a node and moving allowed, move highlighted-nodes along
	{
	     moved = TRUE; project2Dto3D(vi,nx,ny,wx,wy);
	     float deltax = wx-picked_node->getFloat(posx);
	     float deltay = wy-picked_node->getFloat(posy);

	     if (!hl_sel->hasNode(picked_node))		//	moved node is NOT highlighted:  move it separately
               m->moveNode(picked_node,deltax,deltay);	//      ask mapper to move node's graphical representation
	     else					
             { sm->deactivate();                        //	moved node IS highlighted: move all highlight-sel
	       m->moveHighlight(deltax,deltay);		//	as a whole. This is done more efficiently by calling
	       sm->activate();				//      the mapper's moveHighlight() method than moving every node	
	     }						//	separately.
	}
        else if (down_clicked && large_move)		//	draw rubberband, erase previous, if any
	{
		MAKE_CURRENT(vi);
		SWAP_BUFFERS(vi);
                if (!rubber)                            //	1st time: save IV's GL state
                { saveGLState();rubber=TRUE;glColor3f(1,0,0); }
		else drawRubberband(nx_old,ny_old);	//	erase previous rubberband if any
		drawRubberband(nx_old=nx,ny_old=ny);	//	draw current rubberband and remember it
		glFlush();	
	}
	else if (sweep_cb_on && sweep_cb)	        //	sweep event: simply look at what's under the mouse and then call	
	{						//	the user-registered sweep-callback
             initZBufferHit(vi,sz,nx,ny);               //      init zbuffr-hit for correct wireframe-obj picking
	     return_node = vw->pickNode(nx,ny,&zbuffer_hit); 
             return_edge = vw->pickEdge(nx,ny,&zbuffer_hit);  
             brush_txt   = sweep_cb(this,sweep_data);   //      get the ret-value from the sweep-cb. This is the brush-text for us...
	}
        if (brush_txt!="" || vw->store)                 //      if we got now some text to draw or we have something still in
           brushText(brush_txt,vw,nx,ny);               //      the 'store' to restore, do it
   }

   if (e==UP_1)						//UP_1: react on mousebutton-up events
   {
	int px=int(sz[0]*fabs(nx_orig-nx)), py=int(sz[1]*fabs(ny_orig-ny));
	if (moved)				
	   if (px+py<MOVE_THRESHOLD)			//If the move was accidental, restore the node's position
	   {						//to the original one
	      picked_node->set(posx,nx_pos);	
	      picked_node->set(posy,ny_pos);
	      moved = FALSE; 
	   }
	   else						//Move was not accidental. We touch the viewer to
	      vw->touch();				//force it update (since none of its selections have changed),
	if (rubber)					//If we're in rubberband mode:	
	{
		MAKE_CURRENT(vi);			//Explicitly erase the last rubberband. Needed since there's no
		drawRubberband(nx_old,ny_old);		//update of the viewer if e.g. it has no highlight-selection.
		SWAP_BUFFERS(vi);  
		restoreGLState(); 			//set IV's GL state to whatever it was before rubberbanding
		glFlush();
							
	   	Selection* sel = new Selection;		//make new highlight-selection
                Selection* all = sl->getSelection(vw->getSelection());
		vw->pickNodes(sel,nx_orig,ny_orig,nx,ny);//stuff nodes in rubberband in highlight-sel
		Selection* tmp_sel = 0;			 //add also edges (from vw's sel) between picked nodes to highlight-sel
		SelectSetOp::edges(sel,all,tmp_sel);
		delete sel;				//delete sel, we're done with it. Now all is in tmp_sel
		if (shift || ctrl)			//if in shift or ctrl-mode, perform various operations (AND, XOR)
		{					//between the old and new highlight-selections
		   Selection* tmp_sel2 = 0;
		   Selection* hl_sel   = sl->getSelection(vw->getHighlightSelection());
		   if (shift) SelectSetOp::reunion(tmp_sel,hl_sel,tmp_sel2);
		   else       SelectSetOp::xor(tmp_sel,hl_sel,tmp_sel2);
		   delete tmp_sel;
		   tmp_sel = tmp_sel2;
		}
		tmp_sel->ref(vw->getHighlightSelection());//add tmp_sel to sel-list under name of this's highlight-sel
		picked_selection = sl->getSelection(vw->getHighlightSelection());
		return_node = 0; return_edge = 0;	//reset the picked node and edge, since we picked a whole selection	

		obs->update();		                //update possible observers and
		if (user_cb) user_cb(this,user_data);	//call user callback, if any 
	}
	else						//not in rubberband but in single-selection mode:
	if (!moved)
	{
          initZBufferHit(vi,sz,nx,ny);                  //init zbuffer-hit for coreect wireframe-obj picking
	  Node* n = vw->pickNode(nx,ny,&zbuffer_hit);	//find out if we picked a node or an edge
	  Edge* e = vw->pickEdge(nx,ny,&zbuffer_hit);

          if (hl_sel==sl->getNullSelection())		//If 'hl_sel' not pointing to a modifiable Selection yet, make one and name it
          {						//by the name the viewer uses for its highlight-selection. 
	    (new Selection)->ref(vw->getHighlightSelection());       //REMARK: this sets 'hl_sel' to the new Selection, if the name wasn't Selection::NONE,
	    hl_sel = sl->getSelection(vw->getHighlightSelection());  //        and again to the NullSelection, if the name was Selection::NONE.
	  }							     //        However we need this code to have a correct 'hl_sel' in what follows.	

	  if (hl_sel != sl->getNullSelection())		//modify the highlight selection only if not the null one
          {	
	    if (n)					//if we picked a node, toggle its selection state in the mapper and in the selection
	    { 						//for this, the viewer must have a highlight-sel
              if (shift)
              { m->highlight(n,IVMapperBase::TOGGLE); 
	        if (hl_sel->hasNode(n)) hl_sel->delNode(n);
	        else		        hl_sel->addNode(n);
              } 
	      else
	      { m->clearHighlight(); m->highlight(n,IVMapperBase::TOGGLE); 
		hl_sel->reset(); hl_sel->addNode(n); 
	      }
	    }
            else if (e) 				//if we picked an edge, toggle its selection state in the mapper and in the selection
	    { 						//for this, the viewer must have a highlight-sel
	      if (shift)
              { m->highlight(e,IVMapperBase::TOGGLE); 
	        if (hl_sel->hasEdge(e)) hl_sel->delEdge(e);
	        else		        hl_sel->addEdge(e);
              } 
	      else
	      { m->clearHighlight(); m->highlight(e,IVMapperBase::TOGGLE); 
		hl_sel->reset(); hl_sel->addEdge(e); 
	      }
	    }
 	    else					//if we didn't pick a node or an edge, simply clear the highlight-selection
	    {						//also ask the mapper to clear whatever it held selected
              m->clearHighlight();
	      Selection* sel = new Selection;
	      sel->ref(vw->getHighlightSelection());
	      picked_selection = sl->getSelection(vw->getHighlightSelection());
	    }
	  }

	  obs->update();	        		 //update possible observers
	  if (user_cb) user_cb(this,user_data);		 //call user callback, if any
	}
	moved=large_move=down_clicked=rubber=FALSE; 
	picked_node=return_node=0; picked_edge=return_edge=0; picked_selection=0;
   } //UP_1 event
   else
   if (e==KEY)                                          //KEY events:
   {
      //REMARK: Some of the keys below might look odd, (e.g. PAD_8 instead of UP_ARROW alone)
      //        However, under some Windows installations, the up-arrow key produces the
      //        PAD_8 and not the UP_ARROW symbol. Thus the multiple cases.
      
    
      SoKeyboardEvent::Key k = (SoKeyboardEvent::Key)key;  //translate key back to SoKeyboardEvent::Key-type
      switch(k)
      {
        case SoKeyboardEvent::KEY_DELETE:               //DELETE: destroy the stuff in the current hl-sel          
        case SoKeyboardEvent::PAD_DELETE:               
        case SoKeyboardEvent::BACKSPACE:
           hl_sel->flush();
           obs->update();                               //(also update the observers, since stuff was destroyed)
        break;
        case SoKeyboardEvent::LEFT_ARROW:               //LEFT ARROW: pan viewer 
        case SoKeyboardEvent::PAD_4:
           vw->pan(UIViewerBase::INCR_X); break;
        case SoKeyboardEvent::RIGHT_ARROW:              //RIGHT ARROW: pan viewer 
        case SoKeyboardEvent::PAD_6:
           vw->pan(UIViewerBase::DECR_X); break;
        case SoKeyboardEvent::UP_ARROW:                 //UP ARROW: pan viewer 
        case SoKeyboardEvent::PAD_8:
           vw->pan(UIViewerBase::DECR_Y); break;
        case SoKeyboardEvent::DOWN_ARROW:               //DOWN ARROW: pan viewer 
        case SoKeyboardEvent::PAD_2:
           vw->pan(UIViewerBase::INCR_Y); break;
        case SoKeyboardEvent::HOME:                     //HOME: view all
        case SoKeyboardEvent::PAD_7:
           vi->viewAll(); break; 
        
      }
   }  //KEY event
}


void DefaultEventHandler::saveGLState()
{
  glGetIntegerv(GL_SHADE_MODEL,&save_shade);  glShadeModel(GL_FLAT);
  glGetIntegerv(GL_LIGHTING,&save_lighting);  glDisable(GL_LIGHTING);
  glGetIntegerv(GL_LINE_WIDTH,&save_linew);   glLineWidth(1);
  glGetIntegerv(GL_DEPTH_TEST,&save_depth);   glDisable(GL_DEPTH_TEST);
  glGetIntegerv(GL_BLEND,&save_blend);        glEnable(GL_BLEND);
  glGetIntegerv(GL_LOGIC_OP_MODE,&save_logic); glEnable(GL_COLOR_LOGIC_OP); glLogicOp(GL_XOR);
  glGetFloatv(GL_CURRENT_COLOR,save_col);
 


  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,1,0,1,-1,1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


void DefaultEventHandler::restoreGLState()
{
  glShadeModel(save_shade);
  if (save_lighting) glEnable(GL_LIGHTING); else glDisable(GL_LIGHTING);
  glLineWidth(save_linew);
  if (save_depth) glEnable(GL_DEPTH_TEST); else glDisable(GL_DEPTH_TEST);
  if (save_blend) glEnable(GL_BLEND); else glDisable(GL_BLEND);
  glDisable(GL_COLOR_LOGIC_OP);
  glLogicOp(save_logic);
  glColor3fv(save_col);
}


void DefaultEventHandler::drawRubberband(float end_x,float end_y)
{
  glBegin(GL_LINE_LOOP);
  glVertex2f(nx_orig,ny_orig);
  glVertex2f(end_x,ny_orig);
  glVertex2f(end_x,end_y);
  glVertex2f(nx_orig,end_y);
  glEnd();
}


//---------------------  Local utilities  ----------------------------------------------------

void project2Dto3D(SoUIViewer* viewer,float nx,float ny,float& wx,float& wy) 
{
  //Convert 2D norm-screen-coords to 3D coords on the z==0 plane. 
  //We need this conversion to pass the real world-space x,y coords to the EventHandler callback
  //Since a 2D (screen) to 3D (world) coord conversion is not unique, we assume that the 3D coord has z == 0

  const SbViewportRegion& r = viewer->getViewportRegion();

  SbVec2f p(nx,ny);
  SbLine l;
  viewer->getCamera()->getViewVolume(r.getViewportAspectRatio()).projectPointToLine(p,l);  
  const SbVec3f &x = l.getPosition(), &v = l.getDirection();
  float k = -x[2]/v[2];
  wx = x[0] + k*v[0]; wy = x[1]+k*v[1];
}

