#include "GlyphFactoryList.h"
#include "IVGlyphFactory.h"
#include <set>
#include "utils.h"


IVGlyphFactoryList::IVGlyphFactoryList()                //Ctor: initialize
                   :num_def_ngfs(0),num_def_egfs(0)
{  init();   }	


IVGlyphFactoryList::~IVGlyphFactoryList()		//Dtor: delete all Selections in this,
{							//	since we own them.
    int i;
    for(i=0;i<ngf.Count();i++) delete ngf[i];
    for(i=0;i<egf.Count();i++) delete egf[i];
}

void IVGlyphFactoryList::init()				//Add by default one instance of every factory we have.
{
    ngf.Add(def_ngf = new IVBasicNodeGlyphFactory("basic_node_factory"));
    ngf.Add(new IVTclNodeGlyphFactory("Tcl_node_factory"));
    num_def_ngfs = ngf.Count();                         //Remember how many factories are added by init()

    egf.Add(def_egf = new IVBasicEdgeGlyphFactory("basic_edge_factory"));
    egf.Add(new IVTclEdgeGlyphFactory("Tcl_edge_factory"));
    num_def_egfs = egf.Count();                         //Remember how many factories are added by init()
}


void IVGlyphFactoryList::reset()			//Destroys all in this except default-factories
{							//REMARK: this method is implemented based on the
    int i;
    for(i=0;i<ngf.Count();i++) delete ngf[i];
    for(i=0;i<egf.Count();i++) delete egf[i];
    ngf.Flush(); egf.Flush();
    init();                                             //init() again since we've just destroyed the default-factories
}

void IVGlyphFactoryList::save(FILE* fp,std::set<STRING>* tcl_procs)  
{                                                       //Save all glyph-factories to Tcl-file. Gather the Tcl-procs
    int i;                                              //these depend on and return them to caller, if arg given.

    fprintf(fp,"# Glyph factories --------------------------\n\n");
    
    for(i=0;i<ngf.Count();i++)                          //Ask node-glyphs to save their settings
    {
       if (tcl_procs) ngf[i]->getTclDeps(*tcl_procs);   //Ask node-glyphs which Tcl-procs they use
       if (i>=num_def_ngfs) ngf[i]->save(fp);           //Ask them to save themselves, if not the default (hardcoded) ones
    }
       
    for(i=0;i<egf.Count();i++)                          //Ask edge-glyphs to save their settings
    {
       if (tcl_procs) egf[i]->getTclDeps(*tcl_procs);   //Ask edge-glyphs which Tcl-procs they use
       if (i>=num_def_egfs) egf[i]->save(fp);           //Ask them to save themselves, if not the default (hardcoded) ones
    }

    fprintf(fp,"\n\n");
}

IVNodeGlyphFactory* IVGlyphFactoryList::getNGFactory(const char* n)
{
    for(int i=0;i<ngf.Count();i++) if ((*ngf[i])==n) return ngf[i];
    return def_ngf;
}



IVEdgeGlyphFactory* IVGlyphFactoryList::getEGFactory(const char* n)
{
    for(int i=0;i<egf.Count();i++) if ((*egf[i])==n) return egf[i];
    return def_egf;
}



void IVGlyphFactoryList::addFactory(IVNodeGlyphFactory* f)
{   
    const char* name = f->getName();
    if (*def_ngf==name)					//not allowed to add factory with reserved name
    { delete f; return; }				//since caller expects we manage this, delete it.
    for(int i=0;i<ngf.Count();i++)			//search for f with same name as 's'
      if ((*ngf[i])==name)				//if found, delete it (we own it), and put 's' in its place
      {
	 if (ngf[i]==f) return;				//if, by any chance, we try to add the same factory-object,
	 delete ngf[i];					//cancel the operation, the object is already physically in the list.
	 ngf[i] = f;
	 return;
      }
      
    ngf.Add(f);						//if not found, simply add 'f' to list
}


void IVGlyphFactoryList::addFactory(IVEdgeGlyphFactory* f)
{   
    const char* name = f->getName();
    if (*def_egf==name)					//not allowed to add factory with reserved name
    { delete f; return; }				//since caller expects we manage this, delete it.
    for(int i=0;i<egf.Count();i++)			//search for f with same name as 's'
      if ((*egf[i])==name)				//if found, delete it (we own it), and put 's' in its place
      {
	 if (egf[i]==f) return;				//if, by any chance, we try to add the same factory-object,
	 delete egf[i];					//cancel the operation, the object is already physically in the list.
	 egf[i] = f;
	 return;
      }
      
    egf.Add(f);						//if not found, simply add 'f' to list
}


BOOL IVGlyphFactoryList::renameNGFactory(const char* f,const char* nf)
{   
    if (*def_ngf==f) return FALSE;			//not allowed to rename default factory

    IVNodeGlyphFactory*  fact = getNGFactory(f);	//find factory to be renamed 
    if (fact==def_ngf) return FALSE;			//factory doesn't exist
    if (*fact==nf) return TRUE;				//rename with same name, nothing to do

    ngf.Remove(fact);					//remove, but don't delete, factory
    if (*def_ngf==nf)					//if we try to rename factory to a reserved-name,
       delete fact;					//delete it since that name is already there
    else
    {
       fact->setName(nf);				//rename it, and then reinsert renamed factory
       addFactory(fact);				//This may delete an existing factory with same name
    }
    return TRUE;
}

BOOL IVGlyphFactoryList::renameEGFactory(const char* f,const char* nf)
{   
    if (*def_egf==f) return FALSE;			//not allowed to rename default factory
	
    IVEdgeGlyphFactory*  fact = getEGFactory(f);	//find factory to be renamed 
    if (fact==def_egf) return FALSE;			//factory doesn't exist
    if (*fact==nf) return TRUE;				//rename with same name, nothing to do

    egf.Remove(fact);					//remove, but don't delete, factory
    if (*def_egf==nf)					//if we try to rename factory to a reserved-name,
       delete fact;					//delete it since that name is already there
    else
    {
       fact->setName(nf);				//rename it, and then reinsert renamed factory
       addFactory(fact);				//This may delete an existing factory with same name
    }
    return TRUE;
}


BOOL IVGlyphFactoryList::delNGFactory(const char* n)	//destroys factory by name
{
   IVNodeGlyphFactory* f = getNGFactory(n);		//find factory with given name 
   if (f==def_ngf) return FALSE;			//not allowed to delete default factory
   const char* name = f->getName();
   ngf.Remove(f);					//remove factory from this's list
   delete f;						//destroy removed factory
   return TRUE;	
}				
   

BOOL IVGlyphFactoryList::delEGFactory(const char* n)	//destroys factory by name
{
   IVEdgeGlyphFactory* f = getEGFactory(n);		//find factory with given name 
   if (f==def_egf) return FALSE;			//not allowed to delete default factory
   const char* name = f->getName();
   egf.Remove(f);					//remove factory from this's list
   delete f;						//destroy removed factory
   return TRUE;	
}				
   