#include "GraphManager.h"
#include "SelList.h"
#include "ObsList.h"
#include "GenDomainModel.h"
#include "GlyphFactoryList.h"
#include "Registry.h"
#include "CompoLibrary.h"
#include "Ops.h"
#include "IVViewer.h"
#include "GUIManager.h"
#include "PathQualifier.h"
#include "utils.h"



GraphManager* GraphManager::instance = 0;


GraphManager::GraphManager():  
	    def_dmodel(new GenericDomainModel(""))	//make a default generic domain model
{
   instance  = this;
   dmodel    = 0;					//no custom DomainModel yet

   registry  = new Registry;
   sellist   = new SelectionList;
   glyphlist = new IVGlyphFactoryList;
   obslist   = new ObserverList(sellist);
   libs      = new ComponentLibrarySet;
}


GraphManager::~GraphManager()
{
  for(int i=0;i<obslist->numObservers();i++)
     obslist->getObserver(i)->setPersistent(FALSE);

  reset();					//Clear up all data created during this session

  delete libs;
  delete sellist;
  delete obslist; 
  delete glyphlist;
  delete dmodel;				//We assume we own this, in case someone set it.
  delete def_dmodel;				//We own this, we can delete it.
  delete registry;
}



void GraphManager::save(const char* f)          //Save entire system state as Tcl-file
{
   FILE* fp = fopen(f,"w");
   if (!fp) return;

   std::set<STRING> tcl_procs;                  //This gathers all Tcl-proc-names the subsystems depend on (call).
                                                //We need to save these explicitly, if they're not in compo-libs.
   
   STRING f2     = PathQualifier::removeExtension(f);  //Get in f2 the filename f without extension .tcl
   STRING f2_rel = PathQualifier::tail(f2);            //Get tail of f2, i.e. filename (without ext) out of 'f'

   fprintf(fp,"# System state saved by PR\n\n");

   //1. Save domain model. So far, only shallow-save implemented.
   fprintf(fp,"\n# Domain model -----------------------------\n\n");
   getDomainModel()->save(fp);

   //2. Deep-save graph data. Use absolute path since we write new file
   //REMARK: this should be optional.
   STRING graph_f = f2; graph_f += ".dat";
   SaveOp::writeOwnFormat(getSelList()->getFullSelection(),graph_f);
   
   //2.1. Save graph data. Use relative path to 'f', since written file in same dir.
   //REMARK: we might implement a system equivalent to DomainModel.
   //That is, when a graph-file is loaded, we should remember it so that a shallow-save
   //would then simply use that name... We assume of course that a shallow save didn't modify the graph data.
   graph_f = f2_rel; graph_f += ".dat";
   fprintf(fp,"\n# Graph data ---------------------------------\n\n");
   fprintf(fp,"pr_read_own_format %s\n",(const char*)addQuotes(graph_f));

  
   //3. Save selections - both shallow and deep. Both are needed since otherwise
   //it is impossible to restore the actual system state. 
   STRING sel_f = f2; sel_f += ".sel";          //Deep-save sel-file: use absolute name, needed since we write file
   getSelList()->saveDefinition(sel_f);
   sel_f = f2_rel; sel_f += ".sel";             //Shallow-save sel-file: use relative name, written file is in same dir
   fprintf(fp,"\n# Selections ---------------------------------\n\n");
   getSelList()->save(fp,sel_f);
   
   //4. Save viewers and their settings
   fprintf(fp,"\n# Viewers ---------------------------------\n\n");
   IVViewer::save(fp,&tcl_procs);
   
   //5. Save glyph-engines. Also gather Tcl-procs that glyph-engines use.
   fprintf(fp,"\n# Glyph engines ---------------------------------\n\n");
   getGlyphFactoryList()->save(fp,&tcl_procs);

   //6. Save compo-libs. So far, only shallow. Also remove from arg2 Tcl-procs registered as compos.
   fprintf(fp,"\n# Component libraries ---------------------------------\n\n");
   getCompoLibSet()->save(fp,&tcl_procs);

   //7. Separately save the Tcl-procs a) used by some subsystem and b) not registered as compos.
   //   They must be explicitly saved, since there's no other guarantee we have them back 
   //   in the system when the state is loaded.
   fprintf(fp,"\n# Local Tcl procedure definitions ----------------------------\n\n");
   for(std::set<STRING>::iterator it=tcl_procs.begin();it!=tcl_procs.end();it++)
      GUIManager::saveTclProc(fp,*it);
       
   //save registry !!! how exactly?
   //REMARK: this isn't really neede since we don't yet change registry during program execution
   
   //save other Tcl-based settings
   //!!! which ? are there any left we don't save yet?

   fclose(fp);
}





void GraphManager::reset()			//clean up all user data from the system
{
  PathQualifier::reset();                       //reset whatever the path-qualifier might have set

  libs->reset();                                //unload all loaded compo-libs

  sellist->reset();				//delete all selections we created
						//(except the full and empty selections)

  glyphlist->reset();				//delete all glyph-engines

  LoadOp::reset();                              //reset various global data-loading-op flags

  setDomainModel(0);				//delete the current custom domain-model, if any

  obslist->delAllObservers();			//destroy all non-persistent observers
}



void GraphManager::cleanup()			//Called (automatically) when the system is idle. This operation
{						//forces the update of any Observers that, for whatever reasons, have
   obslist->doUpdates();			//to update - mostly because what they observe, e.g. Selections, has changed.
}



void GraphManager::setDomainModel(DomainModel* d)
{  						//deletes current custom DomainModel
   delete dmodel;				//(we own it in any case, if there's one set)
   dmodel = d;  				//make the new custom DomainModel be ours
}



