#include "IVMapper.h"
#include "IVMapperImp.h"
#include "GraphManager.h"
#include "SelList.h"
#include "Selection.h"
#include "Node.h"
#include "Edge.h"
#include "Geometry.h"
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/SoPath.h>




IVMapperBase::IVMapperBase(): pos("pos"),pos_x("pos_x"),pos_y("pos_y"),pos_z("pos_z"),
			      sz_x("pos_dx"),sz_y("pos_dy"),sz_z("pos_dz"),
			      is_modified(TRUE),	//Forces a first full update on this
			      scale(0.5), 		//By default, node-sizes are 0.5 * nodes' dim-attrs
			      move_style(LOW_RES),	//By default, most interactive move style is preferred
			      highlight_r(1),		//The default highlight-color is red
			      highlight_g(0),
			      highlight_b(0),
			      edge_res(DEFAULT_RES),	//By default, use whatever edge-rep the edge comes with
			      edge_smoothing_subdiv(3),	//Choose some reasonable values for the edge smoothing params
			      edge_smoothing_tightness(0.33f)  
{   }

void IVMapperBase::setPosAttr(const char* pos_)
{
   pos   = pos_;
   pos_x = Geometry::getPositionX(pos);			//remember the name of the position-attrs
   pos_y = Geometry::getPositionY(pos);
   pos_z = Geometry::getPositionZ(pos);
   
   sz_x  = pos; sz_x  += "_dx";
   sz_y  = pos; sz_y  += "_dy";
   sz_z  = pos; sz_z  += "_dz";
}

IVMapperBase::~IVMapperBase()
{   }

Node* IVMapperBase::pickNode(SoPickedPoint*,float,float,float)
{  return 0;  }

Edge* IVMapperBase::pickEdge(SoPickedPoint*,float,float,float)
{  return 0;  }

IVMapperBase::SELECT_TYPE IVMapperBase::highlight(Node*,SELECT_TYPE)
{  return DESELECT;  }

IVMapperBase::SELECT_TYPE IVMapperBase::highlight(Edge*,SELECT_TYPE)
{  return DESELECT;  }

void IVMapperBase::clearHighlight()
{  }

void IVMapperBase::setMoveStyle(MOVE_STYLE ms)
{  move_style = ms;  }

void IVMapperBase::setNodeScale(float s)
{  scale = s;  }

void IVMapperBase::moveNode(Node*,float,float)
{   }

void IVMapperBase::moveHighlight(float,float)
{   }

void IVMapperBase::setEdgeResolution(EDGE_RESOLUTION er)
{  edge_res = er;  }

void IVMapperBase::setEdgeSmoothingParams(int k,float t)
{  edge_smoothing_subdiv = k; edge_smoothing_tightness = t; }

BOOL IVMapperBase::isModified() const
{   return is_modified;  }
      

void IVMapperBase::setHighlightColor(float r, float g, float b)
{
   highlight_r = r; highlight_g = g; highlight_b = b;
}

//-------------  IVMapper  ----------------------------------


IVMapper::IVMapper():node_mapper(0),edge_mapper(0)		//Ctor
{  
   base_root   = new SoSeparator; base_root->ref();		//base_root: root for ALL IV stuff in this
   root        = new SoSeparator; base_root->addChild(root);	//root: root for ALL Node/Edge stuff in this
   def_node_mapper = new IVFactoryNodeMapper(this);		//Make some default node/edge mappers
   def_edge_mapper = new IVBasicEdgeMapper(this);
   def_node_mapper->ref(); def_edge_mapper->ref();		//Also own these default mappers
   setNodeMapper(def_node_mapper);				//Make the def-mappers the current mappers
   setEdgeMapper(def_edge_mapper);				
   root->addChild(edge_mapper->getSceneGraph());
   root->addChild(node_mapper->getSceneGraph());
}

IVMapper::~IVMapper()						//Dtor: release the complete IV scenegraph
{   
    base_root->unref();  
    def_node_mapper->unref();					//Release our ownership on the default node and
    def_edge_mapper->unref();					//edge mappers	
    node_mapper->unref();					//Release our ownership of the current node and
    edge_mapper->unref();					//edge mappers, too
    for(int i=0;i<aux_node_mappers.Count();i++)                 //Release our ownership of the aux node mappers too
       aux_node_mappers[i]->unref(); 
}

void IVMapper::setNodeMapper(IVNodeMapper* m)
{
   if (node_mapper) node_mapper->unref();			//first release the current node mapper if any
   node_mapper = (m)? m : def_node_mapper;			//set the current mapper
   node_mapper->ref();						//own the current mapper
   node_mapper->setMainMapper(this);				//inform mapper which IVMapper it uses
}

void IVMapper::addAuxNodeMapper(IVNodeMapper* m, int i)         //add an aux-node-mapper at pos 'i' in the aux-mapper-list
{
   if (m==node_mapper) return;                                  //don't allow adding an aux-mapper twice in any way
   if (aux_node_mappers.Contains(m)!=-1) return;                //(including adding it as aux when it is primary mapper)
   if (i!=-1) aux_node_mappers.Insert(i,m);
   else       aux_node_mappers.Add(m);
   m->ref();                                                    //own this aux mapper
   m->setMainMapper(this);                                      //inform mapper which IVMapper it uses
}

void IVMapper::delAuxNodeMapper(IVNodeMapper* m)
{
   if (aux_node_mappers.Remove(m)==-1) return;                  //is 'm' indeed in this?
   m->unref();                                                  //relinquish ownership of this on m
}


void IVMapper::setEdgeMapper(IVEdgeMapper* m)
{
   if (edge_mapper) edge_mapper->unref();			//first release the current edge mapper if any
   edge_mapper = (m)? m : def_edge_mapper;			//set the current mapper
   edge_mapper->ref();						//own the current mapper
   edge_mapper->setMainMapper(this);				//inform mapper which IVMapper it uses
}

void IVMapper::clearHighlight()					//convenience method to unhighlight all that 
{								//is still highlighted in this. 									
   node_mapper->clearHighlight();				//Simply ask the node and edge mappers to do their
   hl_nodes.clear();						//own specific highlight-clearing
   edge_mapper->clearHighlight();
   hl_edges.clear();

   for(int i=0;i<aux_node_mappers.Count();i++)                  //Ask all aux mappers to unhighlight whatever they're highlighting
      aux_node_mappers[i]->clearHighlight();
}

IVMapperBase::SELECT_TYPE IVMapper::highlight(Node* nd, SELECT_TYPE sel_type)
{
   std::map<Node*,int>::iterator it = nodes.find(nd);		//see if we hold 'nd' as a displayed node
   if (it==nodes.end()) return DEFAULT;
   int idx = (*it).second;
   BOOL hl = (hl_nodes.find(nd)!=hl_nodes.end());

   if (sel_type==DEFAULT) return (hl)? SELECT : DESELECT;	//we're only asked our state, so return it
   
   if (!hl && (sel_type==SELECT || sel_type==TOGGLE)) 		//modify the IV selection group to reflect 
   {								//the new selection state
      std::map<Node*,int>::value_type v(nd,idx);
      hl_nodes.insert(v);
      node_mapper->highlight(idx,TRUE);	
      for(int i=0;i<aux_node_mappers.Count();i++)
         aux_node_mappers[i]->highlight(idx,TRUE);
   }
   else if (hl && (sel_type==DESELECT || sel_type==TOGGLE))
   {
      node_mapper->highlight(idx,FALSE); 
      for(int i=0;i<aux_node_mappers.Count();i++)
         aux_node_mappers[i]->highlight(idx,FALSE);
      hl_nodes.erase(nd);
   }

   return (hl_nodes.find(nd)!=hl_nodes.end())? SELECT : DESELECT;
}

IVMapperBase::SELECT_TYPE IVMapper::highlight(Edge* ed, SELECT_TYPE sel_type)
{
   std::map<Edge*,int>::iterator it = edges.find(ed);		//see if we hold 'ed' as a displayed edge
   if (it==edges.end()) return DEFAULT;
   int idx = (*it).second;
   BOOL hl = (hl_edges.find(ed)!=hl_edges.end());
 
   if (sel_type==DEFAULT) return (hl)? SELECT : DESELECT;
   
   if (!hl && (sel_type==SELECT || sel_type==TOGGLE)) 
   {
      std::map<Edge*,int>::value_type v(ed,idx);
      hl_edges.insert(v);
      edge_mapper->highlight(idx,TRUE);
   }
   else if (hl && (sel_type==DESELECT || sel_type==TOGGLE))
   {
      edge_mapper->highlight(idx,FALSE);
      hl_edges.erase(ed);
   }

   return (hl_edges.find(ed)!=hl_edges.end())? SELECT : DESELECT;
}

void IVMapper::moveNode(Node* n,float dx,float dy)		//A node-position has been moved. This must update, i.e. regenerate
{								//all its visual elements that depend on that position. Precisely,
								//we move the Node-glyph to the new pos-attrs of its Node and rebuild all
								//glyphs for all Edges that the Node has and are in this mapper's selection.
   std::map<Node*,int>::iterator it = nodes.find(n);
   if (it==nodes.end()) return;
   int idx = (*it).second;

   node_mapper->moveNode(n,idx,dx,dy);				//ask node and edge mappers to do
   edge_mapper->moveNode(n,idx,dx,dy);				//the moving of this node...

   for(int i=0;i<aux_node_mappers.Count();i++)                  //do the same for the aux node mappers        
      aux_node_mappers[i]->moveNode(n,idx,dx,dy);
}

void IVMapper::moveHighlight(float dx,float dy)			//Moves the whole highlight selection 
{								//with increments (dx,dy)
								//
   Selection* hl = GraphManager::getInstance()->getSelList()->getSelection(highlight_sel);   	

   node_mapper->moveHighlight(dx,dy);				//Ask node mapper to move highlight
   edge_mapper->moveHighlight(dx,dy);				//Ask edge mapper to move highlight

   for(int i=0;i<aux_node_mappers.Count();i++)                  //do the same for the aux node mappers        
      aux_node_mappers[i]->moveHighlight(dx,dy);
}

void IVMapper::setNodeScale(float s)				//scales all node glyphs
{
   IVMapperBase::setNodeScale(s);				//do inherited job
   node_mapper->setNodeScale(s);				//ask node mapper to move the node...
   for(int i=0;i<aux_node_mappers.Count();i++)                  //do the same for the aux node mappers        
      aux_node_mappers[i]->setNodeScale(s);
}

void IVMapper::setEdgeResolution(EDGE_RESOLUTION er)		//sets hint on how to draw edges
{
   IVMapperBase::setEdgeResolution(er);				//do inherited job
   Selection* sel = GraphManager::getInstance()->getSelList()->getSelection(input_sel);   
   edge_mapper->map(sel);					//REMARK: ask edge_mapper to change edge drawing means remapping...
}


SoSeparator* IVMapper::map(const char* in_sel, const char* hl_sel, const char* pos_name)
{
   input_sel     = in_sel;			//remember which input-selection and highlight-selection we're mapping	
   highlight_sel = hl_sel;
   Node* n; Edge* e; int i;

   if (pos_name && pos_name[0]) 
      setPosAttr(pos_name);			//if caller specifies a position-attr name to use, use it

   Selection* sel = GraphManager::getInstance()->getSelList()->getSelection(input_sel);   
   Selection* hl  = GraphManager::getInstance()->getSelList()->getSelection(highlight_sel);

 
   if (isModified() || sel->isModified())	//1. Update this's part depending on a modified 'sel'
   { 
     root->removeAllChildren();			//Clean up all the constructed stuff in this
   
     nodes.clear(); edges.clear();
     hl_nodes.clear(); hl_edges.clear();
     int idx=0;					//Initialize the 'nodes' map
     for(Selection::NodeIterator* it = sel->iterNodes();(n=it->next());idx++)			
     {					
       std::map<Node*,int>::value_type v(n,idx);
       nodes.insert(v);						
     }
     node_mapper->map(sel);			//Create the actual node-representations
     for(i=0;i<aux_node_mappers.Count();i++)    //in both the primary and aux mappers
        aux_node_mappers[i]->map(sel);
     
     idx=0;					//Initialize the 'edges' map
     for(Selection::EdgeIterator* j = sel->iterEdges();(e=j->next());idx++)
     {
         std::map<Edge*,int>::value_type v(e,idx);
         edges.insert(v);
     }
     edge_mapper->map(sel);			 //Create the actual edge-representations
    
     root->addChild(edge_mapper->getSceneGraph());//Add the new representation back to this
     root->addChild(node_mapper->getSceneGraph());//Add the new representation back to this
     for(i=0;i<aux_node_mappers.Count();i++)      //Add the new aux reps back to this
        root->addChild(aux_node_mappers[i]->getSceneGraph());
   }
   
   if (isModified() || sel->isModified() || hl->isModified())	
   {						//2. Update this's part depending on a modified highlight-sel
   						//(this case is implied by a modified sel)
      clearHighlight();				//First clear all highlights
      Selection::NodeIterator* i=hl->iterNodes();//Highlight what is to be highlighted
      for(;(n=i->next());) highlight(n,SELECT);  //from what we constructed
      Selection::EdgeIterator* j=hl->iterEdges();
      for(;(e=j->next());) highlight(e,SELECT); 
   }

   is_modified = FALSE;				//ok, this is now updated

   return base_root;				//return what we've constructed
}

void IVMapper::delNode(Node* n)			//Removes glyph for given Node.
{						//Searches nodes[] to see if we have 'n'. If n in nodes[],
						//this gives us the position where its glyph is as child of root.
						//So then we simply delete that glyph and also remove 'n' from nodes[]
  std::map<Node*,int>::iterator it = nodes.find(n);
  if (it==nodes.end()) return;
  node_mapper->delNode(n,(*it).second);         //Ask primary and aux mappers to release node
  for(int i=0;i<aux_node_mappers.Count();i++)
     aux_node_mappers[i]->delNode(n,(*it).second);
  nodes.erase(it);
  hl_nodes.erase(hl_nodes.find(n));
}

void IVMapper::delEdge(Edge* e)			//Removes glyph for given Edge.
{						//Searches edges for 'e'. If e in edges, this gives us the position
			 			//where its glyph is as child of root, AFTER the node-glyphs.
						//So then we simply delete that glyph and also remove 'n' from edges
  std::map<Edge*,int>::iterator it = edges.find(e);
  if (it==edges.end()) return;
  edge_mapper->delEdge(e,(*it).second);
  edges.erase(it);
  hl_edges.erase(hl_edges.find(e));
}

Node* IVMapper::pickNode(SoPickedPoint* p,float x,float y,float z)
{  return node_mapper->pickNode(p,x,y,z);  }

Edge* IVMapper::pickEdge(SoPickedPoint* p,float x,float y,float z)
{  return edge_mapper->pickEdge(p,x,y,z);  }

