#include "IVViewer.h"
#include "IVMapper.h"
#include "IVMapperImp.h"
#include "Selection.h"
#include "SelList.h"
#include "ObsList.h"
#include "GUIManager.h"
#include "EventHandler.h"
#include "GraphManager.h"
#include "GUIManager.h"
#include "Registry.h"
#include "Node.h"
#include "Edge.h"
#include "utils.h"
#include "Genrl/minmax.h"
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/details/SoLineDetail.h>
#include <Inventor/events/SoEvents.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include "UI_imp.h"
#include "IVUtils.h"
#include <GL/gl.h>					//include after UI_imp.h, as the latter includes platform stuff on Win32


DARRAY<IVViewer*> IVViewer::viewers;			//class static member decl



void IVViewer::render_cb(void* p,SoAction* a)		//IVViewer render-cb. This enables/disables Z buffering before doing any
{							//rendering as a function on the viewer's get2DRendering().
  if (a->isOfType(SoGLRenderAction::getClassTypeId()))	//We might have done this in the set2DRendering() routine but there we have no
  {							//GL-context for this viewer...		
    IVViewer* this_ = (IVViewer*)p;
    if (this_->get2DRendering()) glDisable(GL_DEPTH_TEST); 
    else glEnable(GL_DEPTH_TEST);
    this_->rendered = TRUE;                             //Mark this as 'rendered' - useful to signal various code that draws directly via OpenGL 
                                                        //into this that this has already re-rendered itself. See for ex. EventHandler.
  }
  
}


const char* IVViewer::make_name(const char* n)          //Make unique viewer-name out of given-name 'n'
{
   static char buf[100];
   int i; STRING nm = n; int cnt = viewers.Count();
   for(;;)
   {
     for(i=0;i<viewers.Count();i++)                     //check if 'nm' is used as viewer name or not
       if (nm==viewers[i]->getName()) break;   
     if (i<viewers.Count())                             //'nm' is used as name. Make a new name and increase cnt
     { nm = "viewer_"; nm += cnt++; }                    // and try again if this is a unique name...
     else break;                                        //'nm' is not used as name, we've found our unique name        
   }
   if (cnt>viewers.Count()) strcpy(buf,nm);             //if we had to change the name (not unique), set new one
   else strcpy(buf,n);                                  //else simply return the original one
   
   return buf;
}


IVViewer::IVViewer(const char* n,const char* sn, const char* pn, const char* hln)
	 : Observer(make_name(n),sn),                   //make Observer-parent with unique name
           pos_name(pn),
	   highlight_name(hln),
	   root(0),
	   output_only(TRUE),				//viewer doesn't allow editing contents (safest default)
	   mapper(0),					//no custom mapper yet				
	   render_2d(FALSE),    			//use 3D render/shading look
           rendered(FALSE),
           store(0),                                    //initialize store_xxx members to nil
           store_sz(0),
           store_x(-1),store_y(-1),store_dx(0),store_dy(0)
{
   

   def_mapper  = new IVMapper; def_mapper->ref();	//make a default mapper and own it
   setMapper(def_mapper);				//also set this to be the current mapper
   fact_node_mapper = new IVFactoryNodeMapper;		//make and own various other specialized
   fact_node_mapper->ref();				//mapper subcomponents. When new mapper classes are written, one 
   basic_node_mapper = new IVBasicNodeMapper;   	//should add code to create & attach them to the viewer, here.	
   basic_node_mapper->ref();
   splat_node_mapper = new IVSplatNodeMapper;
   splat_node_mapper->ref();
   fact_edge_mapper = new IVFactoryEdgeMapper;
   fact_edge_mapper->ref();
   basic_edge_mapper = new IVBasicEdgeMapper;
   basic_edge_mapper->ref();
   def_mapper->setNodeMapper(fact_node_mapper);		//make the factory node/edge mappers the 
   def_mapper->setEdgeMapper(fact_edge_mapper);		//current ones for the IVMapper. This is actually 
							//needed for the getNode/EdgeMapper() routines to work ok.


   evt_handler = new DefaultEventHandler;		//event-handler is the default one so far
   base_root = new SoSeparator; base_root->ref();	//base_root holds EVERYTHING we put in this viewer

   SoCallback* cb = new SoCallback;			//add render-callback to enable/disable Z-buffering before doing
   cb->setCallback(render_cb,this); base_root->addChild(cb);  //any rendering in this
   base_root->addChild(light_model = new SoLightModel);	//lighting model for this viewer

   SoEventCallback* evt_cb = new SoEventCallback; base_root->addChild(evt_cb);
   evt_cb->addEventCallback(SoEvent::getClassTypeId(),event_cb,this);

   base_root->addChild(new SoOrthographicCamera);	//no perspective
   base_root->addChild(root = new SoSeparator);		

   STRING win_name = "Viewer: "; win_name+=getName();	//make name for viewer-window
   SoUIExaminerViewer* vw = new SoUIExaminerViewer;	//make new viewer
   viewer    = vw;		 			//store the new viewer both as a SoUIViewer
   ui_viewer = vw;					//and as a UIViewerBase (needed due to the multiple inheritance)
   viewer->setSize(SbVec2s(450,450));			//needed since in the Coin3D implementation of
							//Inventor, on some Windows platforms, the window
							//buttons look ok ONLY after the 1st resize...
   viewer->setSceneGraph(base_root);			
   viewer->setTitle(win_name);

   Registry::Iterator it (*GraphManager::getInstance()->getRegistry(),"viewer_sweep_cb");
   if (it.num())					//Set, as sweep-cb, the Tcl-proc under the "viewer_sweep_cb" key
   {							
      const DARRAY<STRING>&  e = it.get(0);
      if (e.Count()==1 && e[0]!="")			//Discard entries with empty callback			
         setSweepCallback(e[0]);			//1st item: Tcl sweep callback
   }

   viewer->show();
   viewers.Add(this);				        //add this to the collection of viewers	
}


IVViewer::~IVViewer()
{
   base_root->unref();					//relinquish our use of base_root
   mapper->unref();					//release current mapper
   def_mapper->unref();					//release also our own def_mapper
   fact_node_mapper->unref();
   basic_node_mapper->unref();
   fact_edge_mapper->unref();
   basic_edge_mapper->unref();
   delete (SoUIComponent*)viewer;			//REMARK: upcast is safe, needed since ~SoXtViewer is, oddly, private 
   viewers.Remove(this);				//delete this from the collection of viewers
   delete evt_handler;					//we own def_handler
   delete[] store;                                      //Release 'store' image-backup 
}

void IVViewer::set2DRendering(BOOL b)			//set 2D rendering style
{  
   if (render_2d==b) return;
   render_2d = b;
   light_model->model = (b)? SoLightModel::BASE_COLOR : SoLightModel::PHONG;
}



void IVViewer::setNodeMapper(MAPPER_TYPE nm)
{
   switch(nm)
   {
      case FACTORY:  def_mapper->setNodeMapper(fact_node_mapper);  break;
      case BASIC:    def_mapper->setNodeMapper(basic_node_mapper); break;
      case SPLAT:    def_mapper->setNodeMapper(splat_node_mapper); break;
   }
   touch();                                                                     //this'll have to update
}


void IVViewer::setAuxNodeMapper(MAPPER_TYPE nm,BOOL state)                      //switch on/off the given mapper-type as
{                                                                               //aux-node-mapper for this
   IVNodeMapper* map = (nm==SPLAT)? (IVNodeMapper*)splat_node_mapper :
                       (nm==BASIC)? (IVNodeMapper*)basic_node_mapper :
                       /*FACTORY*/  (IVNodeMapper*)fact_node_mapper;
   if (state)
      def_mapper->addAuxNodeMapper(map);
   else
      def_mapper->delAuxNodeMapper(map);
   touch();                                                                     //this'll have to update
}
  
BOOL IVViewer::getAuxNodeMapper(MAPPER_TYPE nm)                                 //get the on/off state of the aux-mapper indicated
{                                                                               //by 'nm'-type. Since we know we add the aux-mappers
   IVNodeMapper* map = (nm==SPLAT)? (IVNodeMapper*)splat_node_mapper :          //we created ourselves to def_mapper, simply chewck thus        
                       (nm==BASIC)? (IVNodeMapper*)basic_node_mapper :          //if the aux-mapper corresponding to 'nm'-type is added.
                       /*FACTORY*/  (IVNodeMapper*)fact_node_mapper;
   return def_mapper->hasAuxNodeMapper(map);
}

void IVViewer::setEdgeMapper(MAPPER_TYPE em)
{
   switch(em)
   {
      case FACTORY:     def_mapper->setEdgeMapper(fact_edge_mapper); break;
      case BASIC:	def_mapper->setEdgeMapper(basic_edge_mapper); break;
   }
   touch();                                                                     //this'll have to update
}  

IVViewer::MAPPER_TYPE IVViewer::getNodeMapper()
{
   IVNodeMapper* nm = def_mapper->getNodeMapper();
   return (nm==fact_node_mapper)?   FACTORY :
	  (nm==basic_node_mapper)?  BASIC :
                                    SPLAT;    
}

IVViewer::MAPPER_TYPE IVViewer::getEdgeMapper()
{
   IVEdgeMapper* em = def_mapper->getEdgeMapper();
   return (em==fact_edge_mapper)? FACTORY 
	                        : BASIC;
}
   

void IVViewer::setHighlightColor(float r,float g,float b)	//Delegate to mapper
{  getMapper()->setHighlightColor(r,g,b);  }

void IVViewer::getHighlightColor(float& r,float& g,float& b)	//Delegate to mapper
{  getMapper()->getHighlightColor(r,g,b);  }

void IVViewer::setBackgroundColor(float r,float g,float b)
{  viewer->setBackgroundColor(SbColor(r,g,b));  }

void IVViewer::getBackgroundColor(float& r,float& g,float& b)
{  
   const SbColor c = viewer->getBackgroundColor();  
   r=c[0];g=c[1];b=c[2];
}



BOOL  IVViewer::mustBeUpdated() const			//Overrides inherited to monitor changes in both
{							//the basic and highlight selection and also the 'global'
   if (getMapper()->isModified() || alwaysUpdate()) return TRUE;   //state of the mapper
   SelectionList* sl = GraphManager::getInstance()->getSelList();
   Selection*     s1 = sl->getSelection(getSelection());
   Selection*     s2 = sl->getSelection(highlight_name);
   return s1->isModified() || s2->isModified();
}

void IVViewer::setMapper(IVMapperBase* m)
{
   if (mapper) mapper->unref();				//first release the curent mapper if any
   mapper = (m)? m : def_mapper;			//set new mapper or default if nothing given
   mapper->ref();					//we own a ref to the new mapper from now on...
   if (root) reset();					//also, reset all that was in the previous mapper since we're looking
							//at a new one. We leave for the time being update() to be called separately.
}


void IVViewer::reset()					//clears this viewer
{							//(does not necessarily destroy the scene graph, the mapper
   root->removeAllChildren();				//might own it)
}

void IVViewer::update()					//update this upon some change
{  
   reset();						//remove all stuff from this
   root->addChild(getMapper()->map(getSelection(),highlight_name,pos_name));
							//ask this's mapper to rebuild itself completely
							//and then use what it delivers
}

void IVViewer::touch()					//this being modified means, so far,
{							//that this's mapper is modified
   getMapper()->touch();  				
   GraphManager::getInstance()->getObsList()->update();	//also touch the ObserverList since this is an Observer
}


Node* IVViewer::pickNode(float x,float y,IVZBufferHit* zb) //Pick node closest to norm-screen-coords (x,y). 
{                                                          //Wireframe objs are handled correctly.
   return (Node*)pickElement(x,y,zb,TRUE);                 //Ask mapper which Node has the picked IV-stuff
}


Edge* IVViewer::pickEdge(float x,float y,IVZBufferHit* zb) //Pick edge closest to norm-screen-coords (x,y). 
{                                                          //Wireframe objs are handled correctly.
   return (Edge*)pickElement(x,y,zb,FALSE);                //Ask mapper which Edge has the picked IV-stuff
}


Element* IVViewer::pickElement(float x,float y,IVZBufferHit* zb, BOOL node_edge)
{                                                       //Pick node/edge closest to norm-screen-coords (x,y). 
   SoRayPickAction a(viewer->getViewportRegion());  	//Wireframe objs are handled correctly
   a.setNormalizedPoint(SbVec2f(x,y));
   a.apply(base_root);                                  //Cast a ray to determine 1st object at (x,y)
   SoPickedPoint* pt = a.getPickedPoint();
   if (!pt) return 0;					//1. Nothing picked by ray, abort anyway

   SbVec3f pnt;                                         //Now we must check all objects along ray. Declare a new pick-action a2
   SoRayPickAction a2(viewer->getViewportRegion());     //HERE - the object must exist till the function finishes, since the
                                                        //pick-point we get from it is a shallow-pointer to internal data.

   if (zb && !zb->empty())				//2. If the zbuffer-based picking did report something:
   {                                                    //   Get all objects along the pick-ray and their z-intersecions with the
                                                        //   pick-ray. The real picked obj is the one whose z-intersection is the
      SbViewVolume vol =                                //   closest to the zbuffer-nbased z-pick value.
       	             viewer->getCamera()->getViewVolume(viewer->getViewportRegion().getViewportAspectRatio());
      a2.setNormalizedPoint(SbVec2f(x,y));
      a2.setPickAll(TRUE);
      a2.apply(base_root);
      float zbmin = 1000; 
      for(int i=0,N=a2.getPickedPointList().getLength();i<N;i++)
      {                                                 //For all objects picked along the ray:
         SoPickedPoint* pt_  = a2.getPickedPoint(i);
	 const SbVec3f& pnt_ = pt_->getPoint(); 
         SbVec3f ppnt; vol.projectToScreen(pnt_,ppnt);  //Determine homogeneous z-coord of pick-point
	 float z = (ppnt[2]+1)/2;                       //REMARK: this is unclear why, but that's how to get it in the same coord-sys
	 float zmin = 1000;                             //        as the zbuffer z-values, so we can compare them.
	 for(int j=0;j<zb->dimX();j++)                  //Find min-distance, on z, between z-buffer-points and 'z'
          for(int k=0;k<zb->dimY();k++)
          { float diff = fabs(z-(*zb)(j,k)); if (zmin>diff) zmin = diff; }
         if (zmin<zbmin) { zbmin = zmin; pnt = pnt_; pt = pt_; } //Remember in (pt,pnt) ray-object intersection closest to zbuffer
      }
   } 
   else pnt = pt->getPoint();                           //3. No zbuffer-based picking. Simply return the object-based picking.

   if (node_edge==TRUE)
      return getMapper()->pickNode(pt,pnt[0],pnt[1],pnt[2]);
   else 
      return getMapper()->pickEdge(pt,pnt[0],pnt[1],pnt[2]);      
}


void IVViewer::pickNodes(Selection* sel,float x1,float y1,float x2,float y2)
//Adds all nodes whose positions fall in the (x1,y1,x2,y2)-viewport, given
//in norm coords relative to the current view-setings, to selection 'sel'.
{
   //1. Get this's view frustum
   SbViewVolume vol = viewer->getCamera()->
		      getViewVolume(viewer->getViewportRegion().getViewportAspectRatio());

   //2. Restrain the frustum to specified viewport (x1,x2,y1,y2).
   vol = vol.narrow(min(x1,x2),min(y1,y2),max(x1,x2),max(y1,y2));

   //3. Add to 'sel' all nodes whose positions fall within restrained frustum.
   //   We don't really need to delegate this to the mapper, since all computations
   //   are based on node-position-vs-frustum intersections that we can do ourselves.

   Selection* s = GraphManager::getInstance()->getSelList()->getSelection(getSelection());
   Selection::NodeIterator* i = s->iterNodes();
   for(Node* n;(n=i->next());)
   {
      float x = n->getFloat(getMapper()->getPosAttrX()),
            y = n->getFloat(getMapper()->getPosAttrY()),
            z = n->getFloat(getMapper()->getPosAttrZ());
	  if (IVViewVolumeIntersect(vol,SbVec3f(x,y,z))) sel->addNode(n);
   }
}



void IVViewer::event_cb(void* cd,SoEventCallback* cb)	//Dispatch mouse events to this's event-handler, if any
{
  IVViewer*  this_ = (IVViewer*)cd;
  if (!this_->evt_handler) return;			//This viewer is not interested in events
  
  const SoEvent* e = cb->getEvent();			//Get the Inventor event
  const SbVec2f& p = e->getNormalizedPosition(this_->viewer->getViewportRegion());	

  EventHandler::EVENT_TYPE    evt;			//Translate the Inventor event to our event-type
  EventHandler::MODIFIER_TYPE mod = EventHandler::NONE;
  int key = -1;

  if (SoKeyboardEvent::isKeyReleaseEvent(e,SoKeyboardEvent::ANY)) 
  {
     evt = EventHandler::KEY;                           //If key-release event, get key-code and pass it
     key = (int)((SoKeyboardEvent*)e)->getKey();        //further as int...        
     
  }
  else                          
  {
     evt = (e->isOfType(SoLocation2Event::getClassTypeId())) 		         ? EventHandler::MOVE :
        (SoMouseButtonEvent::isButtonPressEvent(e,SoMouseButtonEvent::BUTTON1))  ? EventHandler::DOWN_1 :
        (SoMouseButtonEvent::isButtonReleaseEvent(e,SoMouseButtonEvent::BUTTON1))? EventHandler::UP_1  :
        (SoMouseButtonEvent::isButtonPressEvent(e,SoMouseButtonEvent::BUTTON2))  ? EventHandler::DOWN_2 :
        (SoMouseButtonEvent::isButtonReleaseEvent(e,SoMouseButtonEvent::BUTTON2))? EventHandler::UP_2 :
        EventHandler::OTHER;
  }

  if (e->wasShiftDown()) mod = (EventHandler::MODIFIER_TYPE)(mod | EventHandler::SHIFT);
  if (e->wasCtrlDown())  mod = (EventHandler::MODIFIER_TYPE)(mod | EventHandler::CTRL);
  if (e->wasAltDown())   mod = (EventHandler::MODIFIER_TYPE)(mod | EventHandler::ALT);

  if (evt!=EventHandler::OTHER)				//Call our event-handler !
     this_->evt_handler->event(this_,evt,mod,p[0],p[1],key);
}

void IVViewer::show()					//Show current viewer
{   viewer->show();   }

void IVViewer::pan(UIViewerBase::PAN_DIRECTION dir)	//Pan view with one step in given dir
{   ui_viewer->pan(dir);  }

void IVViewer::addButton(const char* name,const char* cb,const char* tt,BOOL checkbtn)
{   ui_viewer->addButton(name,cb,tt,checkbtn);  }


IVViewer* IVViewer::getViewer(const char* n)		//Get viewer by name
{
    STRING name = n;
    for(int i=0;i<numViewers();i++)
	if (name==getViewer(i)->getName()) return getViewer(i);
    return 0;
}

IVViewer* IVViewer::getViewer(const SoUIViewer* vw)	//Get viewer by implementation-object
{
    for(int i=0;i<numViewers();i++)
	if (vw==getViewer(i)->getViewer()) return getViewer(i);
    return 0;
}



//---------  Various utility functions  -------------------------------------------------------------------

void view_default_callback_delete(EventHandler*,void* cd)//Delete-callback for the viewer's event handler
{
   STRING* s = (STRING*)cd;				//Delete the STRING arg we created
   delete s;						//and added to the callback
}


STRING view_default_callback(EventHandler* hd,void* cd)	//Callback for the viewer's EventHandler
{
   STRING& tcl_code = *((STRING*)cd);			//Get the STRING arg with the Tcl user code
   const Node* nd = hd->getPickedNode();		//Get the node or edge picked at last event
   const Edge* ed = hd->getPickedEdge();
   const Selection* sel = hd->getPickedSelection();
   char buf[1000];

   STRING res;                                          //Return value
   if (nd)						//Have we picked a node?
   {
     sprintf(buf,"%s node %d",(const char*)tcl_code,nd->getId());
     GUIManager::exec(buf);				//Execute modified Tcl user code containing picked-ids
     res = GUIManager::getStringResult();              //Return whatever the Tcl evaluation amounted to
   }
   else if (ed)						//Have we picked an edge?
   {
     sprintf(buf,"%s edge %d",(const char*)tcl_code,ed->getId());
     GUIManager::exec(buf);	
     res = GUIManager::getStringResult();              //Return whatever the Tcl evaluation amounted to
   }
   else if (sel)					//Have we changed the highlight via rubberbanding?
   {
     sprintf(buf,"%s selection %s",(const char*)tcl_code,sel->getName());
     GUIManager::exec(buf);
   }
   return res;
}



void IVViewer::setUserCallback(const char* val)
{
   if (val[0])
   {
      getEventHandler()->setUserCallback(view_default_callback,new STRING(val));
      getEventHandler()->setDeleteCallback(view_default_callback_delete);
   } 
   else getEventHandler()->setUserCallback(0);
}

void IVViewer::setSweepCallback(const char* val)
{
   if (val[0])
   {
      getEventHandler()->setSweepCallback(view_default_callback,new STRING(val));
      getEventHandler()->setSweepDeleteCallback(view_default_callback_delete);
   } 
   else getEventHandler()->setSweepCallback(0);
}


STRING IVViewer::getSweepCallback()						//Get Tcl-code for sweep-callback (of "" if none)
{
   STRING res; 									//We know that this is a string, since we set it so
   EventHandler::Callback cb; void* cd;						//(see the set for "sweep_cb").
   getEventHandler()->getSweepCallback(cb,cd);					//WARNING: this must change when new EventHandlers
   if (cd) res = (*(STRING*)cd);						//with new user-data types will come into use.
   return res;
}


STRING IVViewer::getUserCallback()						//Get Tcl-code for user-callback (of "" if none)
{
   STRING res; 									//We know that this is a string, since we set it so
   EventHandler::Callback cb; void* cd;						//(see the set for "sweep_cb").
   getEventHandler()->getUserCallback(cb,cd);					//WARNING: this must change when new EventHandlers
   if (cd) res = (*(STRING*)cd);						//with new user-data types will come into use.
   return res;
}

void IVViewer::save(FILE* fp,std::set<STRING>* tcl_procs)       //Save settings of this viewer to Tcl-file                      
{
   STRING s;

   for(int i=0;i<numViewers();i++)                              //Save info for all existing viewers
   {
     IVViewer* vw = getViewer(i);                               
     STRING    nm = addQuotes(vw->getName());
     fprintf(fp,"pr_view_new %s %s %s %s\n",                    //Save Tcl-cmd to create viewer with its        
             (const char*)nm,                                   //name, selection, pos-attr, and hl-sel
             (const char*)addQuotes(vw->getSelection()),
             (const char*)addQuotes(vw->getPosAttr()),
             (const char*)addQuotes(vw->getHighlightSelection()));
     fprintf(fp,"pr_view %s set node_scale %8f\n",   (const char*)nm,vw->getMapper()->getNodeScale());
     fprintf(fp,"pr_view %s set edge_res %d\n",      (const char*)nm,(int)vw->getMapper()->getEdgeResolution());
     float r,g,b; vw->getHighlightColor(r,g,b);
     fprintf(fp,"pr_view %s set highlight_col %s\n", (const char*)nm,(const char*)rgb2tkcolor(r,g,b));
     vw->getBackgroundColor(r,g,b);
     fprintf(fp,"pr_view %s set background_col %s\n", (const char*)nm,(const char*)rgb2tkcolor(r,g,b));
     fprintf(fp,"pr_view %s set output_only %d\n",    (const char*)nm,(int)vw->getOutputOnly());
     fprintf(fp,"pr_view %s set render_2d %d\n",      (const char*)nm,(int)vw->get2DRendering());
     fprintf(fp,"pr_view %s set move_style %d\n",     (const char*)nm,(int)vw->getMapper()->getMoveStyle());
     fprintf(fp,"pr_view %s set node_mapper %d\n",    (const char*)nm,(int)vw->getNodeMapper());
     fprintf(fp,"pr_view %s set edge_mapper %d\n",    (const char*)nm,(int)vw->getEdgeMapper());
     fprintf(fp,"pr_view %s set enable_sweep_cb %d\n",(const char*)nm,(int)vw->getEventHandler()->sweepCallbackEnabled());
     
     if (vw->getSweepCallback()!="")
        fprintf(fp,"pr_view %s set sweep_cb %s\n",    (const char*)nm,(const char*)vw->getSweepCallback());
     if (vw->getUserCallback()!="")
     fprintf(fp,"pr_view %s set user_cb %s\n",        (const char*)nm,(const char*)vw->getUserCallback());
     s = vw->getIVFactoryNodeMapper()->getNodeGlyphFactory();
     if (s!="")
        fprintf(fp,"pr_view %s set node_glyph %s\n",     (const char*)nm,(const char*)addQuotes(s));
     s = vw->getIVFactoryEdgeMapper()->getEdgeGlyphFactory();
     if (s!="")
        fprintf(fp,"pr_view %s set edge_glyph %s\n",     (const char*)nm,(const char*)addQuotes(s));

     if (tcl_procs)                                   //Gather Tcl-procs this uses, if asked for  
     {
       if (vw->getUserCallback()!="")  tcl_procs->insert(vw->getUserCallback()); 
       if (vw->getSweepCallback()!="") tcl_procs->insert(vw->getSweepCallback());
     }

     //...other viewer-settings to be saved go here    
     fprintf(fp,"\n");
     
     if (i==numViewers()-1)                           //Save this setting only for last viewer, it's global anyway  
     {
        fprintf(fp,"pr_view %s set pan_step %6f\n",   (const char*)nm,UIViewerBase::getPanStep());
        fprintf(fp,"\n");
     }
   }

}
