#include "Ops.h"
#include "GraphManager.h"
#include "DomainModel.h"
#include "AttrSetType.h"
#include "Node.h"
#include "Edge.h"
#include "Selection.h"
#include "SelList.h"
#include "Geometry.h"
#include "GraphEd.h"
#include "Filter.h"
#include "Genrl/string.h"
#include "Genrl/types.h"
#include "Genrl/minmax.h"
#include "GUIManager.h"
#include "platform.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <map>



STRING GVLayoutOp::tmp_infile  = "./tmp_in";
STRING GVLayoutOp::tmp_outfile = "./tmp_out";
STRING GVLayoutOp::gv_path;
BOOL   LayoutOp::def_keep_dims = FALSE;
float  LayoutOp::def_node_size = 0.1f;
BOOL   LayoutOp::def_ext_lay_use_pos = FALSE;
BOOL   LayoutOp::def_ext_lay_use_dim = FALSE;
BOOL   LayoutOp::def_lay_edges = TRUE;


struct 	LayoutOp_ITEM 				//used locally by LayoutOp::layout()
	{ 
	  Node* node; float x,y,z; BOOL has_x,has_y,has_z; 
	};


LayoutOp::~LayoutOp()
{  }


BOOL LayoutOp::supportsPinning() 		//default implem: layout-class does not
{  return FALSE;  }				//support node-pinning


int LayoutOp::layout(const char* sel, const char* pos)
{
  Selection* s = GraphManager::getInstance()->getSelList()->getSelection(sel);
  return layout(s,pos);
}

int LayoutOp::layout(Selection* sel_, const char* pos_)
{
   sel     = sel_; pos = pos_;				//cache args for subclasses to read - i.e.
   pos_x   = Geometry::getPositionX(pos);		//from now on, the getPositionX() and similar methods 
   pos_y   = Geometry::getPositionY(pos);		//will work ok.
   pos_z   = Geometry::getPositionZ(pos);
   dim_x   = Geometry::getDimensionX(pos);
   dim_y   = Geometry::getDimensionY(pos);
   dim_z   = Geometry::getDimensionZ(pos);
   pos_pin = Geometry::getPinned(pos);

   DARRAY<LayoutOp_ITEM> nodes;
   if (!supportsPinning())				//1. If this layout doesn't have intrinsic pinning support,
   {							//   we must do it ourselves in the LayoutOp baseclass.
      Selection::NodeIterator* i; const Node* n;	//   Collect all nodes whose 'pos_pin' attribute is set
      for(i=sel->iterNodes();(n=i->next());)		//   and store their position-attrs in an array.
         if (n->getInt(getPinning())==1)		//   Here we MUST check for existence of pos-attrs, since
	 {						//   we are BEFORE the layout, so nodes may/may not have them.
	    LayoutOp_ITEM it; const Attr* a;
	    it.has_x = ((a=n->get(getPositionX()))!=0);
	    it.x     = (a)? a->getFloat() : 0;
	    it.has_y = ((a=n->get(getPositionY()))!=0);
	    it.y     = (a)? a->getFloat() : 0;
	    it.has_z = ((a=n->get(getPositionZ()))!=0);
	    it.z     = (a)? a->getFloat() : 0;
	    it.node  = (Node*)n;
	    nodes.Add(it);
	}
   }

   int ret = layout();					//2. Ask subclass to do the real layout

   float xm,ym,xM,yM;					//3. Compute actual bbox layed out. We'll need this for
   if (!getKeepDimensions() || use_bbox)		//   various tasks below.	
      Geometry::bbox2D(sel,pos,xm,ym,xM,yM);		
     
   if (!getKeepDimensions())				//4. If this doesn't take the initiative to set the node-dims,
   {							//   we revert here to the default, i.e. make all nodes of a fraction
     float sz = getDefaultNodeSize()*max(xM-xm,yM-ym);	//   getDefNodeSize() of the layout-size. This ensures that any layout does
     Selection::NodeIterator* i; Node* n;	  	//   set the dim-attrs, in some way.
     for(i=sel->iterNodes();(n=i->next());)	
     {			
         n->set(getDimensionX(),sz);
         n->set(getDimensionY(),sz);
         n->set(getDimensionZ(),sz);
     }
   } 
 
   if (use_bbox)					//5. If we're asked to use the bbox to scale layout within:
   {							//   Scale both layout and node-dims to nicely fit the given bbox,
							//   without disturbing layout's own aspect ratio.
     float k = min((xM-xm>1.0e-3)? getBBoxDx()/(xM-xm):INFINITY,
		   (yM-ym>1.0e-3)? getBBoxDy()/(yM-ym):INFINITY);
     TSLayoutOp ts; 
     ts.setTransl(getBBoxOx()+(getBBoxDx()-k*(xM+xm))/2,
		  getBBoxOy()+(getBBoxDy()-k*(yM+ym))/2);
     ts.setScale(k,k,k);
     ts.useBoundingBox(FALSE);
     ts.LayoutOp::layout(sel,pos);
   }

   for(int i=0;i<nodes.Count();i++)		//6. If this didn't support pinning, then we must restore all
   {						//   pinned nodes back to their initial position...
      Node* n = nodes[i].node;	
      if (nodes[i].has_x) n->set(getPositionX(),nodes[i].x);
      if (nodes[i].has_y) n->set(getPositionY(),nodes[i].y);
      if (nodes[i].has_z) n->set(getPositionZ(),nodes[i].z);
   }

   if (ret) sel_->touch();			//REMARK: Needed since setting the attrs of a selection modifies it
   return ret;
}


//------  GridLayoutOp class  ----------------------------------------------------


int GridLayoutOp::def_nelems = 8;				//Default value of nelems
int GridLayoutOp::def_axis   = 0;				//Default value of axis

GridLayoutOp::GridLayoutOp(): nelems(def_nelems),axis(def_axis)	//Def ctor
{  }

GridLayoutOp::GridLayoutOp(int ne,int ax): nelems(ne),axis(ax) 	//Ctor
{  }

void GridLayoutOp::setGrid(int num)				//Specify we want 'num' nodes along major-axis 'ax'
{  nelems = num; if (nelems<1) nelems=1; }			//Then the # rows/columns along the other axis is implied.

void GridLayoutOp::setAxis(int ax)
{  axis = ax; if (axis!=0 && axis!=1) axis=0; }			//Specify major axis (0==x,1==y)

void GridLayoutOp::setDefaultGrid(int num)			//Def-version of setGrid()
{  def_nelems = num; if (def_nelems<1) def_nelems=1; }			

void GridLayoutOp::setDefaultAxis(int ax)			//Def-version of setAxis()
{  def_axis = ax; if (def_axis!=0 && def_axis!=1) def_axis=0; }



int GridLayoutOp::layout()
{
    GraphManager* gm  = GraphManager::getInstance();
    Selection*    s   = getSelection();

    Node* n; 
    float px    = getBBoxOx(), py = getBBoxOy(); 		//start laying out from bbox-origin
    float Mincr = ((!axis)? getBBoxDx():getBBoxDy())/nelems;	//Mincr = coord-advance along major axis
    int   minl  = int(s->numNodes()/nelems)+1;			//minl  = # rows/columns in minor dimension
    float mincr = ((axis)? getBBoxDx():getBBoxDy())/minl;	//mincr = coord-advance along minor axis
    float xincr = (!axis)? Mincr : mincr;			//xincr,yincr: coord-increments along
    float yincr = (axis)?  Mincr : mincr;			// x,y axes.


    int j = 0;
    Selection::NodeIterator* i;
    for(i=s->iterNodes();(n=i->next());)			//Lay out nodes:
    {
       n->set(getPositionX(),px); 				//Place node at current (px,py,0)
       n->set(getPositionY(),py); 
       n->set(getPositionZ(),0);
       if (!axis) px += xincr; else py += yincr; j++;		//Update px,py using xincr,yincr
       if (j == nelems) 
       { 
          j=0; 
          if (!axis) { px=getBBoxOx(); py+=yincr; } 
                else { py=getBBoxOy(); px+=xincr; } 
       }
    }

    return 1;	       
}


float NeatoLayoutOp::def_weight  = 1.0f; 				//Initialization of default values of NEATO-params
float NeatoLayoutOp::def_epsilon = 0.1f;
float NeatoLayoutOp::def_edgelen = 1.0f;
BOOL  NeatoLayoutOp::def_random  = FALSE;
BOOL  NeatoLayoutOp::def_overlap = TRUE;
BOOL  NeatoLayoutOp::def_compact = FALSE;
int   NeatoLayoutOp::def_maxiter = 2000;


NeatoLayoutOp::NeatoLayoutOp(): weight(def_weight),epsilon(def_epsilon),random(def_random)
{ }									//Ctor

BOOL NeatoLayoutOp::supportsPinning()
{  return TRUE;  }

const char* NeatoLayoutOp::progName() const
{  
   static char s[250]; char* str = s; s[0]='\0';
   if (gv_path[0])  { strcpy(s,gv_path); strcat(s,"/"); }		//path specified, use it
   strcat(s,"neato");
   str = s+strlen(s);

   sprintf(str," -Gmaxiter=%d -Gepsilon=%4.2f -Nshape=box -Elen=%5f -Goverlap=%s %s",
	   def_maxiter,epsilon,def_edgelen,(def_overlap)?"true":"false",(random)? "-Gstart=rand":"");

   return s;
}
  


//--------  DotLayoutOp class  -----------------------------------------------------


float  DotLayoutOp::def_ranksep = 0.75;					//def-values of layout params
float  DotLayoutOp::def_nodesep = 0.25;
int    DotLayoutOp::def_rankdir = 0;					//0==TB,1==BT,2==LR,3==RL
BOOL   DotLayoutOp::def_flip_edges = TRUE;				//flip edges by default
STRING DotLayoutOp::def_level_attr;					//no default node-attr to use as level-value


DotLayoutOp::DotLayoutOp()
{  }

BOOL DotLayoutOp::flipEdges()						//we can customize edge flipping
{  return def_flip_edges;  }

const char* DotLayoutOp::levelAttr()					//we can customize node-levels via node-attrs
{  return def_level_attr;  }

const char* DotLayoutOp::progName() const
{  
   static char s[250]; char* str = s; s[0]='\0';
   if (gv_path[0]!='\0')  { strcpy(s,gv_path); strcat(s,"/"); }		//path specified, use it
   strcat(s,"dot");
   str = s+strlen(s);

   sprintf(str," -Gnodesep=%4.2f -Granksep=%4.2f -Grankdir=%s",
	   def_nodesep,def_ranksep,(def_rankdir<2)? "TB":"LR");

   return s;
}

BOOL DotLayoutOp::flipX()
{  return def_rankdir==3;  }

BOOL DotLayoutOp::flipY()
{  return def_rankdir==1;  }


//----  GVLayoutOp  ------------------------------------------------------

void GVLayoutOp::init(const char* gv_path_)
{  gv_path = gv_path_;  }


GVLayoutOp::GVLayoutOp()
{   }

BOOL GVLayoutOp::flipEdges()					//By default, we don-t want to flip edges 
{  return FALSE;  }

const char* GVLayoutOp::levelAttr()				//By default, we don't use node-levels in the layout
{  return 0;  }

	   


BOOL NeatoLayoutOp::testOverlap(Selection* sel,Node* nd,float cx,float cy)
//Tests overlapping of all nodes in sel except nd with nd shifted at cx,cy
//REMARK: This method is not used. Delete it?
{
   Selection::NodeIterator* i; Node* n;	 
   STRING pos_x = getPositionX(),  pos_y = getPositionY(),
          sz_x  = getDimensionX(), sz_y  = getDimensionY();
   float  bx    = 1.5*nd->getFloat(getDimensionX()),	//test overlap with a bit larger bbox than 'nd' itself,
	  by    = 1.5*nd->getFloat(getDimensionY());	//since we don't want an extremely tight packing
   float  ox    = cx-bx/2, oy = cy-by/2;
   for(i=sel->iterNodes();(n=i->next());)
   {
      if (n==nd) continue;	
      float cxi = n->getFloat(pos_x), cyi = n->getFloat(pos_y),
            bxi = n->getFloat(sz_x),  byi = n->getFloat(sz_y);
      if (Geometry::rectOverlap(cxi-bxi/2,cyi-byi/2,cxi+bxi/2,cyi+byi/2,ox,oy,ox+bx,oy+by))
	 return TRUE;
   }
   return FALSE;
}




void NeatoLayoutOp::compact(int min_sz)			//Compacts the layout by arranging the non-connected
{							//nodes near the connected ones. Improves the quality of a neato layout,
  Selection* all = getSelection();			//as neato produces too loose layouts when given many unconnected, 
  Selection* cs  = new Selection;			//differently sized nodes.
 
  std::multimap<float,Node*> nodes;			//'nodes' holds the unconnected nodes in increasing y-dim value

  Selection::NodeIterator* i; Node* n; float area = 0;
  for(i=all->iterNodes();(n=i->next());)		//gather all unconnected nodes to compact
  {							//also compute the area sum of these nodes
     const EdgeSet& es = n->getEdges(all);
     if (es.isEmpty()) 
     { 
	std::multimap<float,Node*>::value_type v(n->getFloat(getDimensionY()),n);
	nodes.insert(v);
	area += n->getFloat(getDimensionX())*n->getFloat(getDimensionY());
     }
     else cs->addNode(n);				//'cs' will hold the connected nodes
  }

  if (nodes.size()>min_sz)				//Heuristic: compact stuff only when we have a minimum #
  {							//of unconnected nodes.
    float xm,ym,xM,yM,orig_x;				//orig_x says where we start to arrange the 
    if (cs->numNodes())					//unconnected nodes from. This is from the
    {							//right of the connected nodes' bbox, if there
       Geometry::bbox2D(cs,getPosition(),xm,ym,xM,yM);	//are any, otherwise simply from the origin of 
       orig_x = xM + 0.1*(xM-xm);			//the global bbox of all nodes.
    }   
    else 
    {
       Geometry::bbox2D(all,getPosition(),xm,ym,xM,yM); 
       orig_x = xm;
    }
    float crt_x = orig_x, crt_y = yM;
    float max_x = crt_x + 2*sqrt(area);
    for(std::multimap<float,Node*>::iterator i=nodes.begin();i!=nodes.end();i++)
    {
        Node*    nd = (*i).second;
	float dim_x = nd->getFloat(getDimensionX());
	float dim_y = nd->getFloat(getDimensionY());
	nd->set(getPositionX(),crt_x+dim_x/2);
	nd->set(getPositionY(),crt_y-dim_y/2);
	crt_x += 2*dim_x;
	if (crt_x>=max_x)
	{ crt_x = orig_x; crt_y -= 2*dim_y; }
    }
  }
  delete cs;                                            //done with temporary Selection
}




int GVLayoutOp::layout()
{
    GraphManager* gm  = GraphManager::getInstance();

    BOOL use_pos = getDefaultExtLayoutUseCrtPos();		//0. See whether we want to invoke external layout	
    BOOL use_dim = getDefaultExtLayoutUseCrtDim();		//   on the current pos/dim attrs, or simply from scratch. 

    int res = SaveOp::writeGV(getSelection(),tmp_infile,getPosition(),use_pos,use_dim,flipEdges(),levelAttr());
    if (!res) return 0;						//1. Write temporary file with selected graph to lay out

    char prog[200];						//2. Prepare to exec the 'dot' program
    sprintf(prog,"%s -Tplain %s -Gsize=\"%4.2f,%4.2f\" -o %s",	//   Pass it desired bbox-size.
	    progName(),(const char*)tmp_infile,getBBoxDx(),getBBoxDy(),
	    (const char*)tmp_outfile);

    res = GUIManager::spawn(prog);				//3. Exec 'dot' or 'neato'
    if (res) return 0; 						//   if failed to execute, abort too

    if (!read_layout(getSelection(),tmp_outfile)) return 0;	//4. Read temporary file that dot/neato produced
    
    unlink(tmp_infile);						//5. Delete temporary files
    unlink(tmp_outfile);

    return 1;
}




int NeatoLayoutOp::layout()					//Specific layout() for Neato
{
    int ne = getSelection()->numEdges();			//Do the standard GraphViz layout ONLY if there are any edges
    int r = (ne || !getDefaultCompact())?			//or if compaction is off. Otherwise it's useless (and slow), since all
	     GVLayoutOp::layout() : 1;				//unconnected nodes will be repositioned by compaction	

    if (r && getDefaultCompact()) compact((ne)? 3 : 0);		//If we wish to compact the layout, do the compacting phase
								//If there are no edges, we must compact ALL nodes, since we didn't
								//lay them out. Else we compact only from 3 nodes on.

    return r;							//Return whatever the GraphViz layout produced
}


BOOL GVLayoutOp::flipX()
{  return FALSE;  }

BOOL GVLayoutOp::flipY()
{  return FALSE;  }

int GVLayoutOp::read_layout(Selection* all,const char* fname)	//Read 'plain' format produced by dot/neato
{								//We don't put this operation in the public scope of
    FILE* fp = fopen(fname,"r");				//the other graph-data-reading ops since we read data here
    if (!fp) return 0;						//in a special way, knowing what/how we've written...

    float nx,ny,bbox_x,bbox_y; int id,num; char buf[5000],name[50],name2[50];

    fscanf(fp,"%s",buf);					//1. Get 1st keyword - should be "graph" or "stop"
    if (!strcmp(buf,"stop")) return 1;				//   If "stop", we're done	

    if (!strcmp(buf,"graph"))					//2. If "graph", read effective scale & bbox
       fscanf(fp,"%*f %f %f",&bbox_x,&bbox_y);			//

    for(;;)							//3. Read all "node" lines
    {
	 fscanf(fp,"%s",buf);
	 if (strcmp(buf,"node")) break;	

	 fscanf(fp,"%s %f %f %*f %*f",name,&nx,&ny);     	//   Read one node line - we should expect integer node-ids
         fgets(buf,5000,fp);					//   or names starting with '_' since that's how we wrote the file.
	 if (name[0]=='_') continue;				//   All nodes named "_something" are internal nodes, they have to be discarded.
	 id = atoi(name);

	 Node* n = all->getNodeById(id);			//4. Retrieve node with that id in the graph
	 if (!n) return 0;					//   We should find it - if not, abort

	 n->set(getPositionX(),(flipX())? bbox_x-nx:nx); 	//5. Set node's position to whatever dot/neato delivered
	 n->set(getPositionY(),(flipY())? bbox_y-ny:ny); 	//   (in 2D)
	 n->set(getPositionZ(),0);
    }

    static float ex[100],ey[100];                               //ex[],ey[] hold the edge-coords we read from file
    int i;
       
    if (getDefaultLayEdges())					//6. Read all "edge" lines if we lay out edges explicitly
      for(;(!strcmp(buf,"edge"));fscanf(fp,"%s",buf))		//   (otherwise there's nothing to read for edges...)
      {
	fscanf(fp,"%s %s %d",name,name2,&num);
	if (name[0]=='_' || name2[0]=='_') continue;		//   Skip names starting with '_', these are special edges only for the layout
	Node* n1 = all->getNodeById(atoi(name));		//   Get nodes at end of edge
        if (!n1) break;	
	Node* n2 = all->getNodeById(atoi(name2));
	if (!n2) break;
        if (flipEdges())					//   Swap n1,n2 if edges were swapped
	{							//   Also read the coords in opposite order s.th. ex[0],ey[0] are 
	   Node* tmp = n1; n1 = n2; n2 = tmp;			//   ALWAYS the coords close to node n1
	   for(i=num-1;i>=0;i--)                                 
	      fscanf(fp,"%f %f",&ex[i],&ey[i]);
	}	
	else for(i=0;i<num;i++)                                 //   Read edge-coords in direct order s.th. ex[0],ey[0] is the coord
	   fscanf(fp,"%f %f",&ex[i],&ey[i]);			//   close to node n1

	EdgeSet& es2 = n1->getEdges(all,Node::OUT);		//   Get edge 'e' between n1 and n2. Since edges have no ids, we
	for(int j=0;j<es2.Count();j++)				//   simply search for an (unique) edge between n1,n2 in 'all'	
	   if (es2[j]->toNode()==n2) break;
	if (j==es2.Count()) break;				//   No edge found - should not happen
	Edge* e = es2[j];
	
        for(int p=0;p<num;p++)                                  //   Flip edge-coords if needed...
        { if (flipX()) ex[p] = bbox_x-ex[p]; if (flipY()) ey[p] = bbox_y-ey[p]; }
	e->set(getPositionX(),ex,num);                          //   Set edge-coords as vector-attrs
	e->set(getPositionY(),ey,num);
	
	fgets(buf,1000,fp);
      }

    fclose(fp);							//7. At this point, "stop" should have been read, so return
    return 1;
}




float  LevelLayoutOp::def_spacing = 1.0;
STRING LevelLayoutOp::def_level;
LevelLayoutOp::SPACING_METHOD 
       LevelLayoutOp::def_spacing_method = LevelLayoutOp::NORMALIZED;



struct  LevelLayoutOp_ITEM 						//Used locally by LevelLayoutOp::layout()
	{ 	
	  Selection* sel; int level; 
	  LevelLayoutOp_ITEM() {};
	  LevelLayoutOp_ITEM(Selection* s,int l):sel(s),level(l) {}
	};

int LevelLayoutOp::layout()						//Lays out a selection as a set of independently layed out 2D-levels.
{									//The levels are stacked upon each other in 3D
    GraphManager*  gm  = GraphManager::getInstance();
    Selection*     s   = getSelection();
    DARRAY<LevelLayoutOp_ITEM> sels;

    int k; Node* n; Selection::NodeIterator* i;
    for(i=s->iterNodes();(n=i->next());)				//1. for all nodes in sel:
    {
       int l = n->getInt(def_level), j;					//find ITEM for this node's level, if any
       for(j=0;j<sels.Count() && sels[j].level<l;j++);			//if ITEM exists, add node to its sel,
       if (j<sels.Count())						//else create a new ITEM (and selection) and 	
       {								//insert it in sels[] in level-order.
	  if (sels[j].level != l)
	     sels.Insert(j,LevelLayoutOp_ITEM(new Selection,l));			
       }
       else
          sels.Insert(j,LevelLayoutOp_ITEM(new Selection,l));
	
       sels[j].sel->addNode(n);
    }

    Edge* e; Selection::EdgeIterator* j;				//2. for all horiz edges in sel:
    for(j=s->iterEdges();(e=j->next());)				//add them to the sels created before for their nodes
    {									
       if (e->fromNode()->getInt(def_level)!=e->toNode()->getInt(def_level)) 
	  continue;
       int l = e->fromNode()->getInt(def_level), j;
       for(j=0;j<sels.Count() && sels[j].level<l;j++);			//if a horiz-edge found, then its nodes SHOULD already have 
       if (j==sels.Count() || sels[j].level!=l) continue;		//determined the creation of some selections in the previous step.
       sels[j].sel->addEdge(e);	
    }
    if (!sels.Count()) return 1; 					//Just save some trouble if we have nothing to lay out anyway

    float z  = 0;
    float dz = (def_spacing_method==ABSOLUTE)? def_spacing : def_spacing/sels.Count();
 
    NeatoLayoutOp l;							//3. Now we have all nodes/hedges in the input-sel sorted,
    									//   per level, and the levels sorte din increasing order
    for(k=0;k<sels.Count();k++,z+=dz)					//   Lay out every level using the NeatoLayoutOp. The inter-level
    {									//   distance is set according to 'spacing' and 'spacing_method'
       Selection* s = sels[k].sel;
       l.LayoutOp::layout(s,getPosition());   
       Node* n; Selection::NodeIterator* i;
       for(i=s->iterNodes();(n=i->next());)				//   set the z-pos of all nodes in the current selection
          n->set(getPositionZ(),z);					//   (this overwrites any z-pos possibly set by the layout itself)
       delete s;							//   delete the working-selections we created here.
    }

    return 1;	       
}


STRING GEMLayoutOp::tmp_infile  = "./tmp_in";
STRING GEMLayoutOp::tmp_outfile = "./tmp_out";
STRING GEMLayoutOp::gem_dir;

void GEMLayoutOp::init(const char* gem_dir_)
{  gem_dir = gem_dir_;  }

const char* GEMLayoutOp::progName() const				//build gem's cmdline which depends on path
{  
   static char s[250];
   if (gem_dir[0]!='\0')						//path specified, use it
      sprintf(s,"%s/gem -c%s/gem.cfg",(const char*)gem_dir,(const char*)gem_dir);
   else 								//no path specified, assume crt-dir
      sprintf(s,"gem");
   return s;
}

int GEMLayoutOp::layout()
{
    GraphManager* gm  = GraphManager::getInstance();

    BOOL use_pos = getDefaultExtLayoutUseCrtPos();

    int res = SaveOp::writeGraphEd(getSelection(),tmp_infile,getPosition(),use_pos,TRUE);
    if (!res) return 0;							//1. Write temporary file with selected graph to lay out
									//   (and do write node-ids in this file)
    char prog[200];							//2. Prepare to exec the 'gem' program
    sprintf(prog,"%s < %s > %s",progName(),(const char*)tmp_infile,(const char*)tmp_outfile);

    res = GUIManager::spawn(prog);					//3. Exec 'gem'
    if (res) return 0; 							//   if failed to execute, abort too

    if (!read_layout(tmp_outfile)) return 0;				//4. Read temporary file that 'gem' produced
    
    unlink(tmp_infile);							//5. Delete temporary files
    unlink(tmp_outfile);

    return 1;
}


int GEMLayoutOp::read_layout(const char* fname)
//Read positions from GraphEd file produced by 'gem'
{
    FILE* fp = fopen(fname,"r");
    if (!fp) return 0;

    Selection* g = GraphManager::getInstance()->getSelList()->getFullSelection();

    read_graphed_pos(fp,g,getPositionX(),getPositionY(),getPositionZ());		
								//Do the real readin of pos from GraphEd file
    fclose(fp);						
    return 1;
}


//------------  TSLayoutOp  --------------------------------------------



TSLayoutOp::TSLayoutOp()
	   :tx(0), ty(0), tz(0), sx(1), sy(1), sz(1)
{  
   keepDimensions(TRUE);					//we set our own dim-attrs (actually, we scale them)
}


int TSLayoutOp::layout()
{
   STRING pos_x = getPositionX(),  pos_y = getPositionY(),
          dim_x = getDimensionX(), dim_y = getDimensionY(), dim_z = getDimensionZ();

   GraphManager*  gm  = GraphManager::getInstance();
   Selection*     s   = getSelection();

   Selection::NodeIterator* i; Node* n;	Attr* a; float v;
   for(i=s->iterNodes();(n=i->next());)				//translate+scale node-positions and dims
   {
      v = n->getFloat(pos_x); n->set(pos_x,tx+sx*v);		//assume pos-attrs always exist		
      v = n->getFloat(pos_y); n->set(pos_y,ty+sy*v);
      if (getKeepDimensions())				
      {
        a = n->get(dim_x); if (a) a->set(sx*a->getFloat());	//don't assume dim-attrs always exist
        a = n->get(dim_y); if (a) a->set(sy*a->getFloat());
        a = n->get(dim_z); if (a) a->set(sz*a->getFloat());
      }
   }

   Selection::EdgeIterator* j; Edge* e; int sz;			//translate+scale edge-positions, if any 
   for(j=s->iterEdges();(e=j->next());)
   {
      float* v = e->getFloatVec(pos_x,sz);			//don't assume any of the X-pos or Y-pos edge-attrs
      if (v) for(sz--;sz>=0;sz--) v[sz] = tx+sx*v[sz];		//exist. If they do exist, then simply set them in-place,
      v = e->getFloatVec(pos_y,sz);				//since getFloatVec() returns a pointer to the real vector
      if (v) for(sz--;sz>=0;sz--) v[sz] = ty+sy*v[sz];	
   }

   return 1;
}

