#include "Node.h"
#include "Edge.h"
#include "Filter.h"
#include "Selection.h"
#include "GraphManager.h"
#include "DomainModel.h"
#include "SelList.h"
#include <iostream.h>

EdgeSet Node::empty_set;			//empty edge set, used as a return object


//-----------  EdgeSet  --------------------------------------------------------

Edge* EdgeSet::to(Node* n)			//Finds 1st Edge in this set to given Node, or null
{
   for(int i=0;i<Count();i++)
   { Edge* e = (*this)[i]; if (e->toNode()==n) return e; }
   return 0;
}


Edge* EdgeSet::from(Node* n)			//Finds 1st Edge in this set from given Node, or null
{
   for(int i=0;i<Count();i++)
   { Edge* e = (*this)[i]; if (e->fromNode()==n) return e; }
   return 0;
}




//----------  Node  -----------------------------------------------------------

Node::Node(Selection* s,int id,const AttrSetType* ast)
     :Element((ast)? ast : GraphManager::getInstance()->getDomainModel()->getUnknownNodeType(),id), 
      in_dtor(FALSE)                            //Ctor
{  
   Selection* all = GraphManager::getInstance()->getSelList()->getFullSelection();

   if (s && s!=all) s->addNode(this);		//add this to specified selection, if any given (and if not full-sel)
   all->addNode(this);                          //in any case, add this to full-sel
}



Node::~Node()					//Dtor. Deletes this and all Edges connected to this.
{   
}

void Node::destroy()				//Public dtor
{
   int N = numSelections();
 
   for(int i=0;i<N;i++)				//Inform all selections that own this that this goes away
      getSelection(0)->delNode(this);
       
}


Edge* Node::connectTo(Node* n,Selection* s,const AttrSetType* ast_,int id)	
{						//connect this->n in given context, if not already so.
						//return (new) Edge this->n
    Edge* e; 
    const AttrSetType* ast = (ast_)? ast_ : GraphManager::getInstance()->getDomainModel()->getUnknownArcType();

    //WARNING:    Originally we had a test "if ((e=connectionTo(n,s))==0)"
    //		  for the creation of the new edge and its addition "s->addEdge(e)".
    //		  This test is now removed as we must allow several edges
    //		  between the same two nodes - requirement imposed by the
    //		  (structured) RSF data model.
   
    e = new Edge(s,this,n,ast,id);     		//create Edge between this and n and add it to Selection s
    return e;    				
} 


void Node::addOutEdge(Edge* e,Selection* s)	//Add Edge e to out_edges and edges of this. Edge e is NOT yet added to this
{						//(at least not to out_edges)
    EdgeSet* es=0;

    std::multimap<unsigned long,EdgeSet*>::iterator i = out_edges.find((unsigned long)s);
    if (i==out_edges.end())
    {  
	es = new EdgeSet; 
	out_edges.insert(std::multimap<unsigned long,EdgeSet*>::value_type((unsigned long)s,es));
    }
    else
	es = (EdgeSet*)(*i).second;

    es->Add(e);					//Add() is safe, since we know 'e' is not yet added to this in 's'
    addEdge(e,s);				//Now add 'e' to all edges of this in context 's'
}


void Node::addInEdge(Edge* e,Selection* s)	//Add Edge e to in_edges and edges of this. Edge e is NOT yet added to this
{						//(at least not to in_edges)
    EdgeSet* es=0;

    std::multimap<unsigned long,EdgeSet*>::iterator i = in_edges.find((unsigned long)s);
    if (i==in_edges.end())
    {  
	es = new EdgeSet; 
	in_edges.insert(std::multimap<unsigned long,EdgeSet*>::value_type((unsigned long)s,es));
    }
    else
	es = (EdgeSet*)(*i).second;

    es->Add(e);					//Add() is safe, since we know 'e' is not yet added to this in 's' 
    addEdge(e,s);				//Now add 'e' to all edges of this in context 's'
}
       

void Node::addEdge(Edge* e,Selection* s)	//Add Edge e to 'edges' of this
{						//
    EdgeSet* es=0;

    std::multimap<unsigned long,EdgeSet*>::iterator i = edges.find((unsigned long)s);
    if (i==edges.end())
    {  
	es = new EdgeSet;
	edges.insert(std::multimap<unsigned long,EdgeSet*>::value_type((unsigned long)s,es));
    }
    else
    {
	es = (EdgeSet*)(*i).second;
    }

    es->Append(e);				//Add edge to 'context' for 's' of this
    e->nodeRef();				//Mark edge as being used by another node
}




void Node::delFromSelection(Selection* s)	//Enhances inh to remove EdgeSets of this in the context of Selection s
{
   in_dtor = TRUE;
   
   std::multimap<unsigned long,EdgeSet*>::iterator i = edges.find((unsigned long)s);
   if (i!=edges.end())
   {
      EdgeSet* es = (EdgeSet*)(*i).second; 
      for(int j=0;j<es->Count();j++) (*es)[j]->delFromNodes(s); 
      delete es; edges.erase(i); 
   }

   i = in_edges.find((unsigned long)s);
   if (i!=in_edges.end())
   {
      delete (EdgeSet*)(*i).second; in_edges.erase(i);
   }

   i = out_edges.find((unsigned long)s);
   if (i!=out_edges.end())
   {
      delete (EdgeSet*)(*i).second; out_edges.erase(i);
   }

   in_dtor = FALSE;
   Element::delFromSelection(s);		//Call inherited to delete this from selection 's'
}


EdgeSet Node::getEdges(Selection* s,Filter* f)
{
   return f->filter(s,this);
}


void Node::delEdge(Edge* e,Selection* s)
{
   std::multimap<unsigned long,EdgeSet*>::iterator i;
   if (in_dtor)   
   {
      e->nodeUnref(); 
   }
   else
   {
      i = edges.find((unsigned long)s);
      if (i!=edges.end())
      {
        EdgeSet* es = (EdgeSet*)(*i).second; 
	es->Remove(e);
	if (es->isEmpty()) {  delete es; edges.erase(i); }
	e->nodeUnref();
      } 

      i = in_edges.find((unsigned long)s);
      if (i!=in_edges.end())
      {
         EdgeSet* es = (EdgeSet*)(*i).second; 
         es->Remove(e);
         if (es->isEmpty()) { delete es; in_edges.erase(i); }
      }

      i = out_edges.find((unsigned long)s);
      if (i!=out_edges.end())
      {
         EdgeSet* es = (EdgeSet*)(*i).second; 
         es->Remove(e);
         if (es->isEmpty()) { delete es; out_edges.erase(i); }
      }
   }
}

