#include "SelList.h"
#include "Selection.h"
#include "Node.h"
#include "utils.h"


SelectionList::SelectionList()				//Ctor: 	add by default FullSelection and NullSelection with reserved names
{   							//REMARK:	we add these new Selections explicitly here for two reasons.
    sels.Add(new AllSelection);		 		//		First, there's no automatic adding via the GraphManager::cleanup()
    sels.Add(new NoneSelection);			// 		since we're NOT in an Operation-context here. Secondly, even if we 
							//		were, addSelection could NOT be called since these Selections have
							//		reserved names which are rejected by addSelection()
}	

SelectionList::~SelectionList()				//Dtor: delete all Selections in this,
{							//	since we own them.
    for(int i=0;i<sels.Count();i++)
       delete sels[i];
}


void SelectionList::delAllSelections()			//Delete all Selection objects
{
   for(int i=0;i<sels.Count();)
    if (!delSelection(sels[i]->getName())) i++;
}

void SelectionList::reset()
{
   delAllSelections();
   for(int i=0;i<sels.Count();i++)
      sels[i]->reset();
}
  

Selection* SelectionList::getSelection(const char* n)
{
    for(int i=0;i<sels.Count();i++)
       if ((*sels[i])==n) return sels[i];
    return getNullSelection();
}

void SelectionList::addSelection(Selection* s)
{   
    const char* name = s->getName();
    if (Selection::isReservedName(name)) 		//not allowed to add Selections with reserved name
    { delete s; return; }				//since the caller expects we'll manage 's', we simply delete it

    for(int i=0;i<sels.Count();i++)			//search for Selection with same name as 's'
      if ((*sels[i])==name)				//if found, delete it (we own it), and put 's' in its place
      {
	 if (sels[i]==s) return;			//if, by any chance, we try to add the same Selection-object,
	 delete sels[i];				//cancel the operation, the object is already physically in the list.
	 sels[i] = s;
	 return;
      }
      
    sels.Add(s);					//if not found, simply add 's' to list
}


BOOL SelectionList::renameSelection(Selection* sel,const char* ns)
{   
    if (sel->isReserved()) return FALSE;		//not allowed to change a reserved name!
    if (*sel==ns) return TRUE;				//rename with same name, nothing to do

    sels.Remove(sel);					//remove, but don't delete, selection
    if (Selection::isReservedName(ns))			//if we try to rename selection to a reserved-name,
       delete sel;					//simply delete it, since the reserved-sel is already there
    else					
    {							//if we try to rename selection to a normal name,
      sel->setName(ns);					//rename it, and then reinsert renamed selection
      addSelection(sel);				//This may delete an existing selection with same name
    }

    return TRUE;
}



BOOL SelectionList::delSelection(const char* n)		//destroys selection by name
{
   Selection*     s = getSelection(n);			//find selection with given name 
   const char* name = s->getName();

   if (Selection::isReservedName(name))			//not allowed to destroy reserved-name Selection
      return FALSE; 					//or selection with name 'n' not in this SelList

   sels.Remove(s);					//remove selection from this's list
   delete s;						//destroy removed selection
   return TRUE;	
}				
   


void SelectionList::save(FILE* fp,const char* sel_file) //save Tcl-cmd to restore selections from *.sel file
{
   fprintf(fp,"pr_read_sel_file %s\n",(const char*)addQuotes(sel_file));
}



void SelectionList::saveDefinition(const char* sel_file) //save definition of sels in this to *.sel file       
{                                                        //We save only non-reserved selections. The reason is
   FILE* fp = fopen(sel_file,"w");                       //that the default selections (All,None) don't need 
   if (!fp) return;                                      //to be restored, they're implicitly always existent.

   for(int i=0;i<sels.Count();i++)                       //ask every selection to save its definition
      if (!sels[i]->isReserved()) sels[i]->saveDefinition(fp);

   fclose(fp);
}



void SelectionList::read(const char* sel_file)           //load definition of sels into this from *.sel file
{
   FILE* fp = fopen(sel_file,"r");    
   if (!fp) return;

   while (readSelection(fp));                            //Keep calling readSelection() to read a new selection.
                                                         //This returns 0 when the file is finished.
   fclose(fp);
}



Selection* SelectionList::readSelection(FILE* fp)       //Reads new selection from file and adds it to the
{                                                       //SelectionList. If none could be read, returns 0.
   int id; char name[100];
   if (fscanf(fp,"%s",name)!=1) return 0;               //Read sel-name. If none, ready

   Selection* s = new Selection;                        //Make new selection.
   s->ref(name);                                        //Register it under its read name
   s = getSelection(name);                              //Read it back, 's' might have changed if that name was in use

   fscanf(fp,"%*s%*d");                                 //Skip "nodes" and #nodes
   for(;;)                                              //Read node-ids and stuff them in new sel
   { 
      if (fscanf(fp,"%d",&id)!=1) break;                        
      if (id==-1) break;                                //Stop if id -1 was read
      Node* n = getFullSelection()->getNodeById(id);    //Lookup node by id - if found, add it to sel
      if (n) s->addNode(n);
   }
   
   fscanf(fp,"%*s%*d");                                 //Skip "edges" and #edges
   for(;;)                                              //Read edge-ids and stuff them in new sel
   { 
      if (fscanf(fp,"%d",&id)!=1) break;                        
      if (id==-1) break;                                //Stop if id -1 was read
      Edge* e = getFullSelection()->getEdgeById(id);    //Lookup edge by id - if found, add it to sel
      if (e) s->addEdge(e);
   }  

   return s;
}