#include "Ops.h"
#include "GraphManager.h"
#include "AttrSetType.h"
#include "Node.h"
#include "Edge.h"
#include "Filter.h"
#include "TreeTraversal.h"
#include "Selection.h"
#include "SelList.h"
#include "Geometry.h"
#include "Genrl/string.h"
#include "Genrl/types.h"
#include <stdio.h>
#include <stdlib.h>
#include <set>



void SelectSetOp::dag(Selection* sel,Selection* all,Filter* filter,Selection*& res)
{								//Select all nodes in 'all' that descend from nodes in 'sel',
   begin();							//via 'filter', and put them in 'res'.
								//Also select the edges between these nodes.
   GraphManager*  gm = GraphManager::getInstance();
   TreeTraversal* ts = new TreeTraversal(sel,all,filter);
   res		     = new Selection;

   Node* n; TreeTraversal::NodeIterator* i;  				
   for(i=ts->iterNodes();(n=i->next());)
      res->addNode(n);
   Edge* e; TreeTraversal::EdgeIterator* j;			
   for(j=ts->iterEdges();(e=j->next());)
      res->addEdge(e);

   delete ts;
   end();
}



struct Item { Node* n; int l; Item(Node* n_,int l_=0):n(n_),l(l_){} };

struct  Pred
	{ 
	   bool operator()(const Item& i1,const Item& i2) const 
	   { return i1.n<i2.n; }
	};



void SelectSetOp::tree(Selection* sel,Selection* all,DirectionFilter* filter,const char* lev_attr,Selection*& res)
{								//Select all nodes in 'all' that descend from nodes in 'sel'
   begin();							//via 'filter'. Similar to SelectSetOp::dag() except that
								//this produces a strict tree, not a DAG. If a node is reachable via
   GraphManager*  gm = GraphManager::getInstance();		//several paths, only the first one (in order of discovery) is produced.
   res		     = new Selection;

   Selection* root = 0;						//First extract the roots of 'sel'. We need this for the traversal.
   DirectionFilter* rev_filter = (DirectionFilter*)filter->clone();
   rev_filter->setDirection((filter->getDirection()==Node::IN)? Node::OUT : Node::IN);
   SelectSetOp::roots(sel,rev_filter,root);
   delete rev_filter;						

   std::set<Item,Pred> q; std::set<Node*> c;			//Use q,c to perform the traversal instead of Selections, since the
   Node* n; Selection::NodeIterator* i;  			//STL sets are faster.		
   for(i=root->iterNodes();(n=i->next());)			//Put all root-nodes in 'res' as well as in 'q'
   {  if (q.find(n)==q.end()) { q.insert(n); res->addNode(n); } }
   delete root;
   
   for(;q.size();)						//While graph not completely traversed:
   {
      Item  i = *(q.begin()); q.erase(q.begin()); 		//Get 1st not traversed node and put it in 'c'
      Node* n = i.n; c.insert(n);
      int   l = i.l;
      
      if (lev_attr) n->set(lev_attr,l);				//!!!
      
      EdgeSet& es = n->getEdges(all,filter);			//Get all its edges
      for(int j=0;j<es.Count();j++)			
      {
         Edge* e = es[j]; Node* n2 = e->otherNode(n);		//Do we discover a new node or a not yet completely processed node?
	 if (c.find(n2)==c.end() && q.find(n2)==q.end())	
         { 
	   q.insert(n2); 
	   res->addNode(n2); res->addEdge(e); 
	 }
      }
   }
   end();
}



void SelectGenericOp::select(Selection*& res)
{
   begin();
   res = new Selection;
   end();
}


void SelectSetOp::reunion(Selection* s1,Selection* s2,Selection*& res)
{
   begin();

   GraphManager* gm  = GraphManager::getInstance();
   res 		     = new Selection;					//make new "res" selection

   Node* n; Selection::NodeIterator* i;  				//add all stuff in s1 to res
   for(i=s1->iterNodes();(n=i->next());) res->addNode(n);
   for(i=s2->iterNodes();(n=i->next());) res->addNode(n);

   Edge* e; Selection::EdgeIterator* j;					//add all stuff in s2 to res. Since Selection::addNode()
   for(j=s1->iterEdges();(e=j->next());) res->addEdge(e);		//and addEdge() work in append-mode, no duplicates are created here.
   for(j=s2->iterEdges();(e=j->next());) res->addEdge(e);

   end();
}


void SelectSetOp::intersect(Selection* s1,Selection* s2,Selection*& res)
{
   begin();
   GraphManager* gm  = GraphManager::getInstance();
   res 		     = new Selection;					//make new "res" selection
   
   Node* n; Selection::NodeIterator* i;  				//add all stuff in s1 AND s2 to to res
   for(i=s1->iterNodes();(n=i->next());)
      if (s2->hasNode(n)) res->addNode(n);
   Edge* e; Selection::EdgeIterator* j;					
   for(j=s1->iterEdges();(e=j->next());)			
      if (s2->hasEdge(e)) res->addEdge(e);
   end();
}




void SelectSetOp::difference(Selection* s1,Selection* s2,Selection*& res)
{
   begin();
   GraphManager* gm  = GraphManager::getInstance();
   res 		     = new Selection;					//make new "res" selection

   Node* n; Selection::NodeIterator* i;  				//add all stuff in s1 EXCEPT s2 to to res
   for(i=s1->iterNodes();(n=i->next());)
      if (!s2->hasNode(n)) res->addNode(n);
   Edge* e; Selection::EdgeIterator* j;					
   for(j=s1->iterEdges();(e=j->next());)			
      if (!s2->hasEdge(e)) res->addEdge(e);
   end();
}






void SelectSetOp::xor(Selection* s1,Selection* s2,Selection*& res)
{
   begin();
   GraphManager* gm  = GraphManager::getInstance();
   res 		     = new Selection;					//make new "res" selection

   Node* n; Selection::NodeIterator* i;  				//add all stuff in s1 EXCEPT s2 to to res
   for(i=s1->iterNodes();(n=i->next());)
      if (!s2->hasNode(n)) res->addNode(n);
   for(i=s2->iterNodes();(n=i->next());)
      if (!s1->hasNode(n)) res->addNode(n);
   Edge* e; Selection::EdgeIterator* j;					
   for(j=s1->iterEdges();(e=j->next());)			
      if (!s2->hasEdge(e)) res->addEdge(e);
   for(j=s2->iterEdges();(e=j->next());)			
      if (!s1->hasEdge(e)) res->addEdge(e);
   end();
}





void SelectSetOp::edges(Selection* sn,Selection* se,Selection*& res)
{
   begin();

   res = new Selection;						//make new "res" selection

   Node* n; Selection::NodeIterator* i;  			//add all stuff in n1 to res
   for(i=sn->iterNodes();(n=i->next());)			//first, add nodes and all	
   {								//their edges that lead to nodes 
      res->addNode(n);						//in 'sn' too
      EdgeSet& es = n->getEdges(se);
	
      for(int j=0;j<es.Count();j++)
      {
	 Edge* e = es[j];	
	 if (sn->hasNode(e->otherNode(n))) res->addEdge(e);
      }
   }

   Edge* e; Selection::EdgeIterator* j;				//then, add all edges that were
   for(j=sn->iterEdges();(e=j->next());)			//originally in 's1'	
       res->addEdge(e);

   end();
}


void SelectSetOp::nodes(Selection* s,Selection*& res)
{
   begin();
   res 		     = new Selection;				//make new "res" selection

   Node* n; Selection::NodeIterator* i;  			//add only nodes from s to res
   for(i=s->iterNodes();(n=i->next());) res->addNode(n);

   end();
}




void SelectSetOp::roots(Selection* s,Filter* f,Selection*& res)//Selects nodes in 'input' having no descendants
{  								//in 'input'. Descendants are found following filter 'f'
   begin();

   res = new Selection;						//make new "res" selection
   Selection::NodeIterator* i = s->iterNodes();			//find all nodes in 's' that have ONLY descendants
   for(Node* n;(n=i->next());)					//that are NOT in 's'
   {
      EdgeSet p = n->getEdges(s,f);
      if (!p.Count())						//'n' has no descendants at all or all descendants 
         res->addNode(n);					//aren't in 's' -> 'n' is a leaf-node of 's'
   }

   end();
}

