#include "Selection.h"
#include "Node.h"
#include "Edge.h"
#include "GraphManager.h"
#include "SelList.h"
#include "utils.h"
#include <set>


const char* Selection::MAKE_NAME = 0;		//Special name given when we
						//want the Selection to make its own, unique name.
int Selection::make_name_id = 0;		//Used to make unique name_id entries
int Selection::num_selections = 0;		//Total # created Selection objs, used to reset make_name_id
static std::set<int> free_ids;

const char* Selection::ALL  = "All";		//Special names of the full and empty selections
const char* Selection::NONE = "None";


Selection::Selection(const char* n): GVObject(make_name(n,name_id)), in_dtor(FALSE)
{  num_selections++; touch();  }


void Selection::ref(const char* name)				//rename this to new name and register as such in SelectionList
{								
   STRING n = (name!=MAKE_NAME)? name : getName();		//if no name given, register this under its own name	

   if (isReserved()) return;					//check the name under which we want to do the ref() is
   if (isReservedName(n))					//correct. If not and this is a reserved object, simply ignore
   { delete this; return; }					//ref(), else destroy this and return - since the object with
								//the required (reserved) name already exists.

   SelectionList* sl = GraphManager::getInstance()->getSelList();
   Selection*    old = sl->getSelection(n);			//do we kick someone out of SelectionList when we get in?
   if (!old->isReserved())					//yes, we do. Then exchange our name_id with that Selection
   {								//since a) we'll use its name (and thus name_id too) and
      int tmp = name_id;					//b) we release our name (and name_id too)
      name_id = old->name_id; old->name_id = tmp; 
   }
   setName(n);
   sl->addSelection(this);

   touch();							//REMARK: actually this is a bit fake, 
								//we should touch the SelectionList, NOT this selection
}

const char* Selection::make_name(const char* name,int& id)	//Transforms user 'name' into a valid Selection-name.
{								//Basically 'name' stays the same if it is not equal
  static char buf[50];						//to MAKE_NAME, else'name ' is a unique 'name_id' amonst
  if (name!=MAKE_NAME) { id = -1; return name; }		//all existing Selection objects. The numerical id is
								//returned in the last arg.
  if (free_ids.size()) 
  { id = *(free_ids.begin()); free_ids.erase(free_ids.begin()); }
  else
  { id = make_name_id++; }
  sprintf(buf,"%d",id);
  					
  return buf;		
}								
  

Selection::Iterator::~Iterator()				//Dtor
{  }

BOOL Selection::nodesHaveAttr(const char* name,Attr::TYPE* type)
{
  const Node* n; BOOL ret = TRUE;
  NodeIterator* i;
  for(i=iterNodes();((n=i->next())!=0);)
  {
	const Attr* a = n->get(name);
	if ((!a) || (type && (!(a->getType()==*type)))) 
	{ ret = FALSE; break; }
  }
  
  return ret;
}

BOOL Selection::edgesHaveAttr(const char* name,Attr::TYPE* type)
{
  const Edge* e; BOOL ret = TRUE;
  EdgeIterator* i;
  for(i=iterEdges();((e=i->next())!=0);)
  {
	const Attr* a = e->get(name);
	if ((!a) || (type && (!(a->getType()==*type))))
	{ ret = FALSE; break; }
  }

  return ret;
}


void Selection::delNodeAttr(const char* name)
{
  Node* n; NodeIterator* i;
  for(i=iterNodes();(n=i->next());)
     n->del(name);
  touch();
}

void Selection::delEdgeAttr(const char* name)
{
  Edge* e; EdgeIterator* i;
  for(i=iterEdges();(e=i->next());)
     e->del(name);
  touch();
}


Selection::~Selection()					//Dtor
{							//Reset this
   reset();						//If we allocated a 'name_id', free it
   if (name_id!=-1) free_ids.insert(name_id);		//so new Selections can use it

   if (!--num_selections)				//One less selection. If all Selections destroyed,
   {							//we can safely reset the name_id-making system, in
      make_name_id = 0;					//order to recycle the name_ids.
      free_ids.clear();
   }
}


void Selection::reset()
{
   if (in_dtor) return; in_dtor = TRUE;			//prevents recursive calls of this
                                                        //and also accelerates removal of nodes/edges
   for(std::map<int,Node*>::iterator i=node_map.begin();i!=node_map.end();i++)
     ((*i).second)->delFromSelection(this);             //remove nodes. This might trigger edge removal from this too

   for(std::map<int,Edge*>::iterator j=edge_map.begin();j!=edge_map.end();)
   {                                                    //remove edges. Since Edge::delFromSelection() DOES trigger edge removal
     Edge* e = (*j).second; j++;                        //from this (which in turn invalidates the iterator), advance the iterator 
     e->delFromSelection(this);                         //BEFORE calling Edge::delFromSelection()
   }

   node_map.clear();                                     //empty idx-to-node and idx-to-edge maps 
   edge_map.clear();                                     
   touch();

   in_dtor = FALSE;
}

void Selection::delNode(Node* n)
{  
   if (node_map.erase(n->getId()))                      //if node was indeed in this
   { n->delFromSelection(this); touch(); }              //remove node from idx-map too
}

void Selection::delEdge(Edge* e)
{  
   if (edge_map.erase(e->getId()))                      //if edge was indeed in this
   {  e->delFromSelection(this); touch(); }              //remove edge from idx-map too 
}

void Selection::removeEdge(Edge* e)                     //Used only by Edge class
{  /*if (!in_dtor)*/ edge_map.erase(e->getId()); }



void Selection::flush()			//destroys all nodes/edges in this selection.
{					//This is different than reset() which just unreleases the nodes/edges in this.
   for(;;)				//In that case, those nodes/edges can still exist if they are referred to by other selections.
   {					//flush() will completely destroy (remove from the system) the contents of this.
       std::map<int,Node*>::iterator i=node_map.begin();
       if (i!=node_map.end()) ((*i).second)->destroy();
       else break;
   }

   for(;;)
   {
       std::map<int,Edge*>::iterator i=edge_map.begin();
       if (i!=edge_map.end()) ((*i).second)->destroy();
       else break;
   }

   //REMARK: we use the above for(;;) construction instead of a normal iteration over the std::map<> iterators since
   //        Node:: and Edge::destroy() invalidate the current iterators...
}


void Selection::addNode(Node* n)
{  
   if (hasNode(n)) return;				//must check this, addToSelection relies upon
   std::map<int,Node*>::value_type v(n->getId(),n);     //add (node-id,node) for fast node location
   node_map.insert(v);
   n->addToSelection(this); 
   touch();  
}

void Selection::addEdge(Edge* e)
{  
   if (hasEdge(e)) return;				//must check this, addToSelection relies upon
   std::map<int,Edge*>::value_type v(e->getId(),e);     //add (edge-id,edge) for fast edge location
   edge_map.insert(v);
   e->addToSelection(this); 
   touch();  
}



BOOL Selection::hasNode(Node* n) const			//Checks node is in this. For this, we actually ask the node if it is contained
{							//in this, since the latter operation is almust surely faster than searching 
   return n->containedIn((Selection*)this);		//this full selection. Further on: the cast is safe.	
}  

BOOL Selection::hasEdge(Edge* e) const			//Checks edge is in this. Same comments as for hasNode().
{  
   return e->containedIn((Selection*)this);
}



Node* Selection::getNodeById(int id)			//Finds node by id. Returns Node or 0 if none.
{  
   std::map<int,Node*>::iterator i = node_map.find(id);
   return (i!=node_map.end())? (*i).second : 0;
}


   
Node* Selection::getNodeByAttr(const char* name,const char* val_)
{
   STRING val = val_; Node* n=0;
   for(NodeIterator* iter=iterNodes();(n=iter->next());)
   {
      const Attr* a = n->get(name);			//Search for an attr with given name
      if (a && val==a->getString()) break;              //If found, break, we're done
   }
   if (n) delete iter;                                  //If we found sth before iterator is done, must destroy iterator
   return n;                                            //Return found node or 0
}

 

Edge* Selection::getEdgeById(int id)			//Finds edge by id. Returns Edge or 0 if none.
{  
   std::map<int,Edge*>::iterator i = edge_map.find(id);
   return (i!=edge_map.end())? (*i).second : 0;			
}


   
Edge* Selection::getEdgeByAttr(const char* name,const char* val_)
{
   STRING val = val_; Edge* e=0;
   for(EdgeIterator* iter=iterEdges();(e=iter->next());)
   {
      const Attr* a = e->get(name);			//Search for an attr with given name
      if (a && val==a->getString()) break;              //If found, break, we're done
   }
   if (e) delete iter;                                  //If we found sth before iterator is done, must destroy iterator
   return e;                                            //Return found edge or 0
}



BOOL Selection::isReserved() const			//This is not a reserved object
{  return FALSE;  }


void Selection::saveDefinition(FILE* fp)                //Save definition of this in *.sel file
{
   fprintf(fp,"%s\n",(const char*)addQuotes(getName()));
   fprintf(fp,"nodes %d\n",numNodes());
   Node* n; Edge* e;
   int ln=0;
   const int ln_max = 20;                               //Max #items/line to write in the file
   for(NodeIterator* i=iterNodes();(n=i->next());)      //Write node-ids in this	
   {
      fprintf(fp,"%d ",n->getId()); ln++;
      if (ln==ln_max) { fprintf(fp,"\n"); ln=0; }
   }
   ln=0;
   fprintf(fp,"\n-1\n");                                //Finish writing node-ids with a -1 fake entry
   fprintf(fp,"edges %d\n",numEdges());
   for(EdgeIterator* j=iterEdges();(e=j->next());)      //Write edge-ids in this	
   {
      fprintf(fp,"%d ",e->getId()); ln++;
      if (ln==ln_max) { fprintf(fp,"\n"); ln=0; }
   }
   fprintf(fp,"\n-1\n");                                //Finish writing edge-ids with a -1 fake entry
}



//----------  AllSelection, NoneSelection  -----------------------------------------------------

void NoneSelection::addNode(Node*)			//Disabled, can't add nodes to empty-selection
{   }

void NoneSelection::addEdge(Edge*)			//Disabled, can't add edges to empty-selection
{   }

BOOL NoneSelection::isReserved() const			//This is a reserved object
{  return TRUE;  }

BOOL AllSelection::isReserved() const			//This is a reserved object
{  return TRUE;  }


