#
# This file contains the Tk GUIs for setting the various parameters of the
# default glyph engines. Every engine-type has its own tab-panel where the engine-instances
# of that type can be edited individually.
# 
# To add a GUI for a new (node/edge) engine-type, one should:
#
# a) add a registry entry for it in reginit.tcl (see that file)
# b) define in this file 3 procedures. If the last parameter of the registry entry is called, say, xxx, 
#    then these procedures should have the following signatures and meanings:
#
#    pr_gui:glyphs:make_xxx {wid top opts}	constructs the specific widgets for the glyph-type into $top
#								$opts can be used to align their labels.
#    pr_gui:glyphs:select_xxx {wid sel}         update GUI in $wid to reflect the params of the glyph-engine called $sel
#							      (which is guaranteed to be of the correct type for this GUI)
#    pr_gui:glyphs:ok_xxx {wid}			should set all the kernel-params for the selected glyph-engine from the
#								current values present in the GUI fields.
#
#--------------------------------------------------------------------


# pr_gui:glyphs: Main entry point. Builds the glyph GUI
#---------------------------------------------------------------------
proc pr_gui:glyphs {wid} {
   if {[info commands .$wid]!={}} {
      raise .$wid
      return
   }

   toplevel .$wid

   #Make frame for tabs for node glyph engines & fill it in from registry
   frame .$wid.nd
   tixNoteBook .$wid.nd.nb
   set l [pr_register get node_glyph_factory]
   set i 0
   set n [llength $l]
   while {$i<$n} {
      set e      [lindex $l $i]; set label  [lindex $e 0]
      set title  [lindex $e 1];  set type   [lindex $e 2]
      set widget [lindex $e 3]
      .$wid.nd.nb add $widget -label $label -raisecmd [concat "pr_gui:glyphs:set_instances" .$wid.nd.nb $widget $type node_glyph]
      
      #Construct the specific widgets for current engine-type
      pr_gui:glyphs:shell_make [.$wid.nd.nb subwidget $widget] $title $widget   
      incr i
   }
   pack .$wid.nd.nb -side top -expand yes -fill both
   pack .$wid.nd -side top -fill both -expand yes 

   #Make frame for tabs for edge glyph engines & fill it in from registry
   frame .$wid.ed
   tixNoteBook .$wid.ed.nb
   set l [pr_register get edge_glyph_factory]
   set i 0
   set n [llength $l]
   while {$i<$n} {
      set e      [lindex $l $i]; set label  [lindex $e 0]
      set title  [lindex $e 1];  set type   [lindex $e 2]
      set widget [lindex $e 3]
      .$wid.ed.nb add $widget -label $label -raisecmd [concat "pr_gui:glyphs:set_instances" .$wid.ed.nb $widget $type edge_glyph]
      
      #Construct the specific widgets for current engine-type
      pr_gui:glyphs:shell_make [.$wid.ed.nb subwidget $widget] $title $widget   
      incr i
   }
   pack .$wid.ed.nb -side top -expand yes -fill both
   pack .$wid.ed -side top -fill both -expand yes 

   tixButtonBox .$wid.box -orientation horizontal
   .$wid.box add ok -text "Apply"   -command "pr_gui:glyphs:ok .$wid"
   .$wid.box add cc -text "Dismiss" -command "destroy .$wid"
   pack .$wid.box -side bottom -fill x

   bind .$wid <Destroy> { destroy %W }
}


proc pr_gui:glyphs:show_node_engine {wid neng} {
  pr_gui:glyphs $wid
  #!!! not finished - must show neng node-engine's details in wid
}

proc pr_gui:glyphs:show_edge_engine {wid eeng} {
  pr_gui:glyphs $wid
  #!!! not finished - must show eeng edge-engine's details in wid
}

#-------------------------------------------------------------------------------
# Private (implementation) part 
#-------------------------------------------------------------------------------


# pr_gui:glyphs:shell_make: constructs the common part for every 
#                           tab-panel and then calls the specific construction
#--------------------------------------------------------------------------------
proc pr_gui:glyphs:shell_make {wid txt name} {

   frame $wid.top
   set top $wid.top
   pack $top -fill x
   set opts { label.width 15 label.anchor w }
   set cmd  [join [list "pr_gui:glyphs:select_" $name] ""]

   label         $top.l -text $txt -justify center; pack $top.l
   tixComboBox   $top.eng -label "Engine name" -options $opts -dropdown true -editable false\
		     -command [concat $cmd $wid]; pack $top.eng -fill x	

   [join [list "pr_gui:glyphs:make_" $name] ""] $wid $top $opts
}


# pr_gui:glyphs:make_ntcl: constructs the page for the Tcl node-glyph-engine
#------------------------------------------------------------------------------------
proc pr_gui:glyphs:make_ntcl {wid top opts} {
   pr_gui:combo_edit $top.ndcb "Node Tcl proc" $opts [pr_qualify RSCS/edit.gif] "pr_gui:glyphs:edit_node" "Edit Tcl node-glyph script"
}

# pr_gui:glyphs:select_ntcl: refreshes the fields of the Tcl node-glyph-engine 
#                            page to reflect selected instance
#-------------------------------------------------------------------------------
proc pr_gui:glyphs:select_ntcl {wid sel} {
    set top $wid.top
    if {$sel!=""} {
       set ngf [pr_glyphs get node_glyph name $sel]
       pr_gui:combo_edit_set $top.ndcb [pr_glyph_factory $ngf get cb]
    }
}

# pr_gui:glyphs:ok_ntcl: copies params from the Tcl node-glyph-engine 
#                        page to the kernel via the Tcl API.
#------------------------------------------------------------------------------
proc pr_gui:glyphs:ok_ntcl {wid} {
    set top [$wid subwidget ntcl] 
    append top ".top"
    set ngf [$top.eng cget -value]
    if {$ngf==""} return
    set ndcb [pr_gui:combo_edit_get $top.ndcb]
    pr_glyph_factory [pr_glyphs get node_glyph name $ngf] set cb $ndcb
}


# pr_gui:glyphs:make_etcl {parent}: constructs the page for the Tcl edge glyph engine
#--------------------------------------------------------------------------------
proc pr_gui:glyphs:make_etcl {wid top opts} {
   pr_gui:combo_edit $top.edcb "Edge Tcl proc" $opts [pr_qualify RSCS/edit.gif] "pr_gui:glyphs:edit_edge" "Edit Tcl edge-glyph script"
}

# pr_gui:glyphs:select_etcl: as select_ntcl but for the Tcl edge glyph engines
#-------------------------------------------------------------------------------
proc pr_gui:glyphs:select_etcl {wid sel} {
    set top $wid.top
    if {$sel!=""} {
       set egf [pr_glyphs get edge_glyph name $sel]
       pr_gui:combo_edit_set $top.edcb [pr_glyph_factory $egf get cb]
    }
}

# pr_gui:glyphs:ok_etcl: as ok_ntcl but for the Tcl edge glyph engines
#------------------------------------------------------------------------------
proc pr_gui:glyphs:ok_etcl {wid} {
    set top [$wid subwidget etcl] 
    append top ".top"
    set egf [$top.eng cget -value]
    if {$egf==""} return
    set edcb [pr_gui:combo_edit_get $top.edcb]
    pr_glyph_factory [pr_glyphs get edge_glyph name $egf] set cb $edcb
}




# pr_gui:glyphs:make_nbasic {parent}: constructs the page for the basic glyph-engines
#--------------------------------------------------------------------------------
proc pr_gui:glyphs:make_nbasic {wid top opts} {
}

# Refreshes the Basic glyph-engine page
#
#-------------------------------------------------------------------------------
proc pr_gui:glyphs:select_nbasic {wid sel} {
    set top $wid.top
    if {$sel!=""} {
         #update fields if any...
    }
}

proc pr_gui:glyphs:ok_nbasic {wid} {
}

# pr_gui:glyphs:make_ebasic {parent}: constructs the page for the basic glyph-engines
#--------------------------------------------------------------------------------
proc pr_gui:glyphs:make_ebasic {wid top opts} {
}

# Refreshes the Basic glyph-engine page
#-------------------------------------------------------------------------------
proc pr_gui:glyphs:select_ebasic {wid sel} {
    set top $wid.top
    if {$sel!=""} {
         #update fields if any...
    }
}

proc pr_gui:glyphs:ok_ebasic {wid} {
}

# pr_gui:glyphs:ok{parent}: sets all values from the UI's fields to the system via the Tcl API
#-----------------------------------------------------------------------------------------
proc pr_gui:glyphs:ok {wid} {

# get both current node and edge panels and apply them

  set page [$wid.nd.nb raised]
  set cmd [join [list "pr_gui:glyphs:ok_" $page] ""]
  $cmd $wid.nd.nb
  set page [$wid.ed.nb raised]
  set cmd [join [list "pr_gui:glyphs:ok_" $page] ""]
  $cmd $wid.ed.nb

}


# set_instances{wid name type kind}: fills in combo-box of wid with names of all glyph-factory instances
#						 for C++ class 'type' of kind 'kind'
#-------------------------------------------------------------------------------------------------------
proc pr_gui:glyphs:set_instances {wid name type kind} {
  set wid [$wid subwidget $name]  
  [$wid.top.eng subwidget listbox] delete 0 end

  set i 0
  set n [pr_glyphs get $kind number]
  while {$i<$n} {
     set eng [pr_glyphs get $kind idx $i] 
     set tt  [pr_glyph_factory $eng get type]  
     if {$tt==$type} {
       $wid.top.eng insert end [pr_glyph_factory $eng get name] 
     }
     incr i
  }
  $wid.top.eng pick 0
}





# edit_node: Starts an editor for customizing the node-glyph Tcl proc  
#
#--------------------------------------------------------------------
proc pr_gui:glyphs:edit_node {wid} {
    set ndcb [pr_gui:combo_edit_get $wid]
    pr_gui:edit_proc edit $ndcb {node}
}

# edit_edge: Starts an editor for customizing the edge-glyph Tcl proc  
#
#--------------------------------------------------------------------
proc pr_gui:glyphs:edit_edge {wid} {
    set edcb [pr_gui:combo_edit_get $wid]
    pr_gui:edit_proc edit $edcb {edge pos_x pos_y pos_z}
}

