#
# Miscellaneous utility procedures. Users may add/customise these at their leisure.
# These procedures are registered as components. This makes it easy to share them 
# among different application-files.
#----------------------------------------------------------------------------------------


# 1. COMPONENT REGISTRATION SECTION
#    In this section, all components are declared
#------------------------------------------------------------------------------------
pr_compo_lib add_proc pr_viewer_cmd:sweep_cb "Viewer default sweep callback"


# 2. COMPONENT DEFINITION SECTION
#    In this section, all components are defined (the Tcl-procs are defined)
#------------------------------------------------------------------------------------


# pr_viewer_cmd:sweep_cb{opt val} This is the default sweep-cb, called by the viewer when in
#				          sweep-mode. As the mouse moves, the cb gets the node or
#			edge (opt="node" or opt="edge", and val=id thereof). The cb may return a string
#		      which will be displayed (if not "") by the viewer at the mouse-location, as a brush.
#			Newlines in this string are seen as multiline delimiters.



proc pr_viewer_cmd:sweep_cb {opt val} {
  set ret ""
  if {[string equal $opt "node"]} { pr_gui:attrset_refresh_id .nodes $val; return $val }
  if {[string equal $opt "edge"]} { pr_gui:attrset_refresh_id .nodes $val; return $val }
  return ""
}





# pr_examine: simple procedure that pops up the file selector and allows one to
#	        load, lay out, map, and view a graph-file in one step.	
#             Useful for quickly 'browsing' thru a set of graph-files.
#------------------------------------------------------------------------------
proc pr_examine {} {
   pr_reset
   pr_load_dlg
   pr_lay_neato " " pos
   pr_view_new vw " " pos vw_highlight
}




# pr_select_edges_by_type: filter that produces sel2 = sel except the edges of type 'etyp'
#                          (if yes_no==0) or sel2 = sel including only the edges of type 'etyp'
#			         (if yes_no==1)
#------------------------------------------------------------------------------------------
proc pr_select_edges_by_type {sel sel2 etyp yes_no} {

  set res [pr_sel_new $sel2]
  set inp [pr_get_selection name $sel]

  set i [pr_selection $inp iter_nodes]
  set n [pr_iterator $i next_node]

  while {$n!=0} {
    pr_selection $res add_node $n 
    set n [pr_iterator $i next_node]
  }

  set i [pr_selection $inp iter_edges] 
  set e [pr_iterator $i next_edge]

  if {$yes_no==0} {
    while {$e!=0} {
      if {[pr_edge $e type]!=$etyp} {
         pr_selection $res add_edge $e 
      }
      set e [pr_iterator $i next_edge]
    }
  } else {
    while {$e!=0} {
      if {[pr_edge $e type]==$etyp} {
         pr_selection $res add_edge $e 
      }
      set e [pr_iterator $i next_edge]
    }
  }
}


# pr_filter_edges_by_type: very similar to pr_select_edges_by_type, but produces only
#				   the nodes connected by the selected edges, given as a set of types. 
#---------------------------------------------------------------------------------------
proc pr_filter_edges_by_type {sel sel2 etyp} {
  set res [pr_sel_new $sel2]
  set inp [pr_get_selection name $sel]
  set i   [pr_selection $inp iter_edges]
  set e   [pr_iterator $i next_edge]

  while {$e!=0} {
      set j 0; set n [llength $etyp]
      while {$j<$n} {
         set typ [lindex $etyp $j]; set t [pr_edge $e type]	
         if {$typ==$t} {break}
         incr j
      }
      if {$j<$n} {
         pr_selection $res add_edge $e 
	   set n1 [pr_edge $e node from]; set n2 [pr_edge $e node to]
         pr_selection $res add_node $n1; pr_selection $res add_node $n2
      }
      set e [pr_iterator $i next_edge]
  }
}


# pr_select_nodes_by_type: filter that produces sel2 = sel except the nodes of type 'ntyp'
#                          (if yes_no==0) or sel2 = sel including only the nodes of type 'ntyp'
#			         (if yes_no==1)
#------------------------------------------------------------------------
proc pr_select_nodes_by_type {sel sel2 ntyp yes_no} {

  set res [pr_sel_new $sel2]
  set inp [pr_get_selection name $sel]

  set i [pr_selection $inp iter_edges]
  set e [pr_iterator $i next_edge]

  while {$e!=0} {
    pr_selection $res add_edge $e 
    set e [pr_iterator $i next_edge]
  }

  set i [pr_selection $inp iter_nodes] 
  set n [pr_iterator $i next_node]

  if {$yes_no==0} {
    while {$n!=0} {
      if {[pr_node $n type]!=$ntyp} {
         pr_selection $res add_node $n 
      }
      set n [pr_iterator $i next_node]
    }
  } else {
    while {$n!=0} {
      if {[pr_node $n type]==$ntyp} {
         pr_selection $res add_node $n 
      }
      set n [pr_iterator $i next_node]
    }
  }
}



# pr_select_nodes_by_value: filter that produces all nodes in sel (and edges in between) that have 
#                           a given attribute equal to a given value
#--------------------------------------------------------------------------------------------------
proc pr_select_nodes_by_value {sel sel2 attr val} {
  set res [pr_sel_new $sel2]
  set inp [pr_get_selection name $sel]

  set i [pr_selection $inp iter_nodes]; set n [pr_iterator $i next_node]
  while {$n!=0} {
      if {[pr_node $n attr $attr]==$val} { pr_selection $res add_node $n }
      set n [pr_iterator $i next_node]
  }

  set i [pr_selection $inp iter_edges]; set e [pr_iterator $i next_edge]
  while {$e!=0} {
     set f [pr_edge $e node from]; set t [pr_edge $e node to]
     if [expr [pr_selection $res contains node $f] && [pr_selection $res contains node $t]] {
        pr_selection $res add_edge $e
     }
     set e [pr_iterator $i next_edge]
  }
}

# pr_select_edges_by_value: filter that produces all edges that have a given attribute 
#                           equal to a given value. Nodes at edges' ends are added too.
#----------------------------------------------------------------------------------------
proc pr_select_edges_by_value {sel sel2 attr val} {

  set res [pr_sel_new $sel2]
  set inp [pr_get_selection name $sel]

  set i [pr_selection $inp iter_nodes] 
  set n [pr_iterator $i next_node]
  while {$n!=0} {
      pr_selection $res add_node $n
      set n [pr_iterator $i next_node]
  }

  set i [pr_selection $inp iter_edges]
  set e [pr_iterator $i next_edge]
  while {$e!=0} {
     if {[pr_edge $e attr $attr]==$val} {pr_selection $res add_edge $e}
     set e [pr_iterator $i next_edge]
  }
}




# pr_set: simple procedure that sets 'attr' to 'value' of 'type' for all nodes/edges in selection
#	
#
#------------------------------------------------------------------------------
proc pr_set {sel attr val typ} {
  set i [pr_selection $sel iter_nodes]
  set n [pr_iterator $i next_node]
  while {$n!=0} { 
     pr_node $n set $attr $val $typ; set n [pr_iterator $i next_node] 
  }

  set i [pr_selection $sel iter_edges]
  set e [pr_iterator $i next_edge]
  while {$e!=0} {  
     pr_edge $e set $attr $val $typ; set e [pr_iterator $i next_edge] 
  }
}

# pr_unset: simple procedure that deletes 'attr' for all nodes/edges in selection
#	
#
#------------------------------------------------------------------------------
proc pr_unset {sel attr} {
  set i [pr_selection $sel iter_nodes]
  set n [pr_iterator $i next_node]
  while {$n!=0} { pr_node $n unset $attr; set n [pr_iterator $i next_node] }

  set i [pr_selection $sel iter_edges]
  set e [pr_iterator $i next_edge]
  while {$e!=0} { pr_edge $e unset $attr; set e [pr_iterator $i next_edge] }
}



# pr_pin: simple procedure that pins/unpins a given selection of nodes.
#         Useful to exclude some nodes from layout. For example, 
#         use 'pr_pin hl_sel 1', where hl_sel is the highlight-selection
#         of some viewer, to exclude the highlighted nodes from re-layout.
#------------------------------------------------------------------------------
proc pr_pin {sel yes_no} {
  set i [pr_selection [pr_get_selection name $sel] iter_nodes]
  set n [pr_iterator $i next_node]

  while {$n!=0} {
    pr_node $n set pos_pin $yes_no int
    set n [pr_iterator $i next_node]
 }
}


#
#  Arrangement procs. These are counterparts to the pr_lay_xxx procs but work 
#  		      within a prescribed bbox.
#------------------------------------------------------------------------------

proc pr_arrange_dot {sel pos} {
   set bbox [pr_bbox [pr_get_selection name $sel] $pos]
   pr_lay_dot $sel $pos bbox [lindex $bbox 0] [lindex $bbox 1] [lindex $bbox 2] [lindex $bbox 3]
}

proc pr_arrange_neato {sel pos} {
   set bbox [pr_bbox [pr_get_selection name $sel] $pos]
   pr_lay_neato $sel $pos bbox [lindex $bbox 0] [lindex $bbox 1] [lindex $bbox 2] [lindex $bbox 3]
}

proc pr_arrange_gem {sel pos} {
   set bbox [pr_bbox [pr_get_selection name $sel] $pos]
   pr_lay_gem $sel $pos bbox [lindex $bbox 0] [lindex $bbox 1] [lindex $bbox 2] [lindex $bbox 3]
}

proc pr_arrange_grid {sel pos} {
   set bbox [pr_bbox [pr_get_selection name $sel] $pos]
   pr_lay_grid $sel $pos bbox [lindex $bbox 0] [lindex $bbox 1] [lindex $bbox 2] [lindex $bbox 3]
}


#------------------------------------------------------------------------------------------
# pr_gui:combo_edit pre-packed GUI control with an entry-field and a right-hand button with callback.
#------------------------------------------------------------------------------------------

proc pr_gui:combo_edit {wid label opts icon cmd balloon} {
   frame         $wid
   tixLabelEntry $wid.entry -label $label -options $opts
                 pack $wid.entry -fill x -side left -expand yes
   button        $wid.btn   -text  "Edit" -image [image create photo -file $icon] -command "$cmd $wid"
		     pack $wid.btn -side right
   if {$balloon!={}} {
      tixBalloon $wid.balloon -initwait 10
      $wid.balloon bind $wid.btn -msg $balloon
   }
   pack $wid -fill x	
}

proc pr_gui:combo_edit_get {wid} { return [[$wid.entry subwidget entry] get] }

proc pr_gui:combo_edit_set {wid txt} {
   set v [$wid.entry subwidget entry]
   $v  delete 0 [string length [$v get]] 
   $v  insert 0 $txt
}


#-------------------------------------------------------------------------------------------------
# pr_gui:cbox_edit pre-packed GUI control with a combo-box and a right-hand button with callback.
#-------------------------------------------------------------------------------------------------

proc pr_gui:cbox_edit {wid label opts icon cmd cmd2 balloon} {
   frame         $wid
   tixComboBox   $wid.entry -label $label -options $opts -dropdown true -editable false -command $cmd2
                 pack $wid.entry -fill x -side left -expand yes 
   button        $wid.btn   -text  "Edit" -image [image create photo -file $icon] -command "$cmd $wid"
		     pack $wid.btn -side right
   if {$balloon!={}} {
      tixBalloon $wid.balloon -initwait 10
      $wid.balloon bind $wid.btn -msg $balloon
   }
   pack $wid -fill x	
}

proc pr_gui:cbox_edit_get {wid} { return [[$wid.entry subwidget listbox] get active] }

proc pr_gui:cbox_edit_set {wid lst} {
   set v [$wid.entry subwidget listbox]
   $v  delete 0 end
   set l [llength $lst]; set i 0
   while {$i<$l} { $v insert end [lindex $lst $i]; incr i }
   if {$l>0} { $wid.entry pick 0 }
}

proc pr_gui:cbox_edit_pick {wid item} {
   set v [$wid.entry subwidget listbox]
   set i 0; set n [$v index end]
   while {$i<$n} {
     if {[$v get $i]==$item} { break }
     incr i
   }
   if {$i==$n} { $v insert end $item }
   $wid.entry pick $i
}



#--------------------------------------------------------------------------
# Private (implementation) part 
#--------------------------------------------------------------------------

