#include "UI_imp.h"
#include "GraphManager.h"
#include "Registry.h"
#include "GUIManager.h"
#include "IVViewer.h"
#include <Inventor/nodes/SoCamera.h>


//------------  UIViewerBase  ----------------------------------------------------


float     UIViewerBase::pan_step  = 0.1f;		//class static init
const int UIViewerBase::ICON_SIZE = 24;			//

void UIViewerBase::addButton(const char*, const char*, const char*,BOOL)	//to be overriden
{  }

void UIViewerBase::pan(PAN_DIRECTION)			//to be overriden
{  }

UIViewerBase::~UIViewerBase()				//to be overriden
{  }

UIViewerBase::UIViewerBase()				//Ctor
{
  Registry::Iterator it (*GraphManager::getInstance()->getRegistry(),"viewer_activate");
  if (it.num())						//Add the Tcl-proc under the "viewer_activate" key
  {							//as activate-command for this
     const DARRAY<STRING>&  e = it.get(0);
     if (e.Count()==1 && e[0]!="")			//Discard entries with empty callback			
        addActivateCommand(e[0]);			//1st item: Tcl callback
  }
}

void UIViewerBase::addActivateCommand(const char* cmd)	//might be enhanced 
{  cmd_activate = cmd;  }

#ifdef WIN32
//MS-Windows:

#include <commctrl.h> 



//------------  SoUIExaminerViewer  -----------------------------------------------------
//REMARK/WARNING:
//
//  We could have used several methods of the original SoWinExaminerViewer (identical to the ones
//  in the SoXtExaminerViewer) to simplify and streamline the implementation of the SoUIExaminerViewer.
//  However, these methods seem not to work correctly in the Coin3D implementation. For this reason,
//  we simply re-implemented their functionality, as follows.
//
//  SoWinFullViewer::
//  addAppPushButton():		doesn't add the user-buttons starting from the top of the window.
//				We had to reimplement this and thus findAppPushButton() too.
//  SoWinComponent::
//  setWindowCloseCallback():	doesn't get called. Thus we had to override the window's-procedure
//				to explicitly trap the WM_CLOSE events.
//

WNDPROC SoUIExaminerViewer::orig_wnd_proc;		//class static
WNDPROC SoUIExaminerViewer::orig_ttip_proc;		//class static

 
SoUIExaminerViewer::~SoUIExaminerViewer()
{   }



SoUIExaminerViewer::SoUIExaminerViewer()			//Ctor 
{
  HWND        w = getShellWidget();				//Get 'toplevel' window of the viewer
  orig_wnd_proc = (WNDPROC)GetWindowLong(w,GWL_WNDPROC);	//Replace the window-proc with our own
  SetWindowLong(w,GWL_WNDPROC,(LONG)own_wnd_proc);		//and remember the original one

  initialize();							//Initialize viewer-system on demand

  //Make a tooltip-control for the tooltips in this's user-buttons
  toolTip = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP, 
		        CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
		        w,0,(HINSTANCE)GetWindowLong(w,GWL_HINSTANCE),0); 

  //Add user-buttons and their Tcl commands to this viewer, as the registry-key "viewer_cmd" says
  //
  Registry::Iterator it (*GraphManager::getInstance()->getRegistry(),"viewer_cmd");
  for(int i=0;i<it.num();i++)
  {
      const DARRAY<STRING>&  e = it.get(i);
      if (e.Count()<2) continue;				//First two entries are mandatory...
      if (e[1]=="") continue;					//Discard entries with empty callback
      const STRING& btn_bitmap = e[0];				//1st item: name of bitmap-file for button
      const STRING& btn_cb     = e[1];				//2nd item: name ot Tcl-code to execute as button-cb
      const STRING& tooltip    = (e.Count()>2)? e[2]:"";	//3rd item: tooltip (optional)
      BOOL          checkbtn   = (e.Count()>3)? (e[3]=="check") //4th item: button-type, check or not (optional)
	                                      : FALSE;
      addButton(btn_bitmap,btn_cb,tooltip,checkbtn);		//Add button and its callback
  }
}





void SoUIExaminerViewer::initialize()
{
   InitCommonControls();                                        //Initialize Windows' Common Controls system

   INITCOMMONCONTROLSEX icex;                                   // Load the tooltips class from the DLL.
   icex.dwSize = sizeof(icex);
   icex.dwICC  = ICC_BAR_CLASSES;
   InitCommonControlsEx(&icex); //!!!
}


void SoUIExaminerViewer::viewAll()                              //Enhances inh-method from SoWinViewer
{
  SoCamera* cam = getCamera();                                  //First, we reset the orientation of the camera
  cam->position.setValue(SbVec3f(0,0,1));                       //to the default one.
  cam->orientation.setValue(SbVec3f(0,0,-1),0);                 //Then we do the standard viewAll(). 
  SoWinViewer::viewAll();                                       //This is needed since the standard viewAll() does NOT
                                                                //reset the camera-orientation, which is particularly unpleasant
                                                                //when we render 2D objects which do have an orientation
}

void SoUIExaminerViewer::pan(PAN_DIRECTION dir)			//Pan view one step in given direction
{
   SoCamera* cam = getCamera();
   SbVec3f pos = cam->position.getValue();
   const SbViewVolume& vv = cam->getViewVolume();
   SbVec3f v,rot_v; float k;
   if (dir==INCR_X || dir==DECR_X)
   {	   
       k = pan_step*vv.getWidth(); if (dir==DECR_X) k *= -1;
       v = SbVec3f(k,0,0);
   }
   else //dir==INCR_Y || dir==DECR_Y
   {	   
       k = pan_step*vv.getHeight(); if (dir==DECR_Y) k *= -1;
       v = SbVec3f(0,k,0);
   }

   SbRotation r = cam->orientation.getValue();			//rot_v = r*v; This rotates the global-axis-vector v to match the
   r.multVec(v,rot_v);						//camera's axis-vector. Next, pos += rot_v will actually translate the camera's
   pos += rot_v;						//origin with 'rot_v' along one of its axes. The axis to translate along is given
   cam->position.setValue(pos);					//by the rotated 'v', the translation-length by pos_step*k, see above code.		
}


void SoUIExaminerViewer::addButton(const char* fname, const char* cb, const char* tt, BOOL check)	//Impl inh
{
    HWND      w = getAppPushButtonParent();				//Load bitmap from file
    int       n = commands.Count();					//Position new button above last one
    STRING file = fname; file += ".bmp";				//Make platfor-dependent file-name from generic one
									//For MS-Windows, this is a BMP file
    DWORD style = (check)? BS_AUTOCHECKBOX | BS_PUSHLIKE : BS_BITMAP; 
    HANDLE hbm  = LoadImage(0,(const char*)file,IMAGE_BITMAP,0,0,LR_SHARED | LR_LOADFROMFILE | LR_LOADTRANSPARENT); 
    HWND   btn  = CreateWindow("BUTTON",0,				//Create button
		   WS_VISIBLE | WS_CHILD | BS_BITMAP | style,  
		   0,n*ICON_SIZE,ICON_SIZE,ICON_SIZE,w,0,
		   (HINSTANCE) GetWindowLong(w,GWL_HINSTANCE),0);	

    orig_ttip_proc = (WNDPROC)GetWindowLong(btn,GWL_WNDPROC);		//Replace the window-proc with our own
    SetWindowLong(btn,GWL_WNDPROC,(LONG)own_ttip_proc);			//and remember the original one
    SendMessage(btn,BM_SETIMAGE,IMAGE_BITMAP,(LPARAM)hbm);		//Attach bitmap to button

    commands.Add(cb);							//Register Tcl-cmd for button
    buttons.Add(btn);							//Register HWND for button
    tooltips.Add(tt);							//Register tooltip for button	

    if (tt[0]!='\0')							//Add tooltip to this's tooltip-control
    {									//for this button if the tooltip is not an	
      TOOLINFO ti;							//empty string
      ti.cbSize   = sizeof(TOOLINFO);					//
      ti.uFlags   = TTF_IDISHWND; 
      ti.hwnd     = w; 
      ti.hinst    = (HINSTANCE) GetWindowLong(w,GWL_HINSTANCE); 
      ti.uId      = (UINT)btn; 
      ti.lpszText = (LPSTR)(const char*)tooltips[tooltips.Count()-1]; 
      SendMessage(toolTip, TTM_ADDTOOL, 0, (LPARAM)(LPTOOLINFO) &ti);   
    }
}


LRESULT CALLBACK SoUIExaminerViewer::own_wnd_proc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
//This is our 'overloaded' version of the window-procedure, to be called instead the original one
//established by Inventor/Coin3D
{
   //If we want, we can do things with the viewer now...
   SoUIExaminerViewer& this_ = *(SoUIExaminerViewer*)SoWinComponent::getComponent(hwnd);
   const char* command = 0;
   
   //Make the close-button hide (unmap), instead of destroy, the window
   if (uMsg==WM_CLOSE) { this_.hide(); return 0; }

   //Detect clicks on the user-buttons
   if (uMsg==WM_COMMAND && HIWORD(wParam)==BN_CLICKED)		//This is indeed a button-click
   {
      HWND btn  = (HWND)lParam;				
      int  bidx = this_.buttons.Contains(btn);			//Is this from a user button?
      if (bidx>=0)						//If so (Inventor tells us this),
      	 command = this_.commands[bidx];			//then execute the button's Tcl-callback
   }
   if (uMsg==WM_ACTIVATE && LOWORD(wParam)==WA_CLICKACTIVE)	//If we click-activated this window, execute the
      if (this_.cmd_activate!="") command = this_.cmd_activate; //Tcl command 'cmd_activate'
   
   IVViewer* vw = IVViewer::getViewer(&this_);			//Get IVViewer from SoUIViewer (should always work)
   if (vw && command)						//Execute 'command' with vw's name as last argument
   {
     char cmd[500];
     sprintf(cmd,"%s \"%s\"",(const char*)command,(const char*)vw->getName());	
     GUIManager::exec(cmd);					//Exec the Tcl command
     return 0;                                                  //Show that we've processed this message
   }

   //Delegate all other msgs to the original window-proc
   return CallWindowProc(orig_wnd_proc,hwnd,uMsg,wParam,lParam);
}



LRESULT CALLBACK SoUIExaminerViewer::own_ttip_proc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
//This is our own, overloaded window-proc for the user-buttons. We need this proc to handle tooltips
{
   SoUIExaminerViewer* this_ = (SoUIExaminerViewer*)SoWinComponent::getComponent(SoWin::getShellWidget(hwnd));
   
   if (this_ && (uMsg==WM_MOUSEMOVE || uMsg==WM_LBUTTONDOWN || uMsg==WM_LBUTTONUP ||
       uMsg==WM_RBUTTONDOWN || uMsg==WM_RBUTTONUP))
   {
      MSG msg; 
      msg.lParam  = lParam; 
      msg.wParam  = wParam; 
      msg.message = uMsg; 
      msg.hwnd    = hwnd; 
      SendMessage(this_->toolTip, TTM_RELAYEVENT, 0, (LPARAM)(LPMSG) &msg);      
   } 

   //Delegate all other msgs to the original window-proc
   return CallWindowProc(orig_ttip_proc,hwnd,uMsg,wParam,lParam);
}




//end MS-Windows:
#endif

