#ifndef PR_ATTR_H
#define PR_ATTR_H

//  Attr:	Models an atomic data-value in an AttrSet. An Attribute has a value and a type.
//		The type can be string, int, float, float-vector, or pointer (or unset if the Attr-obj 
//		hasn't yet been written to yet). Attrs perform an efficient memory management of their 
//		value-data for an AttrSet.
//


#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include "Genrl/types.h"
#include "FloatVecImp.h"

class Attr	{
		public:
		
		enum 	TYPE	{			//Allowed types of an attribute. Actually UNSET is not
				  INT, FLOAT,		//an allowed type, but is the default-type an Attr is made with.
			          PTR, STRING,		//It allows clients to check whether an Attr has ever been written
				  FLOAT_VEC,		//to or not (if yes, its type will be != UNSET)
				  UNSET			
				};	
			
				Attr();
				Attr(int);
				Attr(float);
				Attr(const void*);
				Attr(const char*);
				Attr(const Attr&);
				Attr(const float*,int);
		       	       ~Attr();	
		Attr&   	operator=(const Attr&);  
		int		operator==(const Attr&) const;	//Succeeds if BOTH type AND value coincide
		int		compare(const Attr&) const;	//Succeeds if values coincide after conversion to this's type
				
		void		set(int);
		void		set(float);
		void		set(const char*);
		void    	set(const void*);
		void		set(const float*,int);
		int     	getInt()	const;		//returns int-rep if type==INT/FLOAT else 0
		float   	getFloat() 	const;		//returns float-rep if type==INT/FLOAT else 0
		const char* 	getString() 	const;		//returns string-rep for all types
		const void*	getPtr()	const;		//returns ptr-rep if type==PTR else 0
		float*		getFloatVec(int&) const;	//returns float-vec if type==FLOAT_VEC else 0
		TYPE		getType()	const;		//returns data-type of this attribute (int,float,etc)
		
		private:
		
		void		dispose();
		double		data;				//data stored by this as a sort of 'union'
		TYPE		type;				//type of data in 'data'
		static char	buf[1024];			//for returning string-data from this's methods
		};




	
inline void Attr::dispose()					//disposed any allocated memory in this
{  
   if (type==STRING) delete[] *((char**)&data);
   else 
   if (type==FLOAT_VEC) delete *((FloatVecImp**)&data);
}

inline void Attr::set(int i)
{  dispose();  *((int*)&data) = i;  type = INT; }

inline void Attr::set(float i)
{  dispose();  *((float*)&data) = i;  type = FLOAT; }

inline void Attr::set(const void* i)
{  dispose();  *((const void**)&data) = i;  type = PTR; }

inline void Attr::set(const char* i)
{  
   dispose();  
   char* c = new char[strlen(i)+1];
   strcpy(c,i);
   *((const char**)&data) = c;  
   type = STRING; 
}

inline void Attr::set(const float* i,int n)			//make a new FloatVecImp object	
{
   dispose();
   *((FloatVecImp**)&data) = new FloatVecImp(i,n);
   type = FLOAT_VEC;
}

inline Attr::TYPE Attr::getType() const
{  return type;  }

inline Attr::Attr():type(UNSET),data(0)    { }		   
inline Attr::Attr(int i):type(INT)         { *((int*)&data) = i; }
inline Attr::Attr(float i):type(FLOAT)     { *((float*)&data) = i; }
inline Attr::Attr(const void* i):type(PTR) { *((const void**)&data) = i; }

inline Attr::Attr(const char* i):type(STRING) 
{ 
   char* c = new char[strlen(i)+1]; strcpy(c,i);
   *((const char**)&data) = c; 
}

inline Attr::Attr(const float* f,int sz): type(FLOAT_VEC)
{  *((const FloatVecImp**)&data) = new FloatVecImp(f,sz); }


inline Attr::Attr(const Attr& a): type(a.type),data(a.data)
{
   if (type==STRING) 
   {
      const char* s = *((const char**)&a.data);			//copy the string-data into this
      char*       c = new char[strlen(s)+1];
      strcpy(c,s); *((const char**)&data) = c;
   }
   else if (type==FLOAT_VEC)					//use FloatVecImp's copy-ctor
      *((const FloatVecImp**)&data) = new FloatVecImp(**((const FloatVecImp**)&a.data));
}  

inline Attr& Attr::operator=(const Attr& a)
{
   if (&a!=this)
   {
     dispose(); type = a.type; data = a.data;
     if (type==STRING)						//make a copy of a's string-data and put it
     {								//into this's data
       const char* s = *((const char**)&a.data);
       char*       c = new char[strlen(s)+1];
       strcpy(c,s); *((const char**)&data) = c;
     }
     else if (type==FLOAT_VEC)					//
     {
       *((FloatVecImp**)&data) = new FloatVecImp(**((const FloatVecImp**)&a.data));
     }   
   } 
   return *this; 
}     

inline int Attr::operator==(const Attr& a) const
{
   if (type!=a.type) return FALSE;
   return (type==STRING)?    !strcmp(*(const char**)&data,*(const char**)&a.data)  :
	  (type==FLOAT_VEC)? **((const FloatVecImp**)&data) == **((const FloatVecImp**)&a.data) :
   		             data==a.data;
}   		        

inline int Attr::compare(const Attr& a) const
{
   if (type==FLOAT_VEC)						//use lightweight FloatVecImp::compare()
   {
      int n; const float* f = a.getFloatVec(n);
      return (*((const FloatVecImp**)&data))->compare(f,n);
   }

   return (type==INT)?       *((int*)&data)         == a.getInt()     :
	  (type==FLOAT)?     *((float*)&data)       == a.getFloat()   :
	  (type==PTR)?       *((const void**)&data) == a.getPtr()	   :
	  (type==STRING)?    !strcmp(*((const char**)&data),a.getString()) : 
	  0;
}

inline Attr::~Attr() { dispose();  }

inline int Attr::getInt() const
{  return (type==INT)?    (int) *((int*)&data)  : 
	  (type==FLOAT)?  (int) *((float*)&data) : 
	  (type==STRING)? (int) atoi(*((const char**)&data)) : 0; 
   
   //REMARK: don't convert FLOAT_VEC to int
}	  

inline float Attr::getFloat() const
{  return (type==INT)?    (float) *((int*)&data)  : 
	  (type==FLOAT)?  (float) *((float*)&data) : 
	  (type==STRING)? (float) atof(*((const char**)&data)) : 0; 
   //REMARK: don't convert FLOAT_VEC to float
}

inline const void* Attr::getPtr() const
{  
   return (type==PTR)? *((const void**)&data) : 0 ;  
   //REMARK: don't convert anything to pointer
}

inline float* Attr::getFloatVec(int& n) const
{
   if (type!=FLOAT_VEC) return 0;

   n = (*((const FloatVecImp**)&data))->size;
   return (*((const FloatVecImp**)&data))->data;

   //REMARK: don't convert anything to FLOAT_VEC
   //	     Also, the value is returned by ref, so the caller
   //	     can directly modify it if wished. This is more efficient
   //	     that get()ting a whole vector, modifying it, and then set()ting
   //	     the modified version.
}  

#endif
