#ifndef PR_ATTR_SET_H
#define PR_ATTR_SET_H

//  AttrSet:		This class models a collection of named data attributes. 
//			The collection is managed as a set of key-value pair with unique (text) keys.
//			Besides this, an AttrSet has a 'type'. The type is an instance of AttrSetType.
//			The semantics of the set() and get() methods is as follows:
//  REMARKS:		- set(value):	always does a deep copy of value into this, in all cases.
//			- get(value):   depending on the method (getFloat,getInt,etc), tries to get 
//					the current value as a float,int, etc. This implies sometimes an
//					internal conversion, so the result is returned by value.
//					However, some get() methods, such as getFloatVec(), do NEVER attempt
//					a conversion, so the returned data is the actual value, by ref.
//					This means the caller can modify the value in-place without calling a set()
//


#include "Attr.h"
#include "Genrl/darray.h"
#include "Genrl/string.h"
#include "Genrl/types.h"


class AttrSetType;


class AttrSet	{
		public:

		virtual	     ~AttrSet();		
		void	      set(const char*,int);		  //Various set methods with various data types.
		void	      set(const char*,float);		  //All do a deep-copy of the data into this.
		void	      set(const char*,double);
		void	      set(const char*,const char*);
		void	      set(const char*,const void*);
		void	      set(const char*,const float*,int);
		void	      setType(const AttrSetType*);	  //set this's type - creates attrs for all type's attr-types
		const Attr*   get(const char*) const;
		Attr*	      get(const char*);			  //get attr by name or 0 if none found
		int	      getInt(const char*) const;	  //get typed attr-value by name or default-value of that type if not found
		float	      getFloat(const char*) const;	  
		const char*   getString(const char*) const;
		float*	      getFloatVec(const char*,int&) const;//see remarks at top of file
		void	      del(const char* =0);		  //delete given attr or all if arg==0
		int	      numAttrs() const;			  //iterator
		Attr*	      getAttrVal(int);			  //iterator	
	        const char*   getAttrName(int);		     
	        const
	        AttrSetType*  getType() const;			  //get this's type
		BOOL	      isModified() const;		  //get this's modification status
		void	      touch();				  //mark this as modified
		void	      untouch();			  //clear this's modification status
			   	
		protected:
		
			      AttrSet(const AttrSetType*);        //Ctor (with given type)
			      AttrSet();			  //Ctor (with no type). Since subclass has no preferences, we shall
								  //simply choose the 'unknown type' of the current DomainModel.
		private:
		
		struct Item   { 
		  	        STRING name; Attr val; 
				Item()				   { }
		  	        Item(const char* n)		   { name.share(n); }
				Item(const char* n,int v):         val(v),name(n) {}
				Item(const char* n,float v):       val(v),name(n) {}
				Item(const char* n,const char* v): val(v),name(n) {}
				Item(const char* n,const void* v): val(v),name(n) {}
				Item(const char* n,const float* f,int sz) : val(Attr(f,sz)),name(n) {}
			      };	
		DARRAY<Item,8>		attrs;			//The attrs in this set as (name,val) pairs
		const AttrSetType*	type;
		BOOL			modified;
		}; 	   	  	

  
  
  
inline AttrSet::AttrSet(const AttrSetType* ast): type(ast),modified(FALSE)
{   }  

inline BOOL AttrSet::isModified() const
{  return modified;  }

inline void AttrSet::untouch()
{  modified = FALSE; }

inline void AttrSet::touch()
{  modified = TRUE;  }

inline Attr* AttrSet::get(const char* name)
{
   for(int i=attrs.Count()-1;i>=0;i--)
      if (attrs[i].name==name) return &attrs[i].val;
   return 0;
}

inline const Attr* AttrSet::get(const char* name) const
{  return ((AttrSet*)this)->get(name);  }

inline void AttrSet::set(const char* name, int val)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,val)); else  a->set(val);
}                 

inline void AttrSet::set(const char* name, float val)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,val)); else  a->set(val);
}                 

inline void AttrSet::set(const char* name, double val)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,(float)val)); else  a->set((float)val);
}                 

inline void AttrSet::set(const char* name, const char* val)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,val)); else  a->set(val);
}                 

inline void AttrSet::set(const char* name, const void* val)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,val)); else  a->set(val);
}                 

inline void AttrSet::set(const char* name, const float* vec, int sz)
{
   Attr* a = get(name); if (!a) attrs.Add(Item(name,vec,sz)); else a->set(vec,sz);
}

inline int AttrSet::getInt(const char* n) const
{  const Attr* a = get(n); return (a)? a->getInt() : 0;  }

inline float AttrSet::getFloat(const char* n) const
{  const Attr* a = get(n); return (a)? a->getFloat() : 0;  }

inline const char* AttrSet::getString(const char* n) const
{  const Attr* a = get(n); return (a)? a->getString() : "";  }

inline float* AttrSet::getFloatVec(const char* n,int& sz) const
{  const Attr* a = get(n); return (a)? a->getFloatVec(sz) : 0; }

inline int AttrSet::numAttrs() const
{  return attrs.Count();  }

inline const char* AttrSet::getAttrName(int i)
{  return attrs[i].name;  }

inline Attr* AttrSet::getAttrVal(int i)
{  return &attrs[i].val;  }

inline const AttrSetType* AttrSet::getType() const
{  return type;  }

          
#endif
