#ifndef PR_COMPO_LIBRARY_H
#define PR_COMPO_LIBRARY_H


//  ComponentLibrary:   Models a collection of 'components'. A component is a software item that is related
//                      to a Tcl-proc, stored in a Tcl-file. A component is related intrinsically to the _code_
//                      of the Tcl-proc, and not only to its name. In other words, a component is a Tcl-proc 
//                      plus some extra information. An example of such a component are glyph-procedures (GlyphProcs).
//                      These are a Tcl-procedure with a given header, plus a description-text.                
//                      Component libs make it easier to manage a set of components as a whole, e.g. load,
//                      save, iterate over components, etc.
//
//  GlyphProc:          One of the component types in a ComponentLibrary. Represents a Tcl-proc name and textual
//                      description.
//
//  ComponentLibrarySet: All ComponentLibraries in the system. This class manages all loaded libraries as a whole.
//
//


#include "Genrl/string.h"
#include "Genrl/darray.h"
#include <set>
#include <stdio.h>


class   GlyphProc       //The glyph-proc component.
        {
        public:
        
                        GlyphProc(const char* proc_name,const char* descr);
                       ~GlyphProc();
        const STRING&   getTclName() const;                             //Get name of Tcl proc
        const STRING&   getDescription() const;                         //Get textual description of this proc

        private:

        STRING          tcl_name;                                       //Name of Tcl-proc for this
        STRING          descr;                                          //Textual description of proc
        };


class   ComponentLibrary  //The component-lib = set of components.
        {
        public:

                        ComponentLibrary(const char* tcl_file);         //Ctor (1st part)
                       ~ComponentLibrary();                             //Dtor
        void            loadFromTclFile();                              //Ctor (2nd part)
        void            addGlyphProc(const char* tcl_name,const char* descr);
        int             numGlyphProcs() const;                          //Iterator
        GlyphProc*      getGlyphProc(int);                              //Iterator
        const STRING&   getTclFile() const;                             
        void            save(FILE*,std::set<STRING>* =0);               //Save cmds to restore this in a Tcl-file
        void            saveDefinition(FILE*);                          //Save body of this (Tcl source)
        

        private:

        DARRAY<GlyphProc*> glyph_procs;                         //the glyph procs this lib has
        STRING          tcl_file;                               //the *.tcl file this resides in
        };

class   ComponentLibrarySet //The component-lib-set = set of component-libs.
        {
        public:

        class           GlyphProcIterator                       //Iterator ovet this's glyph-procs
                        {
                        public:
                        GlyphProc*      next();                 //Next glyph-proc of 0 if ready
                                        GlyphProcIterator(ComponentLibrarySet*);
                        private:
                        int lib; int proc; ComponentLibrarySet* libs;
                        };

                
                        ComponentLibrarySet();                  //Ctor
                       ~ComponentLibrarySet();                  //Dtor
        void            reset();                                //unload all loaded libs
        void            loadLib(const char* tcl_file);          //load lib for given Tcl file
        void            unloadLib(const char* tcl_file);        //unload lib for given Tcl file
        void            addGlyphProc(const char*,const char*);  //calls ComponentLibrary::addGlyphProc for the 'current library'
                                                                //that is undergoing loading
        GlyphProcIterator iterGlyphProcs();                     //Iterate over all glyph-procs in this's libs
        int               numLibs() const;                      //iterator over libs
        ComponentLibrary* getLib(int);                          //iterator over libs
        ComponentLibrary* getLib(const char* tcl_file);         //search a library by name
        void            save(FILE*,std::set<STRING>* =0);       //save Tcl-cmds to restore this. Also remove from arg2 the Tcl-procs
                                                                //registered by this as compos.

        private:

        void            unloadLib(int);                         //unload i-th loaded lib
        
        DARRAY<ComponentLibrary*> libs;                         //the loaded libs, owned by this
        ComponentLibrary*         current_lib;                  //the lib that undergoes loading
        };

//--------------- inline methods  -----------------------------------





inline const STRING& GlyphProc::getTclName() const
{  return tcl_name;  }

inline const STRING& GlyphProc::getDescription() const
{  return descr;  }

inline int ComponentLibrary::numGlyphProcs() const
{  return glyph_procs.Count();  }

inline GlyphProc* ComponentLibrary::getGlyphProc(int i)
{  return glyph_procs[i];  }

inline const STRING& ComponentLibrary::getTclFile() const
{  return tcl_file;  }

inline int ComponentLibrarySet::numLibs() const
{  return libs.Count();  }

inline ComponentLibrary* ComponentLibrarySet::getLib(int i)
{  return libs[i];  }

#endif

