#ifndef PR_EVENT_HANDLER_H
#define PR_EVENT_HANDLER_H

//	EventHandler:		Interface for IVViewer's event handlers. These are objects that the IVViewer
//				uses to process the mouse events it receives. 
//				The handler receives the normalized screen (x,y) coords of the event.
//				Next, the handler can determine which Node/Edge is closest to the event's 
//				(x,y) position, using IVViewer's pickNode/pickEdge().
//				The EventHandler allows the user to add an user-callback+data.
//				Subclass implementations of event() may then call this user-cb.
//
//	DefaultEventHandler:	Default implem of an EventHandler that manages the viewers' highlight-selection
//				by adding/removing nodes/edges clicked by the user in the viewer. This also edits the
//				nodes' positions by dragging them interactively.
//


#include "Genrl/types.h"
#include "Genrl/string.h"
class	IVViewer;
class	Node;
class	Edge;
class	Selection;



class	EventHandler 					
	{						
	public:						
	
	typedef STRING (*Callback)    (EventHandler*,void*);
        typedef void   (*VCallback)   (EventHandler*,void*);
	
	
	enum EVENT_TYPE		 {						//Event type: so far, only mouse events
				   MOVE, UP_1,   DOWN_1,			
				   UP_2, DOWN_2, KEY,    OTHER
				 };
	enum MODIFIER_TYPE	 {						//Modifier type (shift,ctrl,alt keys)
				   NONE  = 0,					//These are OR-ed into the event() method
				   SHIFT = 1,
				   CTRL  = 2,
				   ALT   = 4
				 };
	
				 EventHandler(Callback=0,void* =0);	        //Ctor
	void			 setUserCallback(Callback,void* =0);	        //Specify a user-cb + user-data if desired
	void			 setDeleteCallback(VCallback);		        //Specify an optional cleanup procedure to be called 
										//whenever this releases/modifies its user-data.
	void			 getUserCallback(Callback&,void*&);	        //Get user cb data
	void			 setSweepCallback(Callback,void* =0);	        //Same as setUserCallback but for sweeping
	void			 setSweepDeleteCallback(VCallback);	       	//Same as setDeleteCallback but for sweeping
	void			 getSweepCallback(Callback&,void*&);	        //Get sweep cb data
	void			 enableSweepCallback(BOOL);			//Enable/disable execution of sweep cb
	BOOL			 sweepCallbackEnabled() const;			//Get state of sweep cb (enabled/disabled)
	
	void			 setAllowMove(BOOL);				//Set/get whether this allows moving nodes/edges in the viewer
	BOOL			 getAllowMove() const;
	virtual			~EventHandler();
	virtual void		 event(IVViewer*,EVENT_TYPE,MODIFIER_TYPE,	//Called back when an event happens, with:
				       float,float,int key) =0;	               	// arg1   = viewer in which event happened 
										// arg2   = event-type,
				       						// arg3+4 = 2D norm-screen-coords of event
                                                                                // arg5   = keycode (SoKeyboardEvent) of released key, for key-events
	virtual Node*		getPickedNode();				//For cb-code: ask which was the last picked node. Def=nop
	virtual Edge*		getPickedEdge();				//For cb-code: ask which was the last picked edge. Def=nop
	virtual Selection*	getPickedSelection();

	protected:

	Callback		 user_cb,sweep_cb;				//cb to call after handling event, if desired
	VCallback		 del_cb,del_sweep_cb;				//cb to call whenever user_data gets written to
	void*			 user_data,*sweep_data;				//data to pass back to user-cb
	BOOL			 allow_move;					//allows or not moving objects in the viewer
	BOOL			 sweep_cb_on;					//if TRUE, sweep cb is called, else not
	};




class	DefaultEventHandler : public EventHandler
	{
	public:
	
				DefaultEventHandler(Callback=0, void* =0);	
	void			event(IVViewer*,EVENT_TYPE,MODIFIER_TYPE,	//Impl inh: edits viewer's highlight-selection
				      float,float,int);	        		//and also edits nodes' positions used by viewer

	Node*			getPickedNode();				//Impl inh
	Edge*			getPickedEdge();				//Impl inh
	Selection*		getPickedSelection();				//Impl inh
					    						
	protected:

        void                    brushText(const STRING&,IVViewer*,float,float); //shows a dynamic text 'brush' at mouse-pos
        void                    saveGLState();					//for rubberbanding
	void			restoreGLState();				//for rubberbanding
	void			drawRubberband(float,float);			//for rubberbanding
		
	Node* 			picked_node,*return_node;			//for picking
	Edge* 			picked_edge,*return_edge;			//for picking
	Selection*		picked_selection;				//for picking
	BOOL  			moved,large_move,down_clicked,rubber;		
	float			nx_pos,ny_pos;					
	float			nx_orig,ny_orig;
	float			nx_old,ny_old;	
	int 			save_shade,save_lighting,save_linew,save_depth,	//for saving/restoring
    				save_blend,save_equ,save_logic;			//IV's GL state
	float			save_col[4];
	};



inline void EventHandler::setUserCallback(Callback cb,void* cd)
{   
    user_cb   = cb;					
    if (user_data && del_cb) del_cb(this,user_data);	//if we had a delete callback and some
    user_data = cd;					//user data, ask the delete callback to delete it
}

inline void EventHandler::setDeleteCallback(VCallback cb)
{   del_cb = cb;  }    

inline void EventHandler::getUserCallback(Callback& cb,void*& cd)
{   cb = user_cb; cd = user_data; }

inline void EventHandler::setSweepCallback(Callback cb,void* cd)
{   
    sweep_cb   = cb;					
    if (sweep_data && del_sweep_cb) del_sweep_cb(this,sweep_data);	//if we had a delete callback and some
    sweep_data = cd;							//user data, ask the delete callback to delete it
}


inline void EventHandler::setSweepDeleteCallback(VCallback cb)
{   del_sweep_cb = cb;  }    

inline void EventHandler::getSweepCallback(Callback& cb,void*& cd)
{   cb = sweep_cb; cd = sweep_data; }

inline void EventHandler::enableSweepCallback(BOOL b)
{   sweep_cb_on = b;  }

inline BOOL EventHandler::sweepCallbackEnabled() const
{   return sweep_cb_on;  }



#endif
