#ifndef PR_FILTER_H
#define PR_FILTER_H

//	Filter:		Baseclass for all edge-filter classes. An edge-filter is a small tool that, given
//			a Node, will 'filter' its Edges and extract a subset of them. Edge-filters are mostly
//			used as part of more complex engines, such as Traversals, in order to determine e.g.
//	                how to branch during the traversal of a graph, or in filters, to select a specific subset
//			of a graph.
//			Concrete subclasses simply implement the filter() method in some specific way.
//
//

#include "Genrl/string.h"

class 	Filter							//Baseclass for all Filters
	{
	public:
	
	virtual	       ~Filter();
	virtual EdgeSet filter(Selection*,Node*) =0;		//The job this does...
	virtual Filter* clone()			 =0;		//Clones this

	protected:
			Filter();
	};
	
				
class	DirectionFilter : public Filter				//Filters in or outgoing edges
	{
	public:
			DirectionFilter(Node::DIRECTION);
			DirectionFilter(const DirectionFilter&);
	Filter*		clone();				//Impl inh
	EdgeSet		filter(Selection*,Node*);
        void		setDirection(Node::DIRECTION d) { direction=d; }
	Node::DIRECTION getDirection() const	        { return direction; }

	
	protected:
	
	Node::DIRECTION	direction;				//Filter's direction
	};
	




class	AttrFilter : public DirectionFilter			//Filters edges based on attribute name, value, and edge direction
	{
	public:
			AttrFilter(const char* name,const Attr& value,Node::DIRECTION);
			AttrFilter(const AttrFilter&);
	Filter*		clone();				//Impl inh
	EdgeSet		filter(Selection*,Node*);
	
	protected:
	
	STRING		attr;
	Attr		value;
	};
	



class	TypeFilter : public DirectionFilter			//Filters edges based on type name and edge direction
	{
	public:
			TypeFilter(const char* type_name,Node::DIRECTION);
			TypeFilter(const TypeFilter&);
	Filter*		clone();				//Impl inh
	EdgeSet		filter(Selection*,Node*);
	
	protected:
	
	STRING		type;
	};
	

class	NodeFilter : public DirectionFilter			//Filter selecting all edges between two given nodes
	{							//in the ctx of a given Selection
	public:
			NodeFilter(const Node*,Node::DIRECTION);
			NodeFilter(const NodeFilter&);
	Filter*		clone();
	EdgeSet		filter(Selection*,Node*);		//Impl inh

	protected:

	const Node*	node;
	};


//------------  Inline implementations  -------------------------------------


inline Filter::Filter()
{   }

inline DirectionFilter::DirectionFilter(Node::DIRECTION d): direction(d)
{   }

inline DirectionFilter::DirectionFilter(const DirectionFilter& a): direction(a.direction)
{   }

inline AttrFilter::AttrFilter(const char* a,const Attr& v,Node::DIRECTION d)
		  : DirectionFilter(d), attr(a),value(v)
{   }

inline AttrFilter::AttrFilter(const AttrFilter& a)
                  :DirectionFilter(a),attr(a.attr),value(a.value)
{  }

inline TypeFilter::TypeFilter(const char* t,Node::DIRECTION d)
		  : DirectionFilter(d), type(t)
{   }

inline TypeFilter::TypeFilter(const TypeFilter& f)
		  : DirectionFilter(f), type(f.type)
{   }

inline NodeFilter::NodeFilter(const NodeFilter& f)
		  : DirectionFilter(f), node(f.node)
{   }

inline NodeFilter::NodeFilter(const Node* n,Node::DIRECTION d)
		 : DirectionFilter(d), node(n)
{   }



#endif


