#ifndef GENRL_STRING_H
#define GENRL_STRING_H

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>

// STRING:		General-purpose string class.
//			Encapsulates a char*, which is either owned
//			by this class (in which case a deep-copy is
//			kept) or by some other external owner, 
//			in which case the char* is shared.
//			All mutator operations automatically unshare the data if necessary
// WARNING:		Operator char* is provided for convenience but should be used carefully.
//			Incorrect usage would be for example to assign the result of a STRING f()
//			to a char*, since the STRING obj is temporary and therefore destroyed.

class STRING	{	
		public:
		
		
			STRING(const char* =0);			//Ctor. Arg is deep-copied & owned
			STRING(int len);			//Ctor. Alloc a blank string len chars long
			STRING(const STRING&);			//Copy-ctor. Arg is deep-copied.
		       ~STRING();	
			STRING& operator=(const STRING&);	//Assignment. Arg is deep-copied.
			STRING& operator=(const char*);		//Assignment. Arg is deep-copied.
			STRING& operator=(int);			//Assignment. Converts int to string rep.
			STRING& operator=(double);		//Assignment. Converts dbl to string rep.
			STRING& operator=(float);		//Assignment. Converts flt to string rep.
			int operator==(const STRING&) const;	//Comparison op (deep)
			int operator==(const char*) const;	//Comparison op (deep)
			int operator!=(const STRING&) const;	//Comparison op (deep)
			int operator!=(const char*) const;	//Comparison op (deep)
			int operator<(const STRING&) const;	//Comparison op (deep)
			int operator<(const char*) const;	//Comparison op (deep)
			STRING& share(const STRING&);		//Shallow assignment (sharing char*)
			STRING& share(const char*);		//Shallow assignment (sharing char*)
			STRING& operator+=(const char*);	//Append string
			STRING& operator+=(const STRING&);	//Append string
		        STRING& operator+=(int);      		//Append int
		        STRING& operator+=(float);      	//Append float
		        STRING  operator+(const STRING&) const; //Concat strings
		        STRING  operator+(const char*) const;	//Concat strings
                        char&   operator[](int);                //Indexing
		        int length() const;			//Return string length;
		        int pos(const STRING&) const;		//Returns pos of 1st occurence of arg in this or -1
			int pos(const char*) const;		//- " -		        					

			//--------------Conversion methods--------------
			//operator char*();			//Conversion to char*
			operator const char*() const;		//Conversion to char*
			operator const char*();
			int	 toInt() const;			//Converts this to int (assumes it is ok). On error, returns 0
			float	 toFloat() const;		//Converts this to flt (assumes it is ok). On error, returns 0.0
			double	 toDouble() const;		//Converts this to dbl (assumes it is ok). On error, returns 0.0
			
		//protected:
		
			char* c;
			int   own;
			void  unshare();			//if this doesn't own c, makes an own
								//copy of it for private use.
			char*  alloc(int);			//alloc & return 'int' chars
			char*  copy(const char*);
		};
		
inline char* STRING::alloc(int i)
{
   own = 1;			//always we alloc sth that'll go into our 'c'
   return  new char[i];
}

inline char* STRING::copy(const char* ch) 
{
   char* r = alloc(strlen(ch)+1);	//copy() will be always used to copy sth to be an 'own' char*
   strcpy(r,ch);
   return r;
}

inline STRING::~STRING()
{  if (own) delete[] c;   }

inline STRING::STRING(const STRING& s)
{
   if (!s.own)
   {
      own = 0;
      c = s.c;
   }   
   else c = copy(s.c);
}

inline STRING::STRING(int len)
{  c = alloc(len);  }				//Simply make a buffer of length len


inline STRING::STRING(const char* ch)
{
   if (ch && ch[0]!='\0')  c = copy(ch);	//Copy the arg if not trivial
   else	   { c = ""; own = 0; }
}       
         				
inline STRING& STRING::operator=(const char* ch)
{
   if (own) delete[] c; 
   
   if (ch && ch[0]!='\0') c = copy(ch);		//Free current stuff, then copy arg, if not trivial
   else	   { c = ""; own = 0; }
   
   return *this;
}  

inline STRING& STRING::operator=(int i)
{
   if (own) delete[] c;
   c    = alloc(30);		//should be enough for any number
   sprintf(c,"%d",i);

   return *this;
} 

inline STRING& STRING::operator=(double  d)
{
   if (own) delete[] c; 
   c    = alloc(30);		//should be enough for any number
   sprintf(c,"%lf",d);
   
   return *this;
}  

inline STRING& STRING::operator=(float  f)
{
   if (own) delete[] c;
   c    = alloc(30);		//should be enough for any number
   
   sprintf(c,"%g",f);
   return *this;
}   


inline char& STRING::operator[](int i)
{  return c[i];  }   


inline STRING& STRING::operator=(const STRING& s)
{
   if (&s == this) return *this;
   return operator=(s.c);
}   

inline void STRING::unshare()
{
   if (own) return;		//nothing to unshare, we own c
   
   if (c[0]!='\0') c = copy(c);	//make a private copy of c, if not trivial. 
}				//That is, we never really unshare "" strings

inline STRING& STRING::operator+=(const char* ch) 
{
   if (!ch || ch[0]=='\0') return *this;
   
   int own_ = own;
   char*  s = alloc(strlen(c)+strlen(ch)+1);
   strcpy(s,c);
   strcat(s,ch);
   if (own_) delete[] c;
   c   = s;

   return *this;
} 

inline STRING STRING::operator+(const char* ch) const
{
   STRING ret = c; ret += ch;
   return ret;
}

inline STRING STRING::operator+(const STRING& s) const
{
   STRING ret = c; ret += s.c;
   return ret;
} 

inline STRING& STRING::operator+=(int i)
{
   char buf[100];
   sprintf(buf,"%d",i);
   return operator+=(buf);
}

inline STRING& STRING::operator+=(float f)
{
   char buf[100];
   sprintf(buf,"%g",f);
   return operator+=(buf);
}      
 
inline STRING& STRING::operator+=(const STRING& s)
{  return operator+=(s.c);  }
  

inline STRING& STRING::share(const char* ch)
{
   if (own) delete[] c;			//first get rid of our own c, if we owned it
   
   own = 0;				//remember that from now on we won't own c
   c   = (char*)ch;			//shallow-copy of ch into c
   
   return *this;
}

inline STRING& STRING::share(const STRING& s)
{  return share(s.c);  }

//inline STRING::operator char*()
//{  return c;  }

inline STRING::operator const char*()
{  return c;  }

inline STRING::operator const char*() const
{  return c;  }

inline int STRING::toInt() const
{
   int ret = 0;
   sscanf(c,"%d",&ret);
   return ret;
}

inline float STRING::toFloat() const
{
   float ret = 0.0;
   sscanf(c,"%f",&ret);
   return ret;
}

inline double  STRING::toDouble() const
{
   double  ret = 0.0;
   sscanf(c,"%lf",&ret);
   return ret;
}

inline int STRING::operator==(const char* ch) const
{
   if (!ch)
     if (c[0]=='\0') return 1;			//NULL == ""
     else	     return 0;			
   
   if (ch[0]=='\0' && c[0]=='\0') return 1;	//"" == ""
   
   return !strcmp(c,ch);			//full-blown strcmp()
}

inline int STRING::operator<(const char* ch) const
{
   if (!ch)
      if (c[0]=='\0') return 1;			//NULL == "";
      else            return 0;			//nothing < ""
   
   if (ch[0]=='\0' && c[0]=='\0') return 0;     //"" == ""
   if (c[0]=='\0') return 1;			//"" < anything
   
   return strcmp(c,ch)<0;			//full-blown strcmp()
}

inline int STRING::operator<(const STRING& s) const
{  return operator<(s.c);  }


inline int STRING::operator==(const STRING& s) const
{  return operator==(s.c);  } 

inline int STRING::operator!=(const STRING& s) const
{  return !operator==(s.c);  }  

inline int STRING::operator!=(const char* ch) const
{  return !operator==(ch);  }         

inline int STRING::length() const
{  return (c[0]=='\0')? 0: strlen(c);  }

inline int STRING::pos(const char* s) const
{
   if (!c || c[0]=='\0' || !s) return -1;
   
   char* p = strstr(c,s);
   return (p) ? p-c : -1;
}

inline int STRING::pos(const STRING& s) const
{  return pos(s.c);  }


#endif
