#ifndef PR_GEOMETRY_H
#define PR_GEOMETRY_H

//	Geometry:	Repository of public services pertaining to geometric operations
//			on the geometric (postion/size) attributes of nodes, given as a Selection.
//	REMARK:		It is the caller's responsibility to ensure that the selected nodes have
//			the x,y(,z) attrs. If they don't, the ops will probably read zero values,
//			respectively create new pos and/or size attributes (see class AttrSet...).
//
//

#include "Genrl/string.h"


class   Selection;

class	Geometry
	{
	public:
	
        enum    OVERLAP_TYPE                     //Return codes for various overlap functions, as follows:       
                        {                        //(see specific methods for exact semantics)                         
                           NONE =0,              // no overlap whatsoever
                           EDGE_INTERSECT =1,    // edges (of primitives tested) intersect but no full containment
                           CONTAIN_FIRST =2,     // edges don't intersect but primitive 1 contains primitive 2
                           CONTAIN_SECOND =3     // edges don't intersect but primitive 2 contains primitive 1
                        };
                                   
	
	static void	bbox2D(Selection*,const char* pos,float& xm,float& ym,float& xM,float& yM);
				//Compute bbox of nodes
	static OVERLAP_TYPE     
                        rectOverlap(float x1, float y1, float x2, float y2,
				    float x3, float y3, float x4, float y4);
				//Tests if two rectangles (x1,y1-x2,y2) and (x3,y3-x4,y4) overlap or not.
                                //Returns exact type of overlapping (see OVERLAP_TYPE)
        static BOOL     rectContains(float x1, float y1, float x2, float y2, float xp, float yp);
                                //Tests if point (xp,yp) is in rectangle (x1,y1-x2,y2)
        static void     rectLineClip(float  x1, float  y1, float  x2, float y2,
                                     float& x3, float& y3, float& x4, float& y4);
                                //Computes intersection-point between rectangle (x1,y1-x2,y2) and line (x3,y3-x4,y4)
        static int*     spline(float* Px,float* Py,float* Pz,float t,int K,int& N);
                                //Generates N spline-points out of ctrl-points Px..Pz, spline-params t,K.
				//Caller gets the spline-points as Px|y|z[i[0]]..Px|y|z[i[N-1]] where i[] is the return-value.
				//i[] is owned by this. Caller should ensure that Px..Pz have splineSize(K,N) free positions.
        static int      splineSize(int K,int N);
                                //Estimates how many points spline() would generate for given K and N
	
	static STRING	getPositionX(const char* pos);
	static STRING	getPositionY(const char* pos);
	static STRING	getPositionZ(const char* pos);
	static STRING	getDimensionX(const char* pos);
	static STRING	getDimensionY(const char* pos);
	static STRING	getDimensionZ(const char* pos);
	static STRING	getPinned(const char* pos);

	private:

	static BOOL	rectLineIntersect(float x1, float y1, float x2, float y2,
				          float x3, float y3, float x4, float y4,
                                          unsigned char* =0,unsigned char* =0);
				//Tests if rectangle (x1,y1-x2,y2) intersects line-segment (x3,y3-x4,y4)

   	};



inline STRING Geometry::getPositionX(const char* pos)
{  STRING s = pos; s += "_x"; return s; }

inline STRING Geometry::getPositionY(const char* pos)
{  STRING s = pos; s += "_y"; return s; }
   
inline STRING Geometry::getPositionZ(const char* pos)
{  STRING s = pos; s += "_z"; return s; }

inline STRING Geometry::getPinned(const char* pos)
{  STRING s = pos; s += "_pin"; return s; }

inline STRING Geometry::getDimensionX(const char* pos)
{  STRING s = pos; s += "_dx"; return s; }

inline STRING Geometry::getDimensionY(const char* pos)
{  STRING s = pos; s += "_dy"; return s; }

inline STRING Geometry::getDimensionZ(const char* pos)
{  STRING s = pos; s += "_dz"; return s; }

inline int Geometry::splineSize(int K,int N)
{  return (1<<K) *(N-2)+2; }


inline BOOL Geometry::rectLineIntersect(float x1, float y1, float x2, float y2, 
                                        float x3, float y3, float x4, float y4,
                                        unsigned char* code1_ret, unsigned char* code2_ret)
//Tests intersection rectangle-segment using Cohen-Sutherland algorithm.
//We augment the test to capture also the case when the line is completely inside the rectangle,
//for which we return TRUE as well. Remember this works since we test ONLY axis-aligned rectangles.
{
   unsigned char code1 = 0, code2 = 0;
   if (x3<x1) code1 |=8; if (y3>y2) code1 |=4;
   if (x3>x2) code1 |=2; if (y3<y1) code1 |=1;
   if (x4<x1) code2 |=8; if (y4>y2) code2 |=4;
   if (x4>x2) code2 |=2; if (y4<y1) code2 |=1;
   if (code1_ret) *code1_ret = code1;
   if (code2_ret) *code2_ret = code2;
   return (((code1|code2)==0) || ((code1&code2)==0));
}



inline Geometry::OVERLAP_TYPE Geometry::rectOverlap(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4)
//Tests overlapping of 2 rectangles by testing if edges of 2nd intersect the 1st rectangle
//If this test fails, there is just one possibility for overlap, that is that the 1st rectangle is completely
//inside the 2nd, or conversely. We test this separately. 
//If this fails too, it means that the two rectangles don't overlap at all.
{
   if (rectContains(x3,y3,x4,y4, x1,y1) && rectContains(x3,y3,x4,y4, x2,y2))
      return CONTAIN_SECOND;                                                //rect 2 contains rect 1
   if (rectContains(x1,y1,x2,y2, x3,y3) && rectContains(x1,y1,x2,y2, x4,y4))
      return CONTAIN_FIRST;                                                 //rect 1 contains rect 2
 
   if (rectLineIntersect(x1,y1,x2,y2, x3,y3,x3,y4)) return EDGE_INTERSECT;
   if (rectLineIntersect(x1,y1,x2,y2, x3,y3,x4,y3)) return EDGE_INTERSECT;
   if (rectLineIntersect(x1,y1,x2,y2, x4,y3,x4,y4)) return EDGE_INTERSECT;
   if (rectLineIntersect(x1,y1,x2,y2, x3,y4,x4,y4)) return EDGE_INTERSECT;
   return NONE;                                                             //no overlap whatsoever
}

inline BOOL Geometry::rectContains(float x1, float y1, float x2, float y2, float xp, float yp)
{
   return (xp>x1 && xp<x2 && yp>y1 && yp<y2);
}


#endif

