#ifndef PR_GLYPH_FACTORY_LIST_H
#define PR_GLYPH_FACTORY_LIST_H

// IVGlyphFactoryList:	Maintains a list of all persistent IVGlyphFactory objects in the system.
//			This class is managed by GraphManager. 
//			IVGlyphFactories (with names) can be added to the list. Glyph factories are managed
//			(created,retrieved,destroyed) only via this class. This is similar to SelectionList for Selections.
//			IVGlyphFactoryList automatically creates a few IVGlyphFactories (the 'default' factories) which
//			are guaranteed to exist and not be deleted/replaced/renamed in this.	
//

#include "Genrl/darray.h"
#include "Genrl/types.h"
#include "Genrl/string.h"
#include <set>
#include <stdio.h>

class IVGlyphFactory;
class IVNodeGlyphFactory;
class IVEdgeGlyphFactory;

class IVGlyphFactoryList	 
		  {
		  public:
		  
		  	 	IVGlyphFactoryList();			       //CONSTR/DESTR 	
		  	       ~IVGlyphFactoryList();			       //
		  void		reset();				       //Delete all stuff in this except default factories
                  void          save(FILE*,std::set<STRING>* =0);              //Save Tcl-cmds to restore this in a file
                                                                               //Also return all Tcl-proc-names this depends on in arg2, if given.
					
									       //QUERY METHODS
		  int		numNGFactories() const;			       //get # node/edge factories in list
		  int		numEGFactories() const;			       //
		  IVNodeGlyphFactory*	 
				getNGFactory(int);			       //iterate over node/edge factories
		  IVNodeGlyphFactory* 
				getDefNGFactory();
		  IVEdgeGlyphFactory*
				getEGFactory(int);
		  IVEdgeGlyphFactory*
				getDefEGFactory();
		  IVNodeGlyphFactory*     
				getNGFactory(const char*);		       //get node/edge factory by name or 0 if not found
		  IVEdgeGlyphFactory*
				getEGFactory(const char*);
		  							       //MANAGEMENT METHODS
		  void		addFactory(IVNodeGlyphFactory*);	       //add factories to list. Destroys any factories with same name
		  							       // (if the arg is physically already in the list, nothing happens)
		  void		addFactory(IVEdgeGlyphFactory*);	       //	
		  BOOL		delNGFactory(const char*);		       //destroys factory by name. Returns 1 if op completed ok.
		  BOOL		delEGFactory(const char*);		       //
		  BOOL		renameNGFactory(const char*,const char*);      //renames factory. Deletes any factory with new name.
		  BOOL		renameEGFactory(const char*,const char*); 
		  
		  	
		  private:	
		  
		  void				init();				//
		  DARRAY<IVNodeGlyphFactory*>	ngf;
		  DARRAY<IVEdgeGlyphFactory*>	egf;
		  IVNodeGlyphFactory*		def_ngf;			//default IVNodeGlyphFactory, guaranteed to exist always
		  IVEdgeGlyphFactory*		def_egf;			//default IVEdgeGlyphFactory, guaranteed to exist always	
                  int                           num_def_ngfs;                   //# IVNodeGlyphFactories created automatically by this's init()
                                                                                //These come always first in ngf[].
                  int                           num_def_egfs;                   //# IVEdgeGlyphFactories created automatically by this's init()
                                                                                //These come always first in egf[].
		  };	 		


inline int IVGlyphFactoryList::numNGFactories() const
{  return ngf.Count();  }

inline int IVGlyphFactoryList::numEGFactories() const
{  return egf.Count();  }

inline IVNodeGlyphFactory* IVGlyphFactoryList::getNGFactory(int i) 
{  return ngf[i];  }

inline IVEdgeGlyphFactory* IVGlyphFactoryList::getEGFactory(int i) 
{  return egf[i];  }

inline  IVNodeGlyphFactory* IVGlyphFactoryList::getDefNGFactory()
{   return def_ngf;  }

inline  IVEdgeGlyphFactory* IVGlyphFactoryList::getDefEGFactory()
{   return def_egf;  }

#endif
