#ifndef PR_IV_MAPPER_H
#define PR_IV_MAPPER_H

//	IVMapperBase:	Maps a Selection to an Inventor scenegraph that can be further viewed in an IVViewer. 
//	IVMapper:	Specific subclass implementing the graph-to-Inventor-data mapping by using two separate
//                      helper-classes IVNodeMapper and IVEdgeMapper. These can be set separately to he IVMapper
//                      in order to independently customize the way nodes and edges are drawn.
//                      Moreover, an IVMapper allows a set of 'auxiliary node/edgemappers' to be added.
//                      These are identical to the IVNode|EdgeMapper mentioned above, with the difference that
//                      they don't allow user input (picking). These are generally used to redundantly display
//                      nodes/edges in different ways in the same IVMapper.
//
//			A mapper should be able to answer which is the Node/Edge glyph closest to a given 3D point
//			and/or containing stuff on a given SoPath. This is implemented via the pickNode/pickEdge() methods.
//			If the mapper can't respond to such requests (default), these methods should return 0.
//			
//			Mappers may be able to display a list of visually highlighted nodes/edges, being
//			a subset of the node/edge list given to them to display.
//			Mappers are used by viewers. I'm not sure if it's safe for a mapper to be used by several viewers simultaneously.
//			Mappers should respond to the delNode/delEdge/moveNode/moveHighlight events, by updating themselves
//			accordingly. Only the first 2 events are mandatory to be supported - the 3rd is optional.
//
//      IVNodeMapper:   Helper classes that actually implement the building of Inventor node and edge reps.
//      IVEdgeMapper:   They are used by an IVMapper to perform its displaying of nodes/edges. Specific subclasses
//                      of IVNode|EdgeMapper implement specific ways to draw nodes and edges (see IVMapperImp).
//



#include "Genrl/string.h"
#include "Genrl/types.h"
#include "Genrl/refcount.h"
#include "Genrl/darray.h"
#include <map>

class	Node;
class	Edge;
class   Selection;
class	SoSeparator;
class   SoTransform;
class	SoNode;
class	SoGroup;
class   SoPickedPoint;
class	SoVertexProperty;
class   SoLineSet;
class   IVNodeGlyphFactory;
class   IVEdgeGlyphFactory;
class	IVNodeMapper;
class	IVEdgeMapper;


class	IVMapperBase : public RefCount				//Very generic interface for a mapper - assumes nothing
	{							//about _how_ the mapping is to be done.
	public:
	
	enum		SELECT_TYPE				//Operation to be done for select/deselect()
			{
			   SELECT,
			   DESELECT,
			   TOGGLE,
			   DEFAULT
			};   	
	enum		MOVE_STYLE				//Style in which the mapper does the movement of the nodes/edges.
			{					//Subclasses can implement different move-styles - the low-res style
			   LOW_RES  =0,				//should be the fastest (most interactive) whereas the high-res one	
			   MID_RES  =1,				//should be the one graphically the closest to the still image. 
			   HIGH_RES =2				//Mapper clients can thus choose a tradeoff between interactivity & detail.
			}; 
	enum	        EDGE_RESOLUTION    
			{					//Hint for how to map an edge
                           NODE_CENTERS =0,			// - draw a line between node centers (don't use edge-coord-attrs)
                           STRAIGHT     =1,			// - draw a line between 1st and last edge-coord-attrs
                           DEFAULT_RES  =2,			// - use whatever rep the edge has (e.g. polyline)
                           SMOOTH       =3			// - build a smooth rep (e.g. spline)
                        };					//REMARK: these are hints, actual IVEdgeMappers can interpret them as desired
 
			IVMapperBase();
	virtual void	setNodeScale(float);			 //set the scaling of the nodes. 1 means the size of the whole graph.
								 //should be enhanced by subclasses.
	float		getNodeScale() const;			 //get the scaling of the nodes
	virtual void	setHighlightColor(float,float,float);	 //set/get color used for highlighting stuff in this
	void		getHighlightColor(float&,float&,float&);
	virtual void    setEdgeResolution(EDGE_RESOLUTION);	 //set/get hint on how to draw edges
	EDGE_RESOLUTION getEdgeResolution() const;
	void		setEdgeSmoothingParams(int,float);	 //set/get params used for edge smoothing (see IVEdgeMapper)	
	void		getEdgeSmoothingParams(int&,float&) const;

	//----------   Methods below used only internally by the toolkit   --------------------------------------------------

	virtual void	delNode(Node*) 	=0;			//inform this that a Node is deleted
	virtual void	delEdge(Edge*)	=0;			//inform this that an Edge is deleted
	virtual void	moveNode(Node*,float,float);		//move a Node with dx,dy (def=nop)
	virtual void	moveHighlight(float,float);		//move all highlight-selection with (dx,dy) (def=nop)
	virtual void	setMoveStyle(MOVE_STYLE);		//set the move-style (see styles) (def=set style-member only)
	MOVE_STYLE	getMoveStyle() const;			//get current move-style
	virtual SoSeparator*					//map a Selection and a highlight-Selection
			map(const char*,const char*,		//(with optional position-attr-name) to an 
			    const char* =0) =0;			//SoSeparator which is owned by this)
	void		touch();				//marks this as being modified and needing a full update (see IVViewer)
	BOOL		isModified() const;			//tells if this needs an update of any kind or not
	virtual Node*	pickNode(SoPickedPoint*,float,float,float); //returns the Node whose glyph is closest to world-coord
								//(arg2,arg3,arg4) and has arg1 inside. 
								//Used by IVViewer::pickNode(). Default = ret 0
	virtual Edge*	pickEdge(SoPickedPoint*,float,float,float); //returns the Edge whose glyph is closest to world-coord
								//(arg2,arg3,arg4) and has arg1 inside.			
								//Used by IVViewer::pickEdge(). Default = ret 0		
	const char*	getSelection() const;			//Get name of input-sel
	const char*	getHighlight() const;			//Get name of highlight-sel
	const char*     getPosAttrX() const;			//Get names of pos-attrs that this uses
	const char*	getPosAttrY() const;			//
	const char*	getPosAttrZ() const;			//
        const char*     getPosAttr()  const;
	const char*     getSizeAttrX() const;			//Get names of size-attrs that this uses
	const char*	getSizeAttrY() const;			//
	const char*	getSizeAttrZ() const;			//
								
	virtual SELECT_TYPE					//Highlights/unhighlights Node that this maps and
			highlight(Node*,SELECT_TYPE=SELECT);	//returns crt-sel-state. def=nop						
	virtual SELECT_TYPE					//Highlights/unhighlights Edge that this maps and
			highlight(Edge*,SELECT_TYPE=SELECT);	//returns crt-sel-state. def=nop		
	virtual	void	clearHighlight();
	
								
	protected:
	
		       ~IVMapperBase();	 			//Dtor	
	void		setPosAttr(const char*);		//Tells which attr-prefix of the Selection we should use for the
								//coordinates
	STRING		pos,pos_x,pos_y,pos_z;			//names of pos-prefix and pos-attrs this uses
	STRING		sz_x,sz_y,sz_z;				//names of size-attrs this uses (optional)
	float		scale;					//scale for node-representations
	STRING          input_sel,highlight_sel;
	BOOL		is_modified;				//retains this's need to be updated
	MOVE_STYLE	move_style;				//move style for interactive node dragging
	float		highlight_r,highlight_g,highlight_b;	//color used for highlighting
	EDGE_RESOLUTION edge_res;				//resolution for edge drawing (see IVEdgeMapper)
	int		edge_smoothing_subdiv;			//spline-subdiv used for smoothing edges (see IVEdgeMapper)
	float		edge_smoothing_tightness;		//spline-tightness used for smoothing edges (see IVEdgeMapper)
	};


class	IVMapper : public IVMapperBase				//Mapper using two separate IVNodeMapper and IVEdgeMapper for nodes/edges
	{
	public:
	
				IVMapper();
		               ~IVMapper();  		
	SoSeparator*		map(const char*,const char*,const char* =0);//impl inh
	void			delNode(Node*);			     	//impl inh	
	void			delEdge(Edge*);			     	//impl inh
	void			moveNode(Node*,float,float);    	//impl inh
	void			moveHighlight(float,float);		//impl inh
	Node*			pickNode(SoPickedPoint*,float,float,float); //impl inh
	Edge*			pickEdge(SoPickedPoint*,float,float,float); //impl inh
	SELECT_TYPE		highlight(Node*,SELECT_TYPE=SELECT);    //impl inh
	SELECT_TYPE		highlight(Edge*,SELECT_TYPE=SELECT);    //impl inh
	void			clearHighlight();			//(utility) clear all highlights
	void			setNodeScale(float);		     	//enh  inh
	void			setEdgeResolution(EDGE_RESOLUTION);	//enh  inh
	void			setNodeMapper(IVNodeMapper*);		//set custom node-mapper
	void			setEdgeMapper(IVEdgeMapper*);		//set custom edge-mapper
	IVNodeMapper*		getNodeMapper();			//get current node mapper
	IVEdgeMapper*		getEdgeMapper();			//get current edge mapper
        void                    addAuxNodeMapper(IVNodeMapper*,int=-1); //add/delete/inquire the aux-node-mappers of this 
        void                    delAuxNodeMapper(IVNodeMapper*);        //
        BOOL                    hasAuxNodeMapper(IVNodeMapper*);        //
	
	protected:

	friend class		IVMapperImp;				//IVMapperImp uses a lot of stuff from this
	IVNodeGlyphFactory*	node_glyph;
	IVEdgeGlyphFactory*	edge_glyph;
	SoSeparator*		root;
	SoSeparator*    	base_root;
	std::map<Node*,int>	nodes;					//maps nodes to idxs of their glyphs in 'root'
	std::map<Edge*,int>	edges;					//maps edges to idxs of their glyphs in 'root'
	std::map<Node*,int>	hl_nodes;
	std::map<Edge*,int>	hl_edges;
 
	IVNodeMapper*		node_mapper;				//the current node-mapper
	IVEdgeMapper*		edge_mapper;				//the current edge-mapper
	IVNodeMapper*		def_node_mapper;			//the default (own) node-mapper
	IVEdgeMapper*		def_edge_mapper;			//the default (own) edge-mapper
        DARRAY<IVNodeMapper*>   aux_node_mappers;	
	};


inline float IVMapperBase::getNodeScale() const
{  return scale;  }

inline IVMapperBase::MOVE_STYLE IVMapperBase::getMoveStyle() const
{  return move_style; }

inline IVMapperBase::EDGE_RESOLUTION IVMapperBase::getEdgeResolution() const
{  return edge_res;  }

inline void IVMapperBase::getEdgeSmoothingParams(int& k,float& t) const
{  k = edge_smoothing_subdiv; t = edge_smoothing_tightness; }

inline const char* IVMapperBase::getPosAttrX() const
{  return pos_x;  }

inline const char* IVMapperBase::getPosAttrY() const
{  return pos_y;  }

inline const char* IVMapperBase::getPosAttrZ() const
{  return pos_z;  }

inline const char* IVMapperBase::getPosAttr() const
{  return pos;  }

inline const char* IVMapperBase::getSizeAttrX() const
{  return sz_x;  }

inline const char* IVMapperBase::getSizeAttrY() const
{  return sz_y;  }

inline const char* IVMapperBase::getSizeAttrZ() const
{  return sz_z;  }

inline void IVMapperBase::touch()
{  is_modified = TRUE;  }

inline void IVMapperBase::getHighlightColor(float& r,float& g,float& b)
{  r = highlight_r; g = highlight_g; b = highlight_b; }

inline const char* IVMapperBase::getHighlight() const
{  return highlight_sel; }

inline const char* IVMapperBase::getSelection() const
{  return input_sel; }

inline IVNodeMapper* IVMapper::getNodeMapper()
{  return node_mapper;  }

inline IVEdgeMapper* IVMapper::getEdgeMapper()
{  return edge_mapper;  }

inline BOOL IVMapper::hasAuxNodeMapper(IVNodeMapper* m)
{  return aux_node_mappers.Contains(m)!=-1;  }

#endif
