#ifndef PR_IV_MAPPER_IMP_H
#define PR_IV_MAPPER_IMP_H

//   IVMapperImp:	      Baseclass for the node and edge mapper (IVNodeMapper and IVEdgeMapper). These are
//			      the 'implementation classes' that do the actual mapping of nodes and edges to Inventor	
//			      objects in an IVMapper. An IVMapper uses thus an IVNodeMapper-subclass and an 
//			      IVEdgeMapper-subclass to map its nodes and edges. These concrete subclasses implement
//			      various ways to construct the Inventor stuff from Nodes and Edges.	
//
//

#include "Genrl/types.h"
#include "Genrl/string.h"
#include "Genrl/refcount.h"
#include "Genrl/darray.h"
#include <map>

class	IVMapper;
class	SoGroup;
class	SoSeparator;
class	SoLineSet;
class	SoVertexProperty;
class	SoNode;
class	SoPickedPoint;
class	SoIndexedFaceSet;
class	SoTransform;
class   SoQuadMesh;
class	Selection;
class	Node;
class	Edge;
class	IVNodeGlyphFactory;
class	IVEdgeGlyphFactory;
class	IVHighlightMarker;


class	IVMapperImp : public RefCount				//Implementation baseclass for IVNodeMappers and IVEdgeMappers
	{
	public:

        void			setMainMapper(IVMapper*);	//Bind this to its master, the IVMapper
	virtual void		map(Selection*);		//Map stuff from selection
	virtual void		clearHighlight();		//Clears whatever was highlighted in this
	virtual void		highlight(int,BOOL);		//Highlight ith node/edge to (on|off) state
	virtual void		moveNode(Node*,int,float,float);//Move node (given by ptr and idx) and node's edges
	virtual void		moveHighlight(float,float);	//Move highlight with given (dx,dy) 
	SoGroup*		getSceneGraph();		//Return 'root'

	protected:
				IVMapperImp(IVMapper*);		//Ctor. Makes+refs root
	virtual		       ~IVMapperImp();			//Dtor. Unrefs root.
	friend class		IVHighlightMarker;		//For 'mapper'
	
	std::map<Node*,int>&	nodes();
	std::map<Node*,int>&	hl_nodes();
	std::map<Edge*,int>&	edges();
	std::map<Edge*,int>&	hl_edges();
	

	SoGroup*		root;
	IVMapper*		mapper;
	};



class	IVNodeMapper : public IVMapperImp			//Baseclass for all node mappers	
	{
	public:
			       
        void			map(Selection*);                //Enh inh
	virtual void		setNodeScale(float);		//Set scale of all nodes
	virtual void		delNode(Node*,int);		//Delete node given as ptr and idx
	virtual Node*		pickNode(SoPickedPoint*,float,float,float); //Pick a node given a path to it and a 3D point on it
        static IVNodeMapper*    currentActiveMapper();          //Callbacks can use this to inquire which node-mapper has triggered them
                                                                //(useful e.g. in IVGlyph callbacks that don't get a mapper otherwise)
	protected:
				IVNodeMapper(IVMapper*);	
			       ~IVNodeMapper();			//Dtor
        virtual void            mapImp(Selection*) =0;          //Actual implementation of map(), to be overriden
        void                    shiftNode(Node*,float,float);   //Basic implementation of moveNode() - just updates pos-attrs with dx,dy
        static IVNodeMapper*    active_mapper;                  //Set to the active mapper executing map()
	};


class   IVEdgeMapper : public IVMapperImp			//Baseclass for all edge mappers
	{
	public:
			  
        void			map(Selection*);                //Enh inh
	virtual void		delEdge(Edge*,int);		//Delete edge given as ptr and idx
	virtual Edge*		pickEdge(SoPickedPoint*,float,float,float); //Pick a edge given a path to it and a 3D point on it
        int		        getEdgeCoords(Edge*,const char*,const char*,const char*,float* =0);	
								//Utility: given an Edge and the names of the 3 pos-attrs,
								//extracts the xyz-edge-coords and puts then in the last array arg
								//(if given). Returns # triplets extracted.
                                                                //This can be called by anyone. 
								//We put it here since it doesn't really belong better elsewhere.
        static IVEdgeMapper*    currentActiveMapper();          //Callbacks can use this to inquire which node-mapper has triggered them


	protected:
				IVEdgeMapper(IVMapper*);
			       ~IVEdgeMapper();			//Dtor	
        virtual void            mapImp(Selection*) =0;          //Actual implementation of map(), to be overriden
        static IVEdgeMapper*    active_mapper;                  //Set to the active mapper executing map()
	};


class	IVHighlightMarker					//Utility: implements the highlight-marker used by 
	{							//IVFactoryNodeMapper and IVFactoryEdgeMapper
	public:

				IVHighlightMarker();
                               ~IVHighlightMarker();
	void			setMapper(IVMapperImp*);	
	void			clear();			//Releases this (prior to update or destruction)
	SoGroup*		getSelMarker();			//Get (shared) sel-marker for subclasses
								//This is put here since both FactoryNode and
								//FactoryEdge mappers need it
	private:
	
	SoGroup*		highlight_marker;		//The marker itself
	IVMapperImp*		owner;				//The IVMapperImp owning this
	};



class	IVBasicNodeMapper : public IVNodeMapper		//Node mapper that makes a simple cube per node	
	{
	public:
				IVBasicNodeMapper(IVMapper* =0);	 //Ctor	
	void			clearHighlight();			 //impl inh	
	void			highlight(int,BOOL);			 //impl inh
	void			moveNode(Node*,int,float,float);	 //impl inh
	void			moveHighlight(float,float);		 //impl inh
	void			setNodeScale(float);			 //impl inh
	void			delNode(Node*,int);			 //impl inh
	Node*			pickNode(SoPickedPoint*,float,float,float); //impl inh
	
        protected:

        void			mapImp(Selection*);			 //impl inh

	private:
	
	void			mapNode(Node*,int);
        void                    placeNode(Node*,int);
	std::map<int,Node*>	nodeidx;
	SoVertexProperty*	vprop;
	SoIndexedFaceSet*	faces;
	};



class	IVFactoryNodeMapper : public IVNodeMapper		//Node mapper that uses a IVNodeGlyphFactory per node	
	{
	public:
				IVFactoryNodeMapper(IVMapper* =0);	 //Ctor	
	void			clearHighlight();			 //impl inh	
	void			highlight(int,BOOL);			 //impl inh
	void			moveNode(Node*,int,float,float);	 //impl inh
	void			moveHighlight(float,float);		 //impl inh
	void			setNodeScale(float);			 //impl inh
	void			delNode(Node*,int);			 //impl inh
	Node*			pickNode(SoPickedPoint*,float,float,float); //impl inh
	void			setNodeGlyphFactory(const char*);	 //set the glyph-factory used for making node glyphs.
	const char*		getNodeGlyphFactory();			 //get the node-glyph-factory (def=ret nil)
	
        protected:

        void			mapImp(Selection*);			 //impl inh

	private:
	
	SoNode*			mapNode(Node*);				//uses the glyph-factory to map one node
	SoSeparator*		place(Node*);				//creates SoSeparator for 1 node
	SoGroup*		getNodeGroup(int);			//get unique group for ith Node
	SoTransform*		getNodeTransform(int);			//get transform for ith Node
	SoGroup*		getNodeSelect(int);			//get selection-placeholder for ith Node


	IVHighlightMarker	marker;
	STRING			node_glyph;
	IVNodeGlyphFactory*	cache_glyph_factory;			//local cache for speed
	};




class   IVBasicEdgeMapper : public IVEdgeMapper			//Edge mapper that creates lines for edges	
	{
	public:
				IVBasicEdgeMapper(IVMapper* =0);
	void			clearHighlight();			 //impl inh	
	void			highlight(int,BOOL);			 //impl inh
	void			moveNode(Node*,int,float,float);	 //impl inh
	void			moveHighlight(float,float);		 //impl inh
	void			delEdge(Edge*,int);			 //impl inh
	Edge*			pickEdge(SoPickedPoint*,float,float,float); //impl inh
	
        protected:

        void			mapImp(Selection*);			 //impl inh    

	private:

	void			mapEdge(Edge*,int);			//(re)maps an edge given as ptr and idx
	std::map<int,Edge*>	edgeidx;				//used to map back from idxs to Edges
	SoVertexProperty*	vprop;
	SoLineSet*		lines;
        DARRAY<int>             start;                                  //records idxs where the line-set for each edge starts
	};


class   IVFactoryEdgeMapper : public IVEdgeMapper		//Edge mapper that uses a IVEdgeGlyphFactory per edge
	{
	public:
				IVFactoryEdgeMapper(IVMapper* =0);	 //Ctor
	void			setEdgeGlyphFactory(const char*);	 //set the glyph-factory used for making edge glyphs.
	const char*		getEdgeGlyphFactory();			 //get the edge-glyph-factory (def=ret nil)
	void			clearHighlight();			 //impl inh	
	void			highlight(int,BOOL);			 //impl inh
	void			moveNode(Node*,int,float,float);	 //impl inh
	void			moveHighlight(float,float);		 //impl inh
	void			delEdge(Edge*,int);			 //impl inh
	Edge*			pickEdge(SoPickedPoint*,float,float,float); //impl inh

        protected:

        void			mapImp(Selection*);			 //impl inh

	private:

	SoNode*			mapEdge(Edge*);				//uses the glyph-factory to map one edge
	SoSeparator*		place(Edge*);				//creates SoSeparator for 1 edge
	SoGroup*		getEdgeGroup(int);			//get unique group for ith Edge
	SoNode*			getEdgeGlyph(int);			//get glyph for ith Edge, without transform (may be shared)
	SoGroup*		getEdgeSelect(int);			//get selection-placeholder for ith Edge

	IVHighlightMarker	marker;
	STRING			edge_glyph;
	IVEdgeGlyphFactory*	cache_glyph_factory;			//local cache for speed
	};




class	IVSplatNodeMapper : public IVNodeMapper		//Node mapper that makes a splat field as a quad mesh
	{
	public:
				IVSplatNodeMapper(IVMapper* =0);	 //Ctor	
	void			clearHighlight();			 //impl inh	
	void			highlight(int,BOOL);			 //impl inh
	void			moveNode(Node*,int,float,float);	 //impl inh
	void			moveHighlight(float,float);		 //impl inh
	void			setNodeScale(float);			 //impl inh
	void			delNode(Node*,int);			 //impl inh
	Node*			pickNode(SoPickedPoint*,float,float,float); //impl inh
	
        protected:

                               ~IVSplatNodeMapper();                     //Dtor  
        void			mapImp(Selection*);			 //impl inh

	private:
	
	void			mapNode(Node*,int);
        void                    placeNode(Node*,int);
	SoVertexProperty*	vprop;
	SoQuadMesh*	        mesh;
        int                     verts_per_row,verts_per_col;
        int                     verts_per_dim;
        float                   xm,xM,ym,yM;                            //the 2D bbox of the mapped nodes
        float                   splat_max;                              //max over splat[]
        float*                  splat;                                  //work-array for the mesh splat-values
        int                     splat_sz;                               //allocated size of splat[]
        float                   l2c;                                    //convert from layout to cell coordinates
	};





inline void IVMapperImp::setMainMapper(IVMapper* m)
{ mapper = m; }

inline IVEdgeMapper* IVEdgeMapper::currentActiveMapper()
{  return active_mapper;  }

inline IVNodeMapper* IVNodeMapper::currentActiveMapper()
{  return active_mapper;  }

inline void IVFactoryEdgeMapper::setEdgeGlyphFactory(const char* ef)	//same as above
{   edge_glyph = ef;   }

inline void IVFactoryNodeMapper::setNodeGlyphFactory(const char* nf)	//same as above
{   node_glyph = nf;   }

inline const char* IVFactoryNodeMapper::getNodeGlyphFactory()
{  return node_glyph;  }

inline const char* IVFactoryEdgeMapper::getEdgeGlyphFactory()
{  return edge_glyph;  }

inline void IVHighlightMarker::setMapper(IVMapperImp* m)
{  owner = m;  }

#endif

