#ifndef PR_IV_VIEWER_H
#define PR_IV_VIEWER_H

//	IVViewer:	Implements an Inventor-based viewer. A viewer has both output and input functions:
//			- it displays an input-Selection (nodes+edges) as IV objects. The objects are created by an IVMapper
//			  that the viewer uses. There is a default mapper made, so far.
//			  Together with the input-Selection, a pos-attr-prefix-name is given, specifying the layout attrs.
//			- it displays and allows editing a highlight-Selection as IV objects. This highlight-selection
//			  is a subset of the input-selection. A viewer edits the highlight-selection by an EventHandler.
//			  This is a helper class which maps mouse-events on the displayed objects to edit-events on the
//			  highlight-selection. There is a default EventHandler which allows simple toggle-type selection of
//			  nodes/edges, but one can supply his own EventHandler.
//
//			IVViewers should respond to update/delNode/delEdge messages by
//			updating the view accordingly. So far, these are simply delegated to the viewer's mapper.
//
//			All IVViewers are gathered in a 'viewer collection'. The IVViewer class
//			offers methods to iterate over all its instances existing at any time.
//
//
			
#include "Observer.h"
#include "Genrl/darray.h"
#include "Genrl/types.h"
#include "UI.h"
#include <set>


class	IVMapperBase;
class   IVMapper;
class   IVNodeMapper;
class   IVFactoryNodeMapper;
class	IVBasicNodeMapper;
class   IVSplatNodeMapper;
class   IVFactoryEdgeMapper;
class   IVBasicEdgeMapper;
class   IVZBufferHit;
class	SoSeparator;
class	SoUIViewer;
class   SoAction;
class	SoEventCallback;
class	SoLightModel;
class	EventHandler;
class   Element;



class	IVViewer : public Observer
	{
	public:

	 enum			MAPPER_TYPE 
				{  
				    BASIC   =0,
			            FACTORY =1,
                                    SPLAT   =2
				};
		 
	
	    			IVViewer(const char* vwname, 		//ctor: arg0= own name,
				 	 const char* selname, 		//      arg1= name of sel we display
				 	 const char* posname, 		//	arg2= name of pos-attr-prefix used for layout
				 	 const char* hlname);		//	arg3= name of highlight-sel we display/edit

	 void			update();				//impl inh = redraw graph + highlight Selections
	 void			touch();				//mark this as needing a full update 
									//(needed since viewers depend on other data than Selections only)
	 void			show();					//show viewer window 
	 void			pan(UIViewerBase::PAN_DIRECTION);	//pan the viewer in the X/Y directions
	 void			addButton(const char*,const char*,	//add user-button (icon-filename,Tcl-callback,tooltip-text,
					  const char* ="",BOOL=FALSE);  //checkbox-or-not)
									
	 BOOL			mustBeUpdated() const;			//over inh to monitor both input and highlight selections
	 void			setNodeMapper(MAPPER_TYPE);             //set the mapper-type by making the corresponding own mapper current
	 MAPPER_TYPE		getNodeMapper();
         void                   setAuxNodeMapper(MAPPER_TYPE,BOOL);     //switch on/off, as aux mapper, the own mapper given by the type-arg
         BOOL                   getAuxNodeMapper(MAPPER_TYPE);          //get on/off state of aux mapper of given type
	 void			setEdgeMapper(MAPPER_TYPE);
	 MAPPER_TYPE		getEdgeMapper();
	 IVMapperBase*		getMapper();				//get current mapper	
	 const IVMapperBase*    getMapper() const;	
	 IVFactoryNodeMapper*   getIVFactoryNodeMapper();		//get this's own IVFactoryNodeMapper	
	 IVBasicNodeMapper*     getIVBasicNodeMapper();			//get this's own IVBasicNodeMapper
	 IVFactoryEdgeMapper*   getIVFactoryEdgeMapper();		//get this's own IVFactoryEdgeMapper	
	 IVBasicEdgeMapper*     getIVBasicEdgeMapper();			//get this's own IVBasicEdgeMapper
         IVSplatNodeMapper*     getIVSplatNodeMapper();                 //get this's own IVSplatNodeMapper
	 void			setUserCallback(const char*);           //set Tcl-cb to be called as user-action (upon selection in this)
	 void			setSweepCallback(const char*);          //set Tcl-cb to be called as sweep-action (upon mouse-motion in this)
	 STRING			getUserCallback();
	 STRING			getSweepCallback();


	 void			setHighlightSelection(const char*);	//specify which nodes/edges to be highlighted
	 const char*		getHighlightSelection() const;		//
	 void			setHighlightColor(float,float,float);	//set/get RGB color used tor highlighting stuff in this
	 void			getHighlightColor(float&,float&,float&);
	 void			setBackgroundColor(float,float,float);	//set/get RGB background color of viewer canvas
	 void			getBackgroundColor(float&,float&,float&);
	 void			set2DRendering(BOOL);			//set/get the 2D look (no depth & shading)
	 BOOL			get2DRendering() const;
	 void			setOutputOnly(BOOL);			//set/get output-only viewer state
	 BOOL			getOutputOnly() const;
	 void			setPosAttr(const char*);		//specify pos-attr-prefix-name
	 const char*		getPosAttr() const;			//get pos-attr prefix-name that this uses
	 Node*			pickNode(float,float,IVZBufferHit* =0);	//return Node whose glyph is closest to given screen-pt (in norm-coords)
	 Edge*			pickEdge(float,float,IVZBufferHit* =0);	//return Edge whose glyph is closest to given screen-pt (in norm-coords)
	 void			pickNodes(Selection*,float,float,float,float);
	 								//add Nodes whose 2D positions fall within bbox, given
	 								//in screen norm coords, to given Selection.
	 SoUIViewer*    	getViewer();				//return this's Inventor viewerimplementation
	 EventHandler*  	getEventHandler();			//return this's event-handler

									//VIEWER-LIST QUERY:
	 static int 		numViewers();				//returns # instances of this
	 static IVViewer*	getViewer(int);				//return ith instance of this
	 static IVViewer*	getViewer(const char*);			//return viewer by name or null
	 static IVViewer*	getViewer(const SoUIViewer*);		//return viewer by implem-object
         static void            save(FILE*,std::set<STRING>* =0);       //Save viewer settings to Tcl-file
                                                                        //Also gather Tcl-procs this depends on in arg2, if given
	
	protected:

         friend class           DefaultEventHandler;                    //for store_xxx
	
		       	       ~IVViewer();				//dtor: not public, this should be deleted as an Observer only
	 void			reset();
	 void			setMapper(IVMapperBase*);		//specify custom mapper
         Element*               pickElement(float,float,IVZBufferHit*,BOOL); //implements Node/Edge picking, depending on last arg (T/F)
	 static void		event_cb(void*,SoEventCallback*);	//dispatches mouse events in this
         static void            render_cb(void* p,SoAction* a);         //various data-mgmt to be done before rendering this
         static const char*     make_name(const char* n);               //make unique viewer name        
	
	 IVMapperBase*		mapper;					//current mapper in use
	 IVMapper*		def_mapper;				//default mapper created by this is an IVMapper
	 IVFactoryNodeMapper*   fact_node_mapper;			//the following are several concrete, specialized
	 IVFactoryEdgeMapper*   fact_edge_mapper;			//mapper subcomponents owned by the viewer. The viewer	
	 IVBasicNodeMapper*	basic_node_mapper;                      //creates all of these and lets users switch between the one
	 IVBasicEdgeMapper*     basic_edge_mapper;			//which is in use at a given time.        
         IVSplatNodeMapper*     splat_node_mapper;                      //
									
	 SoSeparator*		root;
	 SoSeparator*   	base_root;
	 SoLightModel*		light_model;			
	 SoUIViewer*		viewer;					//the implem of the IV viewer this uses
	 UIViewerBase*		ui_viewer;				//the implem of the UIViewerBase (identical to the 'viewer' obj but
									//needed here due to the multiple inheritance, see UI.h)
	 BOOL			output_only;				//tells if the viewer is output-only or also allows contents editing
	 BOOL			render_2d;				//tells if this has the 2D or 3D render/shading look
	 STRING			pos_name;				//the prefix-name for the pos-attr this uses to lay out nodes/edges
	 STRING			highlight_name;				//the name of the output highlight-selection
	 EventHandler*  	evt_handler;				//the event-handler this uses to handle mouse events
         BOOL                   rendered;                               //TRUE after this has rendered itself, might be invalidated by direct OpenGL rendering
         unsigned char*         store;                                  //!!!
         int                    store_x,store_y;
         int                    store_sz,store_dx,store_dy;
	 static DARRAY<IVViewer*>
	 			viewers;				//all instances of this
	};





inline IVMapperBase* IVViewer::getMapper()
{  return mapper;  }

inline const IVMapperBase* IVViewer::getMapper() const
{  return mapper;  }

inline IVFactoryNodeMapper* IVViewer::getIVFactoryNodeMapper()
{  return fact_node_mapper;  }

inline IVBasicNodeMapper* IVViewer::getIVBasicNodeMapper()
{  return basic_node_mapper;  }

inline IVSplatNodeMapper* IVViewer::getIVSplatNodeMapper()
{  return splat_node_mapper;  }

inline IVFactoryEdgeMapper* IVViewer::getIVFactoryEdgeMapper()
{  return fact_edge_mapper;  }

inline IVBasicEdgeMapper* IVViewer::getIVBasicEdgeMapper()
{  return basic_edge_mapper;  }

inline const char* IVViewer::getPosAttr() const
{  return pos_name;  }
			
inline const char* IVViewer::getHighlightSelection() const
{  return highlight_name;  }

inline void IVViewer::setHighlightSelection(const char* hn)		
{  highlight_name = hn; }					
									
inline SoUIViewer* IVViewer::getViewer()
{  return viewer;  }

inline EventHandler* IVViewer::getEventHandler()
{  return evt_handler;  }

inline void IVViewer::setPosAttr(const char* pn)
{  pos_name = pn;  }

inline int IVViewer::numViewers()
{  return viewers.Count();  }

inline IVViewer* IVViewer::getViewer(int i)
{  return viewers[i];  }

inline void IVViewer::setOutputOnly(BOOL b)
{  output_only = b; }

inline BOOL IVViewer::getOutputOnly() const
{  return output_only;  }

inline BOOL IVViewer::get2DRendering() const
{  return render_2d;  }

#endif
