#ifndef PR_LAYOUT_OP_H
#define PR_LAYOUT_OP_H

//	LAYOUT 
//	OPERATIONS:	These operations take usually a Selection and set its position-attributes. 
//			Available implementations:	
//
//			LayoutOp:			  Baseclass for graph layouts. layout() computes/refines 2D/3D node/edge coords in [0..1].
//			GridLayoutOp:     		  Lays nodes on 2D regular grid.
//			GVLayoutOp:			  Layouts based on the GraphViz package.
//			DotLayoutOp:			  Directed Sugiyama layout based on DOT.
//			NeatoLayoutOp:			  Spring embedder layout based on NEATO.
//			LevelLayoutOp:			  Lays out nodes on same level in 2D with NEATO, then stacks levels in 3D.
//			GEMLayoutOp:			  Spring embedder layout using the GEM package.
//			TSLayoutOp:			  Affine xform (transl/scale) layout op
//
//

class	Node;
class	Selection;
class   DirectionFilter;

class   LayoutOp : public Operation						//baseclass for layout-algos
			{
			public:
			
					LayoutOp();
			virtual	       ~LayoutOp();
			void		setBoundingBox(float ox,float oy,float dx,float dy);
			void		useBoundingBox(BOOL);			//Use/ignore bbox settings
			float		getBBoxOx()     const;			//Get bounding-box parameters
			float		getBBoxOy()     const;			//(origin and size)
			float		getBBoxDx()	const;			//
			float		getBBoxDy()	const;  		//	
			float		getBBoxOz()	const;			//
			float		getBBoxDz()	const;			//
			void		keepDimensions(BOOL);			//to be called by client to indicate that
										//dim-attrs are not to be set by this layout
			BOOL		getKeepDimensions() const;		//return value of keep_dims						
			static void	setDefaultKeepDimensions(BOOL);		//default (class-wide) version of above
			static BOOL	getDefaultKeepDimensions();
			static void	setDefaultNodeSize(float);		//default node-size as fraction of layout-bbox
			static float	getDefaultNodeSize();
			static void     setDefaultExtLayoutUseCrtPos(BOOL);	//indicates that when an external layout-server is called, the current-values
			static BOOL     getDefaultExtLayoutUseCrtPos();		//of the position-attr are passed as initial state. Some layouts (e.g. NEATO)
										//give different results if they start from a given position or not.
			static void     setDefaultExtLayoutUseCrtDim(BOOL);	//as above, but for the dimension attribute.
			static BOOL     getDefaultExtLayoutUseCrtDim();
			static void	setDefaultLayEdges(BOOL);		//indicates whether the edges should be layed out or not. Some layouts can do this and
			static BOOL	getDefaultLayEdges();			//will produce edge-pos-attrs, but will probably work more for this. If edges are not
										//layed out, mappers will assume edges connect their nodes' positions.
			
			int		layout(Selection*, const char* pos);
			int		layout(const char*,const char* pos);
							
			protected:						
			
			virtual int	layout() =0;		//actual layout implementation
			virtual BOOL    supportsPinning();	//indicates that the subclass has
								//node-pinning capabilities (default=FALSE)

			Selection*	getSelection() const;	//subclasses can inquire on the
			const char*     getPositionX() const;	//layout(sel,posname)'s args
			const char*     getPositionY() const;	//
			const char*     getPositionZ() const;	//
			const char*     getDimensionX() const;	//
			const char*     getDimensionY() const;	//
			const char*     getDimensionZ() const;	//
			const char*	getPosition()  const;	//
			const char*     getPinning()   const;	//

			private:
			
			Selection*	sel;			//caches args of layout(sel,posname)		
			STRING		pos,pos_x,pos_y,	//for subclasses to use
					pos_z,pos_pin,dim_x,	
					dim_y,dim_z;		
			float		bbox_ox,bbox_oy,bbox_oz,
					bbox_dx,bbox_dy,bbox_dz;
			BOOL		use_bbox;		//		
			BOOL		keep_dims;		//if TRUE, dim-attrs are not to be overwritten
								//Def=FALSE, subclasses/callers may customize. 
			static BOOL	def_keep_dims;		//default (class-level) version of above, initialized to FALSE
			static float	def_node_size;		//default node-size, when layout computes node-sizes as fraction of
								//the total bbox (see LayoutOp::layout())
			static BOOL	def_ext_lay_use_pos;	//if TRUE, crt pos-attr passed as input to external layout servers. Def=FALSE
			static BOOL	def_ext_lay_use_dim;	//if TRUE, crt dim-attr passed as input to external layout servers. Def=FALSE					
			static BOOL	def_lay_edges;		//if TRUE, layout should try to produce edge-positions too
			};
			


class   GridLayoutOp : public LayoutOp				//2D regular grid layout
			{
			public:
		
					GridLayoutOp();			//Def ctor
					GridLayoutOp(int ne,int ax);  	//Ctor
			void		setGrid(int ne);	 	//set #elems-on-major-axis
			void		setAxis(int ax);		//set major-axis 
			static void	setDefaultGrid(int ne);		//def-version of setGrid()
			static void	setDefaultAxis(int ax);		//def-version of setAxis()
			static int	getDefaultGrid();		//ret def_nelems
			static int	getDefaultAxis();		//ret def-axis
			
			protected:
			
			int		layout();		//impl inh
			int		nelems;			//# elems along major axis
			int		axis;			//major axis (0==x,1==y)
			static int	def_nelems;		//default # elems along major axis
			static int	def_axis;		//default major axis (0==x,1==y)
			};



class  	GVLayoutOp : public LayoutOp				//stuff common to dot/neato
			{
			public:
								
 		        static void	init(const char* dir);	 //spec where dot/neato execs are located
			
			protected:
			
					GVLayoutOp();
			int		layout();		 //impl inh
			int		read_layout(Selection*,const char* file);
			virtual const char*	
					progName() const =0;
			virtual BOOL	flipX();
			virtual BOOL	flipY();
			virtual BOOL	flipEdges();
			virtual const char*
					levelAttr();
					
			static STRING	tmp_infile,tmp_outfile;
			static STRING   gv_path;
			};				


class 	NeatoLayoutOp : public GVLayoutOp
			{
			public:			
					NeatoLayoutOp();
			static void	setDefaultWeight(float);		
			static float	getDefaultWeight();
			static void	setDefaultEpsilon(float);
			static float	getDefaultEpsilon();
			static void	setDefaultRandom(BOOL);
			static BOOL	getDefaultRandom();
			static void	setDefaultOverlap(BOOL);
			static BOOL	getDefaultOverlap();
			static void	setDefaultMaxIter(int);
			static int	getDefaultMaxIter();
			static void	setDefaultEdgeLength(float);	 
			static float	getDefaultEdgeLength();
			static void	setDefaultCompact(BOOL);
			static BOOL	getDefaultCompact();
					
			BOOL		supportsPinning();		//over inh: returns TRUE		
			protected:
			int		layout();			//enh inh with compaction feature
			const char*	progName() const;		
			void		compact(int);			//post-layout compaction of nodes
			BOOL		testOverlap(Selection*,Node*,float cx,float cy);
			
			float		weight;				//desired edge weight
			float		epsilon;
			BOOL		random;	
			static float	def_weight;
			static float	def_epsilon;
			static BOOL	def_random;
			static int	def_maxiter;
			static BOOL	def_overlap;
			static BOOL	def_compact;
			static float	def_edgelen;
			};			


class 	DotLayoutOp : public GVLayoutOp
			{
			public:			
					DotLayoutOp();		
					
			static void	setDefaultNodeSep(float);
			static float	getDefaultNodeSep();
			static void	setDefaultRankSep(float);
			static float	getDefaultRankSep();
			static void	setDefaultRankDir(int);
			static int	getDefaultRankDir();
			static void	setDefaultEdgeFlipping(BOOL);
			static BOOL	getDefaultEdgeFlipping();
			static void	setDefaultLevelAttr(const char*);
			static const char*
					getDefaultLevelAttr();					
			protected:
			
			const char*	progName() const;
			BOOL		flipX();
			BOOL		flipY();
			BOOL		flipEdges();
			const char*	levelAttr();
			
			static int	def_rankdir;
			static float	def_nodesep;
			static float	def_ranksep;
			static BOOL	def_flip_edges;
			static STRING   def_level_attr;
			};			


class  	GEMLayoutOp : public LayoutOp	
		      {
		      public:
								
		      static void	init(const char* gemdir);	//spec where GEM is installed

		      protected:

		      int		layout();			//impl inh	
		      int		read_layout(const char* file);
		      const char*       progName() const;				    
		      static STRING	tmp_infile,tmp_outfile;
		      static STRING	gem_dir;
		      };				


			
class	LevelLayoutOp : public LayoutOp
			{
			public:
			
			enum 		SPACING_METHOD
					{
					   ABSOLUTE  =0,
					   NORMALIZED=1
					};   
			
					LevelLayoutOp();
			static void	setDefaultSpacing(float);
			static float	getDefaultSpacing();
			static void	setDefaultSpacingMethod(SPACING_METHOD);
			static SPACING_METHOD	
					getDefaultSpacingMethod();
			static void	setDefaultLevelAttr(const char*);
			static const char*	
					getDefaultLevelAttr();

			protected:

			int		layout();			//impl inh
			static STRING	def_level;
			static float	def_spacing;
			static SPACING_METHOD  
					def_spacing_method;
			};			


class	NestedLayoutOp : public LayoutOp
			{
			public:
			
					NestedLayoutOp();
			static void	setXYBorder(float);		//set tightness of nesting in XY plane
			static float	getXYBorder();			//(0==none, 1==max)
			static void	setZBorder(float);		//set tightness of nesting in Z dimension
			static float	getZBorder();			//(0==none, 1==max)
			static void	setDoZLayout(BOOL);		//if set, performs Z layout as function of nesting-level
			static BOOL	getDoZLayout();			//if not, the layout gets a Z==0 coordinate
			static void	setFilter(DirectionFilter*);	//set way to use to traverse to descendants

			protected:

			int		layout();			//impl inh
			void 		layout(Node*,int);	
			void 		layoutNodes(Selection*,float&,float&,int);
			void 		normalize(Selection*,float,float,float,float);
                        void            layoutCrossEdges();             //lays out edges in 'cross' separately

			static float	xyborder,zborder;
			static BOOL     do_z_layout;
			BOOL		detect_bottom;
			float		xyborder_abs;
			int		maxlevel;			//for normalization of Z dimension 
			Selection*	tree;				//filters the input selection to a strict tree
                        Selection*      cross;                          //holds the assoc-edges between nodes in different parents
			static DirectionFilter				//for determining containment-relations		
				        *in_filter,*out_filter;	
			};			

class	TSLayoutOp : public LayoutOp
			{
			public:
			
					TSLayoutOp();
			void		setTransl(float tx,float ty,float tz=0);	
			void		setScale(float sx,float sy,float sz=1);
			
			protected:
			
			int		layout();			//impl inh
			float		tx,ty,tz,sx,sy,sz;		//transl + scaling params
			
			};
						
//-----------  Inline implementations  --------------------------------


inline LayoutOp::LayoutOp()				//Ctor: init def-bbox to [0..1,0..1]
		:bbox_ox(0), bbox_oy(0), 		//	and use it for layout.
		 bbox_dx(1), bbox_dy(1),
		 bbox_oz(0), bbox_dz(1),
		 use_bbox(TRUE),
		 keep_dims(def_keep_dims)		//By default, layouts set both dim- and pos-attrs.
{  }

inline Selection* LayoutOp::getSelection() const
{  return sel;  }

inline const char* LayoutOp::getPinning() const
{  return pos_pin;  }

inline const char* LayoutOp::getPositionX() const
{  return pos_x;  }

inline const char* LayoutOp::getPositionY() const
{  return pos_y;  }

inline const char* LayoutOp::getPositionZ() const
{  return pos_z;  }

inline const char* LayoutOp::getPosition() const
{  return pos;  }

inline const char* LayoutOp::getDimensionX() const
{  return dim_x;  }

inline const char* LayoutOp::getDimensionY() const
{  return dim_y;  }

inline const char* LayoutOp::getDimensionZ() const
{  return dim_z;  }

inline float LayoutOp::getBBoxOx() const
{  return bbox_ox;  }

inline float LayoutOp::getBBoxOy() const
{  return bbox_oy;  }

inline float LayoutOp::getBBoxDx() const
{  return bbox_dx;  }

inline float LayoutOp::getBBoxDy() const
{  return bbox_dy;  }

inline float LayoutOp::getBBoxDz() const
{  return bbox_dz;  }

inline float LayoutOp::getBBoxOz() const
{  return bbox_oz;  }

inline void LayoutOp::setBoundingBox(float ox,float oy,float dx,float dy)
{  bbox_ox=ox; bbox_oy=oy; bbox_dx=dx; bbox_dy=dy;  }

inline void LayoutOp::useBoundingBox(BOOL b)
{  use_bbox = b;  }

inline void LayoutOp::keepDimensions(BOOL b)
{  keep_dims = b;  }

inline BOOL LayoutOp::getKeepDimensions() const
{  return keep_dims;  }

inline void LayoutOp::setDefaultKeepDimensions(BOOL b)
{  def_keep_dims = b;  }

inline BOOL LayoutOp::getDefaultKeepDimensions()
{  return def_keep_dims;  }

inline void LayoutOp::setDefaultNodeSize(float b)
{  def_node_size = b;  }

inline float LayoutOp::getDefaultNodeSize()
{  return def_node_size;  }

inline void  LayoutOp::setDefaultExtLayoutUseCrtPos(BOOL b)
{  def_ext_lay_use_pos = TRUE;  }

inline BOOL  LayoutOp::getDefaultExtLayoutUseCrtPos()
{  return def_ext_lay_use_pos; }

inline void  LayoutOp::setDefaultExtLayoutUseCrtDim(BOOL b)
{  def_ext_lay_use_dim = TRUE;  }

inline BOOL  LayoutOp::getDefaultExtLayoutUseCrtDim()
{  return def_ext_lay_use_dim; }

inline void  LayoutOp::setDefaultLayEdges(BOOL b)
{  def_lay_edges = TRUE;  }

inline BOOL  LayoutOp::getDefaultLayEdges()
{  return def_lay_edges; }



inline void DotLayoutOp::setDefaultNodeSep(float f)
{  def_nodesep = f; }

inline float DotLayoutOp::getDefaultNodeSep()
{  return def_nodesep; }

inline void DotLayoutOp::setDefaultRankSep(float f)
{  def_ranksep = f; }

inline float DotLayoutOp::getDefaultRankSep()
{  return def_ranksep; }

inline void DotLayoutOp::setDefaultRankDir(int i)
{  def_rankdir = i; }

inline int DotLayoutOp::getDefaultRankDir()
{  return def_rankdir; }

inline BOOL DotLayoutOp::getDefaultEdgeFlipping()
{  return def_flip_edges; }

inline void DotLayoutOp::setDefaultEdgeFlipping(BOOL f)
{  def_flip_edges = f; }

inline const char* DotLayoutOp::getDefaultLevelAttr()
{  return def_level_attr; }

inline void DotLayoutOp::setDefaultLevelAttr(const char* la)
{  def_level_attr = la; }


inline void NeatoLayoutOp::setDefaultEpsilon(float e)
{  def_epsilon = e;  }

inline float NeatoLayoutOp::getDefaultEpsilon()
{  return def_epsilon;  }

inline void NeatoLayoutOp::setDefaultMaxIter(int i)
{  def_maxiter = i;  }

inline int NeatoLayoutOp::getDefaultMaxIter()
{  return def_maxiter;  }

inline void NeatoLayoutOp::setDefaultWeight(float w)
{  def_weight = w;  }

inline float NeatoLayoutOp::getDefaultWeight()
{  return def_weight;  }

inline void NeatoLayoutOp::setDefaultRandom(BOOL r)
{  def_random = r;  }

inline BOOL NeatoLayoutOp::getDefaultRandom()
{  return def_random;  }

inline void NeatoLayoutOp::setDefaultOverlap(BOOL r)
{  def_overlap = r;  }

inline BOOL NeatoLayoutOp::getDefaultOverlap()
{  return def_overlap;  }

inline void NeatoLayoutOp::setDefaultEdgeLength(float e)
{  def_edgelen = e;  }

inline float NeatoLayoutOp::getDefaultEdgeLength()
{  return def_edgelen;  }

inline void NeatoLayoutOp::setDefaultCompact(BOOL c)
{  def_compact = c;  }

inline BOOL NeatoLayoutOp::getDefaultCompact()
{  return def_compact;  }

//-------------------  LevelLayoutOp  -------------------------

inline LevelLayoutOp::LevelLayoutOp()
{  }

inline void LevelLayoutOp::setDefaultLevelAttr(const char* l)
{  def_level = l;  }

inline const char* LevelLayoutOp::getDefaultLevelAttr()
{  return def_level;  }

inline void LevelLayoutOp::setDefaultSpacing(float f)
{  def_spacing = f;  }

inline float LevelLayoutOp::getDefaultSpacing()
{  return def_spacing;  }

inline LevelLayoutOp::SPACING_METHOD LevelLayoutOp::getDefaultSpacingMethod()
{  return def_spacing_method;  }

inline void LevelLayoutOp::setDefaultSpacingMethod(SPACING_METHOD sm)
{  def_spacing_method = sm;  }

inline int GridLayoutOp::getDefaultGrid()
{  return def_nelems; }

inline int GridLayoutOp::getDefaultAxis()
{  return def_axis; }

inline NestedLayoutOp::NestedLayoutOp()
{   
   keepDimensions(TRUE);			//We do set the node-dims ourselves.
}

inline void NestedLayoutOp::setXYBorder(float b)
{  xyborder = b;  }

inline float NestedLayoutOp::getXYBorder()
{  return xyborder;  }

inline void NestedLayoutOp::setZBorder(float b)
{  zborder = b;  }

inline float NestedLayoutOp::getZBorder()
{  return zborder;  }

inline void NestedLayoutOp::setDoZLayout(BOOL b)
{  do_z_layout = b;  }

inline BOOL NestedLayoutOp::getDoZLayout()
{  return do_z_layout;  }

inline void TSLayoutOp::setTransl(float tx_,float ty_,float tz_)
{  tx = tx_; ty = ty_; tz=tz_;  }

inline void TSLayoutOp::setScale(float sx_,float sy_,float sz_)
{  sx = sx_; sy = sy_; sz = sz_;  }



#endif


