#ifndef PR_OBSERVER_H
#define PR_OBSERVER_H

//	Observer:	An Observer attaches itself to a selection-name that
//			that names a Selection in the SelectionList.
//			The Observer monitors the selection's contents in some way, e.g. by displaying it.
//			The specific implementation of the monitoring is done by the concrete Observer subclasses in update().
//			However, update() refreshes the whole Observer from its Selection. To do this more gradually,
//			Observers can implement delNode/delEdge, which are automatically called by the system when a node/edge/level
//			is deleted. The Observer can then refresh its state partially.
//			The Observer refers to selections _by name_ and not directly. 
//			This allows other parts of the system to e.g. delete a Selection, create a new one
//			with same name, and let the Observer mechanism still work.
//
//

#include "GVObject.h"
#include "Genrl/string.h"
#include "Genrl/types.h"


class	Node;
class	Edge;
class   Selection;
class   ObserverList;

class	Observer : public GVObject
	{
	public:
	
	void		setSelection(const char*);		//specifies which Selection-obj we observe
	const char*	getSelection() const;
	virtual void	update() =0;			        //UPDATE METHODS (ONLY FOR SelectionList)
	BOOL		persistent() const;			//Returns persistence flag (see ObserverList)	
        void            setPersistent(BOOL);			//Sets persistence flag (see ObserverList)
	virtual BOOL    mustBeUpdated() const;			//Tells if this must be updated - the logic within is based on a combination of
								//alwaysUpdate() and the selection(s) this monitors being modified
								//May be overriden if subclasses monitor more than 1 selection and/or other data
	
	protected:
	
	friend class	ObserverList;				//for ~Observer
			Observer(const char* name,		//ctor: arg0=name of this, arg1=name of sel we observe
				 const char* selname);
	void		updateAlways(BOOL);			//Marks this to be always updated or only on selection-change
	virtual	       ~Observer();				//dtor: not public, this should be deleted only via ObsList
	BOOL		alwaysUpdate() const;			//This Observer is to be updated even if its selection didn't change
								//(e.g. because it monitors something indirectly...)

	private:

	STRING		sel;					//name of sel we observe
	BOOL	        persist;				//persistence flag
	BOOL		always_update;				//update type: on selection change or always
	};


		
inline void Observer::setSelection(const char* s)
{  sel = s;  }

inline const char* Observer::getSelection() const
{  return sel;  }

inline BOOL Observer::persistent() const
{  return persist;  }

inline void Observer::setPersistent(BOOL b)
{  persist = b;  }

inline BOOL Observer::alwaysUpdate() const
{  return always_update;  }

inline void Observer::updateAlways(BOOL b)
{  always_update = b;  }
			

#endif
