#ifndef PR_SELECTION_H
#define PR_SELECTION_H

#include "Genrl/types.h"
#include "Genrl/darray.h"
#include "Attr.h"
#include "GVObject.h"

//	Selection:	A selection is a set of nodes and edges.
//			One can iterate over a Selection to set/get its nodes/edges.
//			Selections should be able to update themselves when the Graph structure changes. For this, they should
//			implement the delNode/delEdge messages. These are called automatically to ask Selections to remove from
//			themselves given nodes/edges. Selections are AttrSets, so one can set 'global' attr-values
//			on a Selection, without setting them on the nodes/edges contained in a Selection. This is useful
//			for specifying properties of the selection itself.
//
//

class Node;
class Edge;
class SelectionList;
class AttrSetType;

#include "AttrSet.h"
#include <map>
#include <iostream.h>


class Selection	 : public GVObject, public AttrSet 
		  {
		  public:
		  static const char*					       //Use this as a name arg to indicate that	
		  		MAKE_NAME;				       //the Selection must create its own, unique name
		  
		  	 	Selection(const char* =MAKE_NAME);             //Ctor (optional name). No type is given - type is not important.
		  virtual      ~Selection();				       //Dtor
		  void		reset();				       //Releases all nodes/edges in this
		  void		flush();				       //Destroys all nodes/edges in this	
                  void		ref(const char* =MAKE_NAME);  		       //Inserts this in the SelectionList under given name
		  
		  int		numNodes() const;			       //get # nodes in selection			
		  int		numEdges() const;			       //get # edges in selection
		  BOOL		nodesHaveAttr(const char*,Attr::TYPE* =0);     //check that all selected nodes have attr named arg1 
		  							       //optionally check if the attr-type is arg2
		  BOOL		edgesHaveAttr(const char*,Attr::TYPE* =0);     //check that all selected edges have attr named arg1 
		  							       //optionally check if the attr-type is arg2
		  BOOL		hasNode(Node*) const;			       //get if this has a given Node (def=with iterators)
		  BOOL          hasEdge(Edge*) const;			       //get if this has a given Edge (def=with iterators)
		  void		delNodeAttr(const char*);		       //delete given attr from all nodes of this
		  void		delEdgeAttr(const char*);		       //delete given attr from all edges of this
		  
		  class Iterator						//Baseclass for node/edge iterators. 
				{public:					
				 virtual ~Iterator();				//Needed for destructing via Iterator*'s
				 protected:
					  Iterator(Selection*);
					  Selection* sel; };
		  class NodeIterator : public Iterator			       //Iterator over nodes
		  		{public:  Node* next();   
		  			  NodeIterator(Selection*); 
				 friend class Selection;
		  		 private: std::map<int,Node*>::iterator i;
                                };

		  class EdgeIterator : public Iterator			       //Iterator over edges
		  		{public:  Edge* next();   
		  			  EdgeIterator(Selection*); 
				 friend class Selection;
		  		 private: std::map<int,Edge*>::iterator i;
                                };    
		  		
		  NodeIterator* iterNodes();				       //Return specific iterator for nodes
		  EdgeIterator* iterEdges();				       //Return specific iterator for edges
		  Node*		getNodeById(int);			       //Get node by id - may be faster than an iterative search
		  Node*		getNodeByAttr(const char*,const char*);        //Get node that has attr <arg1> equal to <name2>; returns 0 if not found
		  Edge*		getEdgeById(int);			       //Get edge by id - may be faster than an iterative search
		  Edge*		getEdgeByAttr(const char*,const char*);        //Get edge that has attr <arg1> equal to <name2>; returns 0 if not found


		  
		  								//EDIT METHODS
		  virtual void	addNode(Node*);					//Add Node to this 
		  virtual void	addEdge(Edge*);					//Add Edge to this
		  void		delNode(Node*);					//Remove Node from this	
		  void		delEdge(Edge*);					//Remove Edge from this
		  static const char				
		  		*ALL, *NONE;					//reserved names of the full and empty selections
		  		  		  				  
		  protected:
		  
		  friend class	SelectionList;					//For ~Selection,isReservedName,saveDefinition
		  friend class  NodeIterator;
		  friend class  EdgeIterator;
		  friend class  Node;
		  friend class	Edge;						//For removeEdge()
		  
		  void          saveDefinition(FILE*);                          //Save definition of this in *.sel file
		  static  BOOL	isReservedName(const char*);			//Tells if arg is a reserved-sel-name such as ALL, NONE
		  virtual BOOL  isReserved() const;				//Tells if this is a reserved-object (i.e an AllSelection or NoneSelection)
										//Default=False, subclasses may override.
		  void		removeEdge(Edge*);				//Same as delEdge() but without informing Edge further
		  
		  
		  private:  
		  
		  const char*   make_name(const char*,int&);			//Generate conforming names from any strings
		  
		  std::map<int,Node*> node_map;                                 //Store nodes and edges in this for fast lookup via id.
                  std::map<int,Edge*> edge_map;                                 //Useful for accelerating getNode/EdgeById()
		  BOOL		in_dtor;					//Flag for accelerating destruction process
		  int		name_id;					//Unique id among all Selection objects
		  
		  static int	make_name_id;					//Used to make unique name_ids
		  static int	num_selections;					//Total # Selections, used to reset make_name_id
		  };	 		


class NoneSelection : public Selection						//Special empty-selection
		  {
		  public:   

				NoneSelection();				//Ctor (name=="None")
		  void		addNode(Node*);					//Over inh: nop
		  void		addEdge(Edge*);					//Over inh: nop

		  protected:

		  BOOL		isReserved() const;				//True
		  };


class AllSelection : public Selection						//Special full-selection
		  {
		  public:
				AllSelection();

		  protected:
			  
		  BOOL		isReserved() const;				//True
		  };



//-----------------  Selection  --------------------------------


inline BOOL Selection::isReservedName(const char* n)
{
  if (!n) return FALSE;
  return (!strcmp(n,ALL) || !strcmp(n,NONE));
}  


//------------------  Selection  -------------------------

inline int Selection::numNodes() const
{  return node_map.size();  }

inline int Selection::numEdges() const
{  return edge_map.size();  }

inline Selection::NodeIterator* Selection::iterNodes()
{  return new NodeIterator(this);   }

inline  Selection::Iterator::Iterator(Selection* s): sel(s)
{  }

inline Selection::NodeIterator::NodeIterator(Selection* s): Iterator(s),i(s->node_map.begin())
{  }

inline Node* Selection::NodeIterator::next()                    //Return 'next' node. Actually returns the
{                                                               //'current' node in iterator i and then advances i,
   Node* ret = (i!=sel->node_map.end())? (*i).second : 0;       //if not at end. If at end, this is deleted.
   if (!ret) delete this; else i++;
   return ret;
}

inline Selection::EdgeIterator* Selection::iterEdges()
{  return new EdgeIterator(this);   }

inline Selection::EdgeIterator::EdgeIterator(Selection* s): Iterator(s),i(s->edge_map.begin())
{  }

inline Edge* Selection::EdgeIterator::next()                    //Return 'next' edge. See NodeIterator::next()
{  
   Edge* ret = (i!=sel->edge_map.end())? (*i).second : 0;
   if (!ret) delete this; else i++;
   return ret;
}

//------------------  AllSelection, NoneSelection  -------------------------

inline AllSelection::AllSelection(): Selection(ALL)
{  }

inline NoneSelection::NoneSelection(): Selection(NONE)
{  }

#endif
